package lcsb.mapviewer.converter.model.sbgnml;

import java.awt.Color;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import javax.xml.bind.JAXBException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLStreamException;

import org.apache.commons.lang3.RandomStringUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.sbgn.ArcClazz;
import org.sbgn.GlyphClazz;
import org.sbgn.Language;
import org.sbgn.RenderUtil;
import org.sbgn.bindings.Arc;
import org.sbgn.bindings.Arc.End;
import org.sbgn.bindings.Arc.Next;
import org.sbgn.bindings.Arc.Start;
import org.sbgn.bindings.Bbox;
import org.sbgn.bindings.Glyph;
import org.sbgn.bindings.Label;
import org.sbgn.bindings.Map;
import org.sbgn.bindings.Port;
import org.sbgn.bindings.RenderInformation;
import org.sbgn.bindings.SBGNBase;
import org.sbgn.bindings.Sbgn;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import lcsb.mapviewer.common.comparator.DoubleComparator;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.common.geometry.ColorParser;
import lcsb.mapviewer.converter.ConverterException;
import lcsb.mapviewer.converter.annotation.XmlAnnotationParser;
import lcsb.mapviewer.converter.graphics.bioEntity.element.species.SpeciesConverter;
import lcsb.mapviewer.converter.graphics.bioEntity.reaction.ReactionConverter;
import lcsb.mapviewer.converter.model.celldesigner.CommonXmlParser;
import lcsb.mapviewer.model.LogMarker;
import lcsb.mapviewer.model.ProjectLogEntryType;
import lcsb.mapviewer.model.graphics.ArrowType;
import lcsb.mapviewer.model.graphics.PolylineData;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.modifier.Catalysis;
import lcsb.mapviewer.model.map.modifier.Inhibition;
import lcsb.mapviewer.model.map.modifier.Modulation;
import lcsb.mapviewer.model.map.modifier.PhysicalStimulation;
import lcsb.mapviewer.model.map.modifier.Trigger;
import lcsb.mapviewer.model.map.reaction.AbstractNode;
import lcsb.mapviewer.model.map.reaction.AndOperator;
import lcsb.mapviewer.model.map.reaction.AssociationOperator;
import lcsb.mapviewer.model.map.reaction.DissociationOperator;
import lcsb.mapviewer.model.map.reaction.Modifier;
import lcsb.mapviewer.model.map.reaction.NodeOperator;
import lcsb.mapviewer.model.map.reaction.OrOperator;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.reaction.ReactionNode;
import lcsb.mapviewer.model.map.reaction.SplitOperator;
import lcsb.mapviewer.model.map.reaction.TruncationOperator;
import lcsb.mapviewer.model.map.reaction.type.CatalysisReaction;
import lcsb.mapviewer.model.map.reaction.type.DissociationReaction;
import lcsb.mapviewer.model.map.reaction.type.HeterodimerAssociationReaction;
import lcsb.mapviewer.model.map.reaction.type.InhibitionReaction;
import lcsb.mapviewer.model.map.reaction.type.KnownTransitionOmittedReaction;
import lcsb.mapviewer.model.map.reaction.type.ModifierReactionNotation;
import lcsb.mapviewer.model.map.reaction.type.ModulationReaction;
import lcsb.mapviewer.model.map.reaction.type.NegativeInfluenceReaction;
import lcsb.mapviewer.model.map.reaction.type.PhysicalStimulationReaction;
import lcsb.mapviewer.model.map.reaction.type.PositiveInfluenceReaction;
import lcsb.mapviewer.model.map.reaction.type.ReducedModulationReaction;
import lcsb.mapviewer.model.map.reaction.type.ReducedNotation;
import lcsb.mapviewer.model.map.reaction.type.ReducedPhysicalStimulationReaction;
import lcsb.mapviewer.model.map.reaction.type.ReducedTriggerReaction;
import lcsb.mapviewer.model.map.reaction.type.TriggerReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownCatalysisReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownInhibitionReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownNegativeInfluenceReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownPositiveInfluenceReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownReducedModulationReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownReducedPhysicalStimulationReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownReducedTriggerReaction;
import lcsb.mapviewer.model.map.reaction.type.UnknownTransitionReaction;
import lcsb.mapviewer.model.map.species.AntisenseRna;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Degraded;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Ion;
import lcsb.mapviewer.model.map.species.Phenotype;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.SimpleMolecule;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.map.species.TruncatedProtein;
import lcsb.mapviewer.model.map.species.Unknown;
import lcsb.mapviewer.model.map.species.field.AbstractRegionModification;
import lcsb.mapviewer.model.map.species.field.AbstractSiteModification;
import lcsb.mapviewer.model.map.species.field.ModificationResidue;
import lcsb.mapviewer.model.map.species.field.StructuralState;
import lcsb.mapviewer.modelutils.map.ElementUtils;

/**
 * Class used to export SBGN-ML files from the model.
 * 
 * @author Michał Kuźma
 *
 */
public class SbgnmlXmlExporter {

  /**
   * Side margin for units of information.
   */
  private static final double UNIT_OF_INFORMATION_MARGIN = 10.0;
  /**
   * Height of generated units of information.
   */
  private static final double UNIT_OF_INFORMATION_HEIGHT = 12.0;
  /**
   * Height and width of generated operators.
   */
  private static final double OPERATOR_SIZE = 40.0;
  /**
   * Distance between operator circle and port point.
   */
  private static final double OPERATOR_PORT_DISTANCE = 20.0;
  /**
   * Distance between process glyph and port point.
   */
  private static final double PROCESS_PORT_DISTANCE = 10.0;
  /**
   * Length of random alphabetic string added in the beginning of ID, if it is a
   * number. SBGN-ML doesn't accept numbers as ID.
   */
  private static final int ID_RANDOM_STRING_LENGTH = 5;

  /**
   * Default class logger.
   */
  private Logger logger = LogManager.getLogger();
  /**
   * Counter of the arcs parsed so far, used in generating arc's id.
   */
  private int arcCounter;
  /**
   * Set of all the operators parsed so far, used in generating operator's id.
   */
  private Set<NodeOperator> parsedOperators;
  /**
   * Map of operator IDs used when parsing arcs targeting an operator.
   */
  private java.util.Map<NodeOperator, String> operatorIds;
  /**
   * Map of all glyphs and ports generated so far with id as key.
   */
  private java.util.Map<String, Object> sourceTargetMap;

  RenderInformation renderInformation;

  ColorParser colorParser = new ColorParser();

  XmlAnnotationParser xap = new XmlAnnotationParser(new ArrayList<>(), true);

  private java.util.Map<Element, String> exportElementIds = new HashMap<>();

  /**
   * Transforms model into SBGN-ML xml.
   * 
   * @param model
   *          model that should be transformed
   * @return SBGM-ML xml string for the model
   * @throws ConverterException
   */
  public Sbgn toSbgnml(final Model model) throws ConverterException {
    // Reset global variables
    arcCounter = 0;
    parsedOperators = new HashSet<>();
    operatorIds = new HashMap<>();
    sourceTargetMap = new HashMap<>();
    renderInformation = new RenderInformation();

    Map map = new Map();
    map.setId(model.getIdModel());
    map.setLanguage(Language.PD.getName());

    Sbgn sbgnData = new Sbgn();

    for (Element element : model.getElements()) {
      if (element instanceof Species) {
        if (((Species) element).getComplex() == null) {
          Glyph newGlyph = elementToGlyph(element);
          map.getGlyph().add(newGlyph);
        }
      } else if (element instanceof Compartment) {
        Glyph newGlyph = elementToGlyph(element);
        map.getGlyph().add(newGlyph);
      }
    }
    for (Species element : model.getSpeciesList()) {
      Glyph speciesGlyph = ((Glyph) (sourceTargetMap.get(element.getElementId())));
      if (element.getCompartment() != null) {
        Glyph compartmentGlyph = ((Glyph) (sourceTargetMap.get(element.getCompartment().getElementId())));
        speciesGlyph.setCompartmentRef(compartmentGlyph);
      }
    }

    for (Reaction reaction : model.getReactions()) {
      try {
        Glyph glyph = getProcessGlyphFromReaction(reaction);
        if (glyph != null) {
          map.getGlyph().add(glyph);
          map.getArc().addAll(getArcsFromReaction(reaction, map.getGlyph()));
        } else {
          Arc arc = getArcFromReducedReaction(reaction);
          if (arc != null) {
            map.getArc().add(arc);
          }
        }
      } catch (Exception ex) {
        logger.warn(new LogMarker(ProjectLogEntryType.EXPORT_ISSUE, reaction),
            "Unexpected problem with exporting reaction.", ex);
      }
    }
    try {
      RenderUtil.setRenderInformation(map, renderInformation);
    } catch (XMLStreamException | JAXBException | ParserConfigurationException | SAXException | IOException e) {
      throw new ConverterException("Problem with providing render information");
    }

    try {
      StringBuilder notes = new StringBuilder();
      notes.append("<html xmlns=\"http://www.w3.org/1999/xhtml\"><head><title/></head><body>");
      if (model.getNotes() != null) {
        notes.append(new CommonXmlParser().getNotesXmlContent(model.getNotes()));
      }
      notes.append("</body></html>");

      DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
      dbf.setNamespaceAware(true);
      DocumentBuilder db;
      db = dbf.newDocumentBuilder();
      Document doc = db.parse(new ByteArrayInputStream(notes.toString().getBytes()));

      map.setNotes(new SBGNBase.Notes());
      map.getNotes().getAny().add((org.w3c.dom.Element) (doc.getFirstChild()));
    } catch (Exception e) {
      logger.warn("Problem with processing notes");
    }

    sbgnData.getMap().add(map);
    return sbgnData;
  }

  /**
   * Creates new glyph element with all parameters from given alias.
   * 
   * @param element
   *          alias with all parameters for the glyph
   * @return newly created glyph
   */
  private Glyph elementToGlyph(final Element element) {
    Glyph newGlyph = new Glyph();
    String elementId = getExportElementId(element);
    newGlyph.setId(elementId);
    newGlyph.setClazz(getGlyphClazzFromElement(element).getClazz());
    newGlyph.setLabel(getGlyphLabelFromAlias(element));

    Bbox bbox = new Bbox();
    bbox.setX(element.getX().floatValue());
    bbox.setY(element.getY().floatValue());
    bbox.setW(element.getWidth().floatValue());
    bbox.setH(element.getHeight().floatValue());
    newGlyph.setBbox(bbox);

    if (element instanceof Species) {
      Species species = (Species) element;
      if (element instanceof Protein) {
        Protein protein = (Protein) element;
        for (ModificationResidue mr : protein.getModificationResidues()) {
          Glyph stateVariableGlyph = parseStateVariable(mr);
          stateVariableGlyph.setId(newGlyph.getId().concat("-").concat(stateVariableGlyph.getId()));
          newGlyph.getGlyph().add(stateVariableGlyph);
        }
        if (protein.getStructuralState() != null && !protein.getStructuralState().getValue().isEmpty()) {
          newGlyph.getGlyph().add(createStateVariableForStructuralState(protein.getStructuralState()));
        }
      }

      Glyph unitOfInformationGlyph = getUnitOfInformationGlyph(species);
      if (unitOfInformationGlyph != null) {
        newGlyph.getGlyph().add(unitOfInformationGlyph);
      }

      if (element instanceof Complex) {
        Complex complex = (Complex) element;
        for (Species child : complex.getElements()) {
          Glyph childGlyph = elementToGlyph(child);
          newGlyph.getGlyph().add(childGlyph);
        }
        if (complex.getStructuralState() != null && !complex.getStructuralState().getValue().isEmpty()) {
          newGlyph.getGlyph().add(createStateVariableForStructuralState(complex.getStructuralState()));
        }
      }
    }
    addRenderInformation(newGlyph, element);
    addExtensions(newGlyph, element);

    sourceTargetMap.put(element.getElementId(), newGlyph);
    return newGlyph;
  }

  private String getExportElementId(final Element element) {
    if (element == null) {
      return null;
    }
    String elementId = exportElementIds.get(element);
    if (elementId == null) {
      boolean idIsANumber = StringUtils.isNumeric(element.getElementId().substring(0, 1));
      if (idIsANumber) {
        elementId = RandomStringUtils.randomAlphabetic(ID_RANDOM_STRING_LENGTH).concat(element.getElementId());
      } else {
        elementId = element.getElementId();
      }
      exportElementIds.put(element, elementId);
    }
    return elementId;
  }

  private void addExtensions(final Glyph newGlyph, final Element element) {
    SBGNBase.Extension extension = new SBGNBase.Extension();
    newGlyph.setExtension(extension);
    org.w3c.dom.Element annotation = getAnnotationExtension(element);
    if (annotation != null) {
      extension.getAny().add(annotation);
    }

  }

  private org.w3c.dom.Element getAnnotationExtension(final Element element) {
    try {
      String content = xap.dataSetToXmlString(element.getMiriamData(), element.getElementId());
      content = "<annotation>" + content + "</annotation>";
      DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
      dbf.setNamespaceAware(true);
      DocumentBuilder db;
      db = dbf.newDocumentBuilder();
      Document doc = db.parse(new ByteArrayInputStream(content.getBytes()));

      return doc.getDocumentElement();
    } catch (Exception e) {
      logger.error(new LogMarker(ProjectLogEntryType.EXPORT_ISSUE, element), "Problem with exporting annotations", e);
      return null;
    }
  }

  private void addRenderInformation(final Glyph glyph, final Element element) {
    float lineWidth = 1.0f;
    if (element instanceof Species) {
      lineWidth = ((Species) element).getLineWidth().floatValue();
    } else if (element instanceof Compartment) {
      lineWidth = (float) ((Compartment) element).getThickness();
    }
    RenderUtil.addStyle(renderInformation, glyph.getId(), colorParser.colorToHtml(element.getFillColor()),
        colorParser.colorToHtml(element.getBorderColor()), lineWidth);
  }

  private void addRenderInformation(final Glyph glyph, final PolylineData line) {
    float lineWidth = (float) line.getWidth();
    RenderUtil.addStyle(renderInformation, glyph.getId(), colorParser.colorToHtml(Color.WHITE),
        colorParser.colorToHtml(line.getColor()), lineWidth);
  }

  private void addRenderInformation(final Arc arc, final PolylineData line) {
    float lineWidth = (float) line.getWidth();
    RenderUtil.addStyle(renderInformation, arc.getId(), colorParser.colorToHtml(Color.WHITE),
        colorParser.colorToHtml(line.getColor()), lineWidth);
  }

  private Glyph createStateVariableForStructuralState(final StructuralState structuralState) {
    Glyph glyph = new Glyph();
    glyph.setId(structuralState.getSpecies().getElementId() + "-state");
    glyph.setClazz(GlyphClazz.STATE_VARIABLE.getClazz());

    Glyph.State state = new Glyph.State();
    if (structuralState.getValue().indexOf("@") >= 0) {
      state.setValue(structuralState.getValue().substring(0, structuralState.getValue().indexOf("@")));
      state.setVariable(structuralState.getValue().substring(structuralState.getValue().indexOf("@") + 1));
    } else {
      state.setValue(structuralState.getValue());
    }

    glyph.setState(state);

    Bbox bbox = new Bbox();

    float width = structuralState.getWidth().floatValue();
    float height = structuralState.getHeight().floatValue();
    bbox.setH(height);
    bbox.setW(width);

    bbox.setX((float) structuralState.getPosition().getX());
    bbox.setY((float) structuralState.getPosition().getY());

    glyph.setBbox(bbox);

    return glyph;
  }

  /**
   * Creates new glyph element with all parameters from given operator.
   * 
   * @param operator
   *          operator with all parameters for the glyph
   * @return newly created glyph
   */
  private Glyph operatorToGlyph(final NodeOperator operator) {
    Glyph newGlyph = new Glyph();
    newGlyph.setId("operator".concat(Integer.toString(parsedOperators.size())));
    parsedOperators.add(operator);
    operatorIds.put(operator, newGlyph.getId());
    if (operator instanceof AndOperator) {
      newGlyph.setClazz(GlyphClazz.AND.getClazz());
    } else if (operator instanceof OrOperator) {
      newGlyph.setClazz(GlyphClazz.OR.getClazz());
    }

    Bbox bbox = new Bbox();
    bbox.setX((float) (operator.getLine().getStartPoint().getX() - OPERATOR_SIZE / 2));
    bbox.setY((float) (operator.getLine().getStartPoint().getY() - OPERATOR_SIZE / 2));
    bbox.setW((float) OPERATOR_SIZE);
    bbox.setH((float) OPERATOR_SIZE);
    newGlyph.setBbox(bbox);

    Port port2 = new Port();
    port2.setId(newGlyph.getId().concat(".2"));
    Point2D centerPoint = operator.getLine().getStartPoint();
    Point2D nextPoint = operator.getLine().getLines().get(0).getP2();
    double dx = nextPoint.getX() - centerPoint.getX();
    double dy = nextPoint.getY() - centerPoint.getY();
    double portToCenterDistance = OPERATOR_SIZE / 2 + OPERATOR_PORT_DISTANCE;

    double dx2, dy2;
    if (dx != 0) {
      dx2 = Math.sqrt(Math.pow(portToCenterDistance, 2) / (1 + Math.pow(dy / dx, 2)));
      dy2 = dx2 * dy / dx;
    } else {
      dx2 = 0;
      if (dy > 0) {
        dy2 = portToCenterDistance;
      } else {
        dy2 = -portToCenterDistance;
      }
    }

    port2.setX((float) (centerPoint.getX() + dx2));
    port2.setY((float) (centerPoint.getY() + dy2));
    sourceTargetMap.put(port2.getId(), port2);
    newGlyph.getPort().add(port2);

    Port port1 = new Port();
    port1.setId(newGlyph.getId().concat(".1"));
    Point2D portPoint1 = new Point2D.Double(2 * centerPoint.getX() - port2.getX(),
        2 * centerPoint.getY() - port2.getY());
    port1.setX((float) portPoint1.getX());
    port1.setY((float) portPoint1.getY());
    sourceTargetMap.put(port1.getId(), port1);
    newGlyph.getPort().add(port1);

    sourceTargetMap.put(newGlyph.getId(), newGlyph);

    return newGlyph;
  }

  /**
   * Returns GlyphClazz adequate for given element, based on it's class.
   * 
   * @param element
   *          element to extract GlyphCLazz from
   * @return GlyphClazz adequate for given alias
   */
  private GlyphClazz getGlyphClazzFromElement(final Element element) {
    if (element instanceof Protein) {
      Protein protein = (Protein) element;
      if (protein.getHomodimer() == 1) {
        return GlyphClazz.MACROMOLECULE;
      } else {
        return GlyphClazz.MACROMOLECULE_MULTIMER;
      }
    }

    if (element instanceof SimpleMolecule) {
      SimpleMolecule simpleMolecule = (SimpleMolecule) element;
      if (simpleMolecule.getHomodimer() == 1) {
        return GlyphClazz.SIMPLE_CHEMICAL;
      } else {
        return GlyphClazz.SIMPLE_CHEMICAL_MULTIMER;
      }
    }

    if (element instanceof Ion) {
      Ion ion = (Ion) element;
      if (ion.getHomodimer() == 1) {
        return GlyphClazz.SIMPLE_CHEMICAL;
      } else {
        return GlyphClazz.SIMPLE_CHEMICAL_MULTIMER;
      }
    }

    if (element instanceof Gene) {
      Gene gene = (Gene) element;
      if (gene.getHomodimer() == 1) {
        return GlyphClazz.NUCLEIC_ACID_FEATURE;
      } else {
        return GlyphClazz.NUCLEIC_ACID_FEATURE_MULTIMER;
      }
    }

    if (element instanceof Rna) {
      Rna rna = (Rna) element;
      if (rna.getHomodimer() == 1) {
        return GlyphClazz.NUCLEIC_ACID_FEATURE;
      } else {
        return GlyphClazz.NUCLEIC_ACID_FEATURE_MULTIMER;
      }
    }

    if (element instanceof AntisenseRna) {
      AntisenseRna rna = (AntisenseRna) element;
      if (rna.getHomodimer() == 1) {
        return GlyphClazz.NUCLEIC_ACID_FEATURE;
      } else {
        return GlyphClazz.NUCLEIC_ACID_FEATURE_MULTIMER;
      }
    }

    if (element instanceof Complex) {
      Complex complexSpecies = (Complex) element;
      if (complexSpecies.getHomodimer() == 1) {
        return GlyphClazz.COMPLEX;
      } else {
        return GlyphClazz.COMPLEX_MULTIMER;
      }
    }

    if (element instanceof Degraded) {
      return GlyphClazz.SOURCE_AND_SINK;
    }

    if (element instanceof Phenotype) {
      return GlyphClazz.PHENOTYPE;
    }

    if (element instanceof Compartment) {
      return GlyphClazz.COMPARTMENT;
    }
    if (element instanceof Unknown) {
      return GlyphClazz.UNSPECIFIED_ENTITY;
    }

    logger.warn(new LogMarker(ProjectLogEntryType.EXPORT_ISSUE, element),
        "Element type is not supported by SBGN-ML format. Unspecified Entity type assigned.");
    return GlyphClazz.UNSPECIFIED_ENTITY;
  }

  /**
   * Returns label for a glyph with name extracted from given element.
   * 
   * @param element
   *          element with the name
   * @return label for a glyph
   */
  private Label getGlyphLabelFromAlias(final Element element) {
    Label label = new Label();
    label.setText(element.getName());
    Bbox box = new Bbox();
    box.setX(element.getNameX().floatValue());
    box.setY(element.getNameY().floatValue());
    box.setW(element.getNameWidth().floatValue());
    box.setH(element.getNameHeight().floatValue());
    label.setBbox(box);
    return label;
  }

  /**
   * Returns state variable glyph parsed from {@link ModificationResidue}.
   * 
   * @param mr
   *          {@link ModificationResidue} to be parsed
   * @return state variable glyph
   */
  private Glyph parseStateVariable(final ModificationResidue mr) {
    Glyph glyph = new Glyph();
    glyph.setId(mr.getIdModificationResidue());
    glyph.setClazz(GlyphClazz.STATE_VARIABLE.getClazz());

    double width = SpeciesConverter.DEFAULT_MODIFICATION_DIAMETER;
    double height = SpeciesConverter.DEFAULT_MODIFICATION_DIAMETER;
    if (mr instanceof AbstractSiteModification) {
      AbstractSiteModification modification = (AbstractSiteModification) mr;
      Glyph.State state = new Glyph.State();
      state.setVariable(mr.getName());
      glyph.setState(state);
      if (modification.getState() != null) {
        state.setValue(modification.getState().getAbbreviation());
      }
    } else if (mr instanceof AbstractRegionModification) {
      width = ((AbstractRegionModification) mr).getWidth();
      height = ((AbstractRegionModification) mr).getHeight();
    }

    Bbox bbox = new Bbox();

    bbox.setH((float) width);
    bbox.setW((float) height);

    bbox.setX((float) (mr.getPosition().getX() - width / 2));
    bbox.setY((float) (mr.getPosition().getY() - height / 2));

    glyph.setBbox(bbox);

    return glyph;
  }

  /**
   * Returns glyph with unit of information extracted from the alias or null if
   * no unit of information was found.
   * 
   * @param species
   *          input species
   * @return glyph with unit of information extracted from the alias or null if
   *         no unit of information was found
   */
  private Glyph getUnitOfInformationGlyph(final Species species) {
    Glyph uoiGlyph = null;

    String uoiText = "";
    int homodir = species.getHomodimer();
    if (homodir > 1) {
      uoiText = "N:".concat(Integer.toString(homodir));
    }

    if (species instanceof TruncatedProtein) {
      if (!uoiText.equals("")) {
        uoiText = uoiText.concat("; ");
      }
      uoiText = uoiText.concat("ct:truncatedProtein");
    }

    if ((species.getStateLabel() != null) && (species.getStatePrefix() != null)) {
      if (!uoiText.equals("")) {
        uoiText = uoiText.concat("; ");
      }
      if (!species.getStatePrefix().equals("free input")) {
        uoiText = uoiText.concat(species.getStatePrefix()).concat(":");
      }
      uoiText = uoiText.concat(species.getStateLabel());
    }

    if (!uoiText.contains("ct:")) {
      if (species instanceof Rna) {
        if (!uoiText.equals("")) {
          uoiText = uoiText.concat("; ");
        }
        uoiText = uoiText.concat("ct:RNA");
      }
      if (species instanceof AntisenseRna) {
        if (!uoiText.equals("")) {
          uoiText = uoiText.concat("; ");
        }
        uoiText = uoiText.concat("ct:antisenseRNA");
      }
      if (species instanceof Gene) {
        if (!uoiText.equals("")) {
          uoiText = uoiText.concat("; ");
        }
        uoiText = uoiText.concat("ct:gene");
      }
    }

    if (!uoiText.equals("")) {
      uoiGlyph = new Glyph();
      uoiGlyph.setClazz(GlyphClazz.UNIT_OF_INFORMATION.getClazz());
      uoiGlyph.setId(species.getElementId().concat("uoi"));

      Label label = new Label();
      label.setText(uoiText);
      uoiGlyph.setLabel(label);

      Bbox bbox = new Bbox();
      bbox.setX((float) (species.getX() + UNIT_OF_INFORMATION_MARGIN));
      bbox.setY((float) (species.getY() - UNIT_OF_INFORMATION_HEIGHT / 2));
      bbox.setH((float) UNIT_OF_INFORMATION_HEIGHT);
      bbox.setW((float) (species.getWidth() - 2 * UNIT_OF_INFORMATION_MARGIN));
      uoiGlyph.setBbox(bbox);
    }

    return uoiGlyph;
  }

  /**
   * Returns process glyph created based on reaction's center point.
   * 
   * @param reaction
   *          reaction to be parsed
   * @return process glyph for given reaction
   */
  private Glyph getProcessGlyphFromReaction(final Reaction reaction) {
    Glyph processGlyph = new Glyph();
    processGlyph.setId(reaction.getIdReaction());
    GlyphClazz clazz = getGlyphClazzFromReaction(reaction);
    if (clazz == null) {
      return null;
    }
    processGlyph.setClazz(clazz.getClazz());
    Bbox bbox = new Bbox();

    PolylineData line = reaction.getLine();
    Point2D startPoint = line.getStartPoint();
    Point2D endPoint = line.getEndPoint();

    double pointX = (startPoint.getX() + endPoint.getX()) / 2;
    double pointY = (startPoint.getY() + endPoint.getY()) / 2;
    Point2D pointNW = new Point2D.Double(pointX - ReactionConverter.RECT_SIZE / 2,
        pointY - ReactionConverter.RECT_SIZE / 2);
    bbox.setX((float) pointNW.getX());
    bbox.setY((float) pointNW.getY());
    bbox.setH((float) ReactionConverter.RECT_SIZE);
    bbox.setW((float) ReactionConverter.RECT_SIZE);

    processGlyph.setBbox(bbox);

    Port reactantPort = new Port();
    reactantPort.setId(reaction.getIdReaction().concat(".1"));
    sourceTargetMap.put(reactantPort.getId(), reactantPort);

    Port productPort = new Port();
    productPort.setId(reaction.getIdReaction().concat(".2"));
    sourceTargetMap.put(productPort.getId(), productPort);

    // Set glyph orientation and ports' coordinates.
    double reactantDxAverage = 0.0;
    double reactantDyAverage = 0.0;
    double productDxAverage = 0.0;
    double productDyAverage = 0.0;

    for (Reactant r : reaction.getReactants()) {
      Point2D reactantPoint;
      PolylineData reactantArcPoints = r.getLine();
      if (reaction instanceof HeterodimerAssociationReaction) {
        reactantPoint = reactantArcPoints.getEndPoint();
      } else {
        reactantPoint = reactantArcPoints.getLines().get(reactantArcPoints.getLines().size() - 1).getP1();
      }
      reactantDxAverage += reactantPoint.getX() - pointX;
      reactantDyAverage += reactantPoint.getY() - pointY;
    }
    reactantDxAverage /= reaction.getReactants().size();
    reactantDyAverage /= reaction.getReactants().size();

    for (Product p : reaction.getProducts()) {
      PolylineData productArcPoints = p.getLine();
      Point2D productPoint;
      if (reaction instanceof DissociationReaction) {
        productPoint = productArcPoints.getStartPoint();
      } else {
        productPoint = productArcPoints.getLines().get(0).getP2();
      }
      productDxAverage += productPoint.getX() - pointX;
      productDyAverage += productPoint.getY() - pointY;
    }
    productDxAverage /= reaction.getProducts().size();
    productDyAverage /= reaction.getProducts().size();

    boolean horizontalOrientation;
    if ((reactantDxAverage * productDxAverage < 0) && (reactantDyAverage * productDyAverage < 0)) {
      horizontalOrientation = (Math.abs(productDxAverage - reactantDxAverage) > Math
          .abs(productDyAverage - reactantDyAverage));
    } else {
      horizontalOrientation = (reactantDxAverage * productDxAverage < 0);
    }
    if (horizontalOrientation) {
      processGlyph.setOrientation("horizontal");
      reactantPort.setY((float) pointY);
      productPort.setY((float) pointY);
      if (reactantDxAverage < 0) {
        reactantPort.setX((float) (pointX - PROCESS_PORT_DISTANCE));
        productPort.setX((float) (pointX + PROCESS_PORT_DISTANCE));
      } else {
        reactantPort.setX((float) (pointX + PROCESS_PORT_DISTANCE));
        productPort.setX((float) (pointX - PROCESS_PORT_DISTANCE));
      }
    } else {
      reactantPort.setX((float) pointX);
      productPort.setX((float) pointX);
      if (reactantDyAverage < 0) {
        reactantPort.setY((float) (pointY - PROCESS_PORT_DISTANCE));
        productPort.setY((float) (pointY + PROCESS_PORT_DISTANCE));
      } else {
        reactantPort.setY((float) (pointY + PROCESS_PORT_DISTANCE));
        productPort.setY((float) (pointY - PROCESS_PORT_DISTANCE));
      }
      processGlyph.setOrientation("vertical");
    }

    processGlyph.getPort().add(reactantPort);
    processGlyph.getPort().add(productPort);

    sourceTargetMap.put(processGlyph.getId(), processGlyph);

    addRenderInformation(processGlyph, reaction.getLine());

    return processGlyph;
  }

  /**
   * Returns {@link GlyphClazz} appropriate to given reaction.
   * 
   * @param reaction
   *          {@link Reaction} to extract {@link GlyphClazz} from
   * @return {@link GlyphClazz} appropriate to given reaction
   */
  GlyphClazz getGlyphClazzFromReaction(final Reaction reaction) {
    if (reaction instanceof HeterodimerAssociationReaction) {
      return GlyphClazz.ASSOCIATION;
    }
    if (reaction instanceof DissociationReaction) {
      return GlyphClazz.DISSOCIATION;
    }
    if (reaction instanceof KnownTransitionOmittedReaction) {
      return GlyphClazz.OMITTED_PROCESS;
    }
    if (reaction instanceof UnknownTransitionReaction) {
      return GlyphClazz.UNCERTAIN_PROCESS;
    }
    if (reaction instanceof ReducedNotation) {
      return null;
    }
    if (reaction instanceof ModifierReactionNotation) {
      return null;
    }
    return GlyphClazz.PROCESS;
  }

  /**
   * Returns arc extracted from given reduced notation reaction.
   * 
   * @param reaction
   *          reduced notation reaction
   * @return arc extracted from given reduced notation reaction
   */
  Arc getArcFromReducedReaction(final Reaction reaction) {
    if ((reaction.getReactants().size() != 1) || (reaction.getProducts().size() != 1)) {
      logger.warn(new LogMarker(ProjectLogEntryType.PARSING_ISSUE, reaction),
          "Reduced notation does not allow to have multiple reactants/products. Reaction skipped");
      return null;
    }
    Arc arc = new Arc();
    arc.setId(reaction.getIdReaction());

    ArcClazz clazz = getArcClassForReducedNotationReaction(reaction);
    if (clazz != null) {
      arc.setClazz(clazz.getClazz());
    } else {
      logger.warn(new LogMarker(ProjectLogEntryType.EXPORT_ISSUE, reaction), "Invalid arc type."
          + " Reduced notation reaction found of type that is not compliant with SBGN-ML format.");
      return null;
    }

    arc.setSource(sourceTargetMap.get(reaction.getReactants().get(0).getElement().getElementId()));
    arc.setTarget(sourceTargetMap.get(reaction.getProducts().get(0).getElement().getElementId()));

    List<Point2D> pointList = new ArrayList<>();
    pointList.add(reaction.getReactants().get(0).getLine().getStartPoint());
    for (Line2D line : reaction.getReactants().get(0).getLine().getLines()) {
      pointList.add(line.getP2());
    }

    pointList.add(reaction.getProducts().get(0).getLine().getStartPoint());
    for (Line2D line : reaction.getProducts().get(0).getLine().getLines()) {
      pointList.add(line.getP2());
    }

    removeRedundantPoints(pointList);

    Start start = new Start();
    start.setX((float) pointList.get(0).getX());
    start.setY((float) pointList.get(0).getY());
    arc.setStart(start);

    End end = new End();
    end.setX((float) pointList.get(pointList.size() - 1).getX());
    end.setY((float) pointList.get(pointList.size() - 1).getY());
    arc.setEnd(end);

    for (int i = 1; i < pointList.size() - 1; i++) {
      Point2D nextPoint = pointList.get(i);
      Next next = new Next();
      next.setX((float) nextPoint.getX());
      next.setY((float) nextPoint.getY());
      arc.getNext().add(next);
    }

    return arc;
  }

  ArcClazz getArcClassForReducedNotationReaction(final Reaction reaction) {
    if (reaction instanceof NegativeInfluenceReaction) {
      return ArcClazz.INHIBITION;
    } else if (reaction instanceof ReducedModulationReaction) {
      return ArcClazz.MODULATION;
    } else if (reaction instanceof ReducedTriggerReaction) {
      return ArcClazz.NECESSARY_STIMULATION;
    } else if (reaction instanceof ReducedPhysicalStimulationReaction) {
      return ArcClazz.STIMULATION;
    } else if (reaction instanceof CatalysisReaction) {
      return ArcClazz.CATALYSIS;
    } else if (reaction instanceof InhibitionReaction) {
      return ArcClazz.INHIBITION;
    } else if (reaction instanceof ModulationReaction) {
      return ArcClazz.MODULATION;
    } else if (reaction instanceof PhysicalStimulationReaction) {
      return ArcClazz.STIMULATION;
    } else if (reaction instanceof PositiveInfluenceReaction) {
      return ArcClazz.STIMULATION;
    } else if (reaction instanceof TriggerReaction) {
      return ArcClazz.NECESSARY_STIMULATION;
    } else if (reaction instanceof UnknownCatalysisReaction) {
      return ArcClazz.CATALYSIS;
    } else if (reaction instanceof UnknownInhibitionReaction) {
      return ArcClazz.INHIBITION;
    } else if (reaction instanceof UnknownNegativeInfluenceReaction) {
      return ArcClazz.INHIBITION;
    } else if (reaction instanceof UnknownReducedModulationReaction) {
      return ArcClazz.MODULATION;
    } else if (reaction instanceof UnknownReducedPhysicalStimulationReaction) {
      return ArcClazz.STIMULATION;
    } else if (reaction instanceof UnknownPositiveInfluenceReaction) {
      return ArcClazz.STIMULATION;
    } else if (reaction instanceof UnknownReducedTriggerReaction) {
      return ArcClazz.NECESSARY_STIMULATION;
    }
    return null;
  }

  /**
   * Removes redundant points from the list.
   * 
   * @param pointList
   *          list of points to be fixed
   */
  private void removeRedundantPoints(final List<Point2D> pointList) {
    boolean allDone = false;
    while (!allDone) {
      allDone = true;
      for (int i = 1; i < pointList.size() - 1; i++) {
        double dx1 = pointList.get(i).getX() - pointList.get(i - 1).getX();
        double dy1 = pointList.get(i).getY() - pointList.get(i - 1).getY();
        double dx2 = pointList.get(i + 1).getX() - pointList.get(i).getX();
        double dy2 = pointList.get(i + 1).getY() - pointList.get(i).getY();

        DoubleComparator doubleComparator = new DoubleComparator();
        if (((doubleComparator.compare(dy1 / dx1, dy2 / dx2) == 0)
            || (pointList.get(i - 1).getY() == pointList.get(i).getY())
                && (pointList.get(i).getY() == pointList.get(i + 1).getY()))
            && between(pointList.get(i).getX(), pointList.get(i - 1).getX(), pointList.get(i + 1).getX())) {
          pointList.remove(i);
          allDone = false;
        }
      }
    }
  }

  /**
   * Checks if argument x is between s1 and s2.
   * 
   * @param x
   *          number to checked if it is in the middle
   * @param s1
   *          one side argument
   * @param s2
   *          second side argument
   * @return true if x is between s1 and s2
   */
  private boolean between(final double x, final double s1, final double s2) {
    if ((x >= s1) && (x <= s2)) {
      return true;
    }
    if ((x <= s1) && (x >= s2)) {
      return true;
    }
    return false;
  }

  /**
   * Returns set of all arcs used in the reaction.
   * 
   * @param reaction
   *          the reaction to extract arcs from
   * @param glyphList
   *          list of all glyphs in the map; used only for parsing operators
   * @return set of all arcs used in the reaction
   */
  private List<Arc> getArcsFromReaction(final Reaction reaction, final List<Glyph> glyphList) {
    List<Arc> arcList = new ArrayList<>();

    // Parse all nodes except NodeOperators
    for (ReactionNode node : reaction.getReactionNodes()) {
      try {
        arcList.add(getArcFromNode(node, glyphList));
      } catch (InvalidArgumentException ex) {
        logger.warn(new LogMarker(ProjectLogEntryType.EXPORT_ISSUE, node),
            new ElementUtils().getElementTag(node.getElement()) +
                "Node skipped in export process, since it is not compliant with SBGN-ML format.");
        continue;
      }
    }

    // Now parse NodeOperators
    for (AbstractNode node : reaction.getOperators().stream().filter(o -> {
      if (o.getInputs().stream().filter(i -> i instanceof Reactant).count() > 0) {
        return false;
      }
      if (o instanceof SplitOperator || o instanceof AssociationOperator || o instanceof DissociationOperator) {
        return false;
      }
      return true;
    }).collect(Collectors.toList())) {
      try {
        arcList.add(getArcFromNode(node, glyphList));
      } catch (InvalidArgumentException ex) {
        logger.warn("Node skipped in export process, since it is not compliant with SBGN-ML format."
            + node.getClass().getName());
        continue;
      }
    }

    return arcList;

  }

  /**
   * Returns arc for given node.
   * 
   * @param node
   *          node to parse arc from
   * @param glyphList
   *          list of all glyphs in the map; used only for parsing operators
   * @return SBGN-ML arc for given node
   */
  private Arc getArcFromNode(final AbstractNode node, final List<Glyph> glyphList) {
    Arc arc = new Arc();
    arc.setId("arc".concat(Integer.toString(arcCounter)));
    arcCounter += 1;

    boolean logicArc = false;
    if (node instanceof Modifier) {
      for (NodeOperator operator : node.getReaction().getOperators()) {
        if (operator.getInputs().contains(node)) {
          logicArc = true;
        }
      }
    }
    if (logicArc) {
      arc.setClazz(ArcClazz.LOGIC_ARC.getClazz());
    } else {
      arc.setClazz(getArcClazzFromNode(node).getClazz());
    }

    if (node instanceof Reactant) {
      arc.setSource(sourceTargetMap.get(((Reactant) node).getElement().getElementId()));
      arc.setTarget(sourceTargetMap.get(node.getReaction().getIdReaction().concat(".1")));
    } else if (node instanceof Product) {
      arc.setSource(sourceTargetMap.get(node.getReaction().getIdReaction().concat(".2")));
      arc.setTarget(sourceTargetMap.get(((Product) node).getElement().getElementId()));
    } else if (node instanceof Modifier) {
      arc.setSource(sourceTargetMap.get(((Modifier) node).getElement().getElementId()));
      if (!node.getLine().getEndAtd().getArrowType().equals(ArrowType.NONE)) {
        arc.setTarget(sourceTargetMap.get(node.getReaction().getIdReaction()));
      } else {
        boolean found = false;
        for (NodeOperator operator : node.getReaction().getOperators()) {
          if (operator.getInputs().contains(node)) {
            if (!parsedOperators.contains(operator)) {
              Glyph newOperator = operatorToGlyph(operator);
              glyphList.add(newOperator);
            }
            arc.setTarget(sourceTargetMap.get(operatorIds.get(operator).concat(".1")));
            found = true;
          }
        }
        if (!found) {
          throw new InvalidArgumentException(
              new ElementUtils().getElementTag((Modifier) node) + "Problem with exporting modifier");
        }
      }
    } else if (node instanceof NodeOperator) {
      if ((node instanceof DissociationOperator) || (node instanceof AssociationOperator)
          || (node instanceof TruncationOperator)) {
        throw new InvalidArgumentException();
      }
      arc.setSource(sourceTargetMap.get(operatorIds.get(node).concat(".2")));
      if (!node.getLine().getEndAtd().getArrowType().equals(ArrowType.NONE)) {
        arc.setTarget(sourceTargetMap.get(node.getReaction().getIdReaction()));
      } else {
        for (NodeOperator operator : node.getReaction().getOperators()) {
          if (operator.getInputs().contains(node)) {
            if (!parsedOperators.contains(operator)) {
              Glyph newOperator = operatorToGlyph(operator);
              glyphList.add(newOperator);
            }
            arc.setTarget(sourceTargetMap.get(operatorIds.get(operator).concat(".1")));
          }
        }
      }
    }

    List<Point2D> arcPoints = new ArrayList<>();
    arcPoints.add(node.getLine().getStartPoint());
    for (Line2D line : node.getLine().getLines()) {
      arcPoints.add(line.getP2());
    }
    Start start = new Start();
    if ((node instanceof Product) || (node instanceof NodeOperator)) {
      Port sourcePort = (Port) arc.getSource();
      start.setX(sourcePort.getX());
      start.setY(sourcePort.getY());
    } else {
      start.setX((float) arcPoints.get(0).getX());
      start.setY((float) arcPoints.get(0).getY());
    }
    arc.setStart(start);

    End end = new End();
    if ((node instanceof Reactant) || ((node instanceof Modifier) && (arc.getTarget() instanceof Port))) {
      Port targetPort = (Port) arc.getTarget();
      end.setX(targetPort.getX());
      end.setY(targetPort.getY());
    } else {
      Point2D lastPoint = arcPoints.get(arcPoints.size() - 1);
      end.setX((float) lastPoint.getX());
      end.setY((float) lastPoint.getY());
    }
    arc.setEnd(end);

    if ((node instanceof Product) && (node.getReaction() instanceof DissociationReaction)) {
      Point2D nextPoint = arcPoints.get(0);
      Next next = new Next();
      next.setX((float) nextPoint.getX());
      next.setY((float) nextPoint.getY());
      arc.getNext().add(next);
    }

    DoubleComparator doubleComparator = new DoubleComparator();
    for (int i = 1; i < arcPoints.size() - 1; i++) {
      Point2D nextPoint = arcPoints.get(i);
      if ((doubleComparator.compare(nextPoint.getX(), new Double(start.getX())) == 0)
          && (doubleComparator.compare(nextPoint.getY(), new Double(start.getY())) == 0)) {
        arc.getNext().clear();
        continue;
      }
      if ((doubleComparator.compare(nextPoint.getX(), new Double(end.getX())) == 0)
          && (doubleComparator.compare(nextPoint.getY(), new Double(end.getY())) == 0)) {
        break;
      }
      Next next = new Next();
      next.setX((float) nextPoint.getX());
      next.setY((float) nextPoint.getY());
      arc.getNext().add(next);
    }

    if ((node instanceof Reactant) && (node.getReaction() instanceof HeterodimerAssociationReaction)) {
      Point2D nextPoint = arcPoints.get(arcPoints.size() - 1);
      Next next = new Next();
      next.setX((float) nextPoint.getX());
      next.setY((float) nextPoint.getY());
      arc.getNext().add(next);
    }

    addRenderInformation(arc, node.getLine());
    return arc;
  }

  /**
   * Returns {@link ArcClazz} for given node.
   * 
   * @param node
   *          Node to extract {@link ArcClazz} from
   * @return {@link ArcClazz} for given node
   */
  private ArcClazz getArcClazzFromNode(final AbstractNode node) {
    if (node instanceof Reactant) {
      return ArcClazz.CONSUMPTION;
    }
    if (node instanceof Product) {
      return ArcClazz.PRODUCTION;
    }
    if (node instanceof Catalysis) {
      return ArcClazz.CATALYSIS;
    }
    if (node instanceof Inhibition) {
      return ArcClazz.INHIBITION;
    }
    if (node instanceof Modulation) {
      return ArcClazz.MODULATION;
    }
    if (node instanceof Trigger) {
      return ArcClazz.NECESSARY_STIMULATION;
    }
    if (node instanceof PhysicalStimulation) {
      return ArcClazz.STIMULATION;
    }
    if (node instanceof NodeOperator) {
      ArrowType arrowType = node.getLine().getEndAtd().getArrowType();
      switch (arrowType) {
        case BLANK:
          return ArcClazz.STIMULATION;
        case BLANK_CROSSBAR:
          return ArcClazz.NECESSARY_STIMULATION;
        case CIRCLE:
          return ArcClazz.CATALYSIS;
        case CROSSBAR:
          return ArcClazz.INHIBITION;
        case DIAMOND:
          return ArcClazz.MODULATION;
        case NONE:
          return ArcClazz.LOGIC_ARC;
        default:
          throw new InvalidArgumentException();
      }
    }
    throw new InvalidArgumentException();
  }
}
