package lcsb.mapviewer.converter.model.sbml.species;

import java.awt.Color;
import java.util.Objects;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.sbml.jsbml.Model;
import org.sbml.jsbml.ext.layout.*;
import org.sbml.jsbml.ext.render.*;

import lcsb.mapviewer.converter.model.sbml.SbmlModelUtils;
import lcsb.mapviewer.model.LogMarker;
import lcsb.mapviewer.model.ProjectLogEntryType;
import lcsb.mapviewer.model.graphics.ArrowType;
import lcsb.mapviewer.model.map.species.field.*;
import lcsb.mapviewer.modelutils.map.ElementUtils;

public class ModificationResidueExporter {

  /**
   * Default diameter of the modification residues.
   */
  public static final int DEFAULT_MODIFICATION_DIAMETER = 15;

  private static Logger logger = LogManager.getLogger();

  private SbmlModelUtils sbmlModelUtils;

  public ModificationResidueExporter(Model sbmlModel) {
    this.sbmlModelUtils = new SbmlModelUtils(sbmlModel);
  }

  public LocalStyle createStyle(ModificationResidue element) {
    if (element instanceof Residue) {
      return createEllipseStyle();
    } else if (element instanceof ModificationSite) {
      return createModificationSiteStyle((ModificationSite) element);
    } else if (element instanceof TranscriptionSite) {
      return createTranscriptionSiteStyle((TranscriptionSite) element);
    } else if (element instanceof RegulatoryRegion) {
      return createRectangleStyle();
    } else if (element instanceof BindingRegion) {
      return createRectangleStyle();
    } else if (element instanceof CodingRegion) {
      return createRectangleStyle();
    } else if (element instanceof ProteinBindingDomain) {
      return createRectangleStyle();
    } else {
      logger.warn(new LogMarker(ProjectLogEntryType.EXPORT_ISSUE, element.getSpecies()),
          "Don't know how to render modification residue: " + new ElementUtils().getElementTag(element));
      return createEllipseStyle();
    }
  }

  private LocalStyle createRectangleStyle() {
    LocalStyle style = createGenericStyle();

    Rectangle rectangle = new Rectangle();
    rectangle.setRx(new RelAbsVector(0, 0));
    rectangle.setRy(new RelAbsVector(0, 0));
    rectangle.setX(new RelAbsVector(0, 0));
    rectangle.setY(new RelAbsVector(0, 0));
    rectangle.setWidth(new RelAbsVector(0, 100));
    rectangle.setHeight(new RelAbsVector(0, 100));

    style.getGroup().addElement(rectangle);

    return style;
  }

  private LocalStyle createTranscriptionSiteStyle(TranscriptionSite transcriptionSite) {
    LocalStyle style = createGenericStyle();

    ColorDefinition color;
    if (transcriptionSite.getActive()) {
      color = sbmlModelUtils.getColorDefinition(Color.RED);
    } else {
      color = sbmlModelUtils.getColorDefinition(Color.BLACK);
    }
    if (transcriptionSite.getSpecies() != null && transcriptionSite.getSpecies().getFillColor().equals(Color.WHITE)) {
      color = sbmlModelUtils.getColorDefinition(Color.BLACK);
    }

    style.getGroup().setStroke(color.getId());
    style.getGroup().setEndHead(ArrowType.FULL.name());
    style.getGroup().setFill(color.getId());

    double y = transcriptionSite.getPosition().getY();
    double x = transcriptionSite.getPosition().getX();
    if (Objects.equals(transcriptionSite.getDirection(), "RIGHT")) {
      x += transcriptionSite.getWidth();
    }
    Polygon line = new Polygon();

    RenderPoint p1 = line.createRenderPoint();
    p1.setX(new RelAbsVector(Math.round(x - transcriptionSite.getPosition().getX()), 50));
    p1.setY(new RelAbsVector(Math.round(y - transcriptionSite.getPosition().getY()), 50));

    y -= DEFAULT_MODIFICATION_DIAMETER;
    RenderPoint p2 = line.createRenderPoint();
    p2.setX(new RelAbsVector(Math.round(x - transcriptionSite.getPosition().getX()), 50));
    p2.setY(new RelAbsVector(Math.round(y - transcriptionSite.getPosition().getY()), 50));

    double delta = 0;
    if (Objects.equals(transcriptionSite.getDirection(), "RIGHT")) {
      x -= transcriptionSite.getWidth();
      delta = 6;
    } else {
      x += transcriptionSite.getWidth();
      delta = -6;
    }
    RenderPoint p3 = line.createRenderPoint();
    p3.setX(new RelAbsVector(Math.round(x - transcriptionSite.getPosition().getX()), 50));
    p3.setY(new RelAbsVector(Math.round(y - transcriptionSite.getPosition().getY()), 50));

    RenderPoint p4 = line.createRenderPoint();
    p4.setX(new RelAbsVector(Math.round(x + delta - transcriptionSite.getPosition().getX()), 50));
    p4.setY(new RelAbsVector(Math.round(y - 3 - transcriptionSite.getPosition().getY()), 50));

    RenderPoint p5 = line.createRenderPoint();
    p5.setX(new RelAbsVector(Math.round(x + delta - transcriptionSite.getPosition().getX()), 50));
    p5.setY(new RelAbsVector(Math.round(y + 3 - transcriptionSite.getPosition().getY()), 50));

    RenderPoint p6 = line.createRenderPoint();
    p6.setX(new RelAbsVector(Math.round(x - transcriptionSite.getPosition().getX()), 50));
    p6.setY(new RelAbsVector(Math.round(y - transcriptionSite.getPosition().getY()), 50));

    RenderPoint p7 = line.createRenderPoint();
    p7.setX(new RelAbsVector(p2.getX()));
    p7.setY(new RelAbsVector(p2.getY()));

    style.getGroup().addElement(line);

    return style;
  }

  private LocalStyle createModificationSiteStyle(ModificationSite modificationSite) {
    LocalStyle style = createGenericStyle();

    double diameter = DEFAULT_MODIFICATION_DIAMETER;

    Ellipse ellipse = new Ellipse();
    ellipse.setCx(new RelAbsVector(0, 50));
    ellipse.setCy(new RelAbsVector(Math.round(-diameter), 50));
    ellipse.setRx(new RelAbsVector(Math.round(diameter / 2)));
    ellipse.setRy(new RelAbsVector(Math.round(diameter / 2)));

    style.getGroup().addElement(ellipse);

    Polygon line = new Polygon();

    RenderPoint p1 = line.createRenderPoint();
    p1.setX(new RelAbsVector(0, 50));
    p1.setY(new RelAbsVector(Math.round(-diameter / 2), 50));

    RenderPoint p2 = line.createRenderPoint();
    p2.setX(new RelAbsVector(0, 50));
    p2.setY(new RelAbsVector(0, 50));

    style.getGroup().addElement(line);

    return style;
  }

  public LocalStyle createStyle(StructuralState structuralState) {
    return createEllipseStyle();
  }

  private LocalStyle createEllipseStyle() {
    LocalStyle style = createGenericStyle();

    Ellipse ellipse = new Ellipse();
    ellipse.setCx(new RelAbsVector(0, 50));
    ellipse.setCy(new RelAbsVector(0, 50));
    ellipse.setRx(new RelAbsVector(0, 50));
    ellipse.setRy(new RelAbsVector(0, 50));

    style.getGroup().addElement(ellipse);

    return style;
  }

  public LocalStyle createTextStyle(ModificationResidue mr) {
    return createTextStyle();
  }

  private LocalStyle createTextStyle() {
    LocalStyle style = createGenericStyle();

    style.getGroup().setVTextAnchor(VTextAnchor.MIDDLE);
    style.getGroup().setTextAnchor(HTextAnchor.MIDDLE);
    return style;
  }

  private LocalStyle createGenericStyle() {
    LocalStyle style = sbmlModelUtils.createStyle();
    ColorDefinition color = sbmlModelUtils.getColorDefinition(Color.BLACK);
    ColorDefinition background = sbmlModelUtils.getColorDefinition(Color.WHITE);
    style.getGroup().setStroke(color.getId());
    style.getGroup().setStrokeWidth(1.0);
    style.getGroup().setFill(background.getId());
    style.getGroup().setFontSize((short) 10);
    return style;
  }

  public LocalStyle createTextStyle(StructuralState structuralState) {
    return createTextStyle();
  }

  public BoundingBox createBoundingBoxForModification(ModificationResidue mr) {
    double width = getModificationWidth(mr);
    double height = getModificationHeight(mr);
    double x = mr.getPosition().getX() - width / 2;
    double y = mr.getPosition().getY() - height / 2;
    BoundingBox boundingBox = new BoundingBox();
    boundingBox.setPosition(new Point(x, y, mr.getSpecies().getZ() + 1));
    Dimensions dimensions = new Dimensions();
    dimensions.setWidth(width);
    dimensions.setHeight(height);
    boundingBox.setDimensions(dimensions);
    return boundingBox;
  }

  private double getModificationHeight(ModificationResidue mr) {
    if (mr instanceof AbstractRegionModification) {
      return ((AbstractRegionModification) mr).getHeight();
    } else {
      return 12;
    }
  }

  private double getModificationWidth(ModificationResidue mr) {
    if (mr instanceof AbstractRegionModification) {
      return ((AbstractRegionModification) mr).getWidth();
    } else {
      return 12;
    }
  }

  public BoundingBox createBoundingBoxForModificationText(ModificationResidue mr) {
    BoundingBox result = createBoundingBoxForModification(mr);
    if (mr instanceof ModificationSite) {
      result.getPosition().setY(result.getPosition().getY() - DEFAULT_MODIFICATION_DIAMETER);
    }
    result.getPosition().setX(result.getPosition().getX() + 2);
    result.getPosition().setY(result.getPosition().getY() + 2);
    result.getDimensions().setWidth(result.getDimensions().getWidth() - 4);
    result.getDimensions().setHeight(result.getDimensions().getHeight() - 4);
    result.getPosition().setZ(result.getPosition().getZ() + 1);
    return result;
  }

  public BoundingBox createBoundingBoxForStructuralState(StructuralState structuralState) {
    double width = structuralState.getWidth();
    double height = structuralState.getHeight();
    double x = structuralState.getPosition().getX();
    double y = structuralState.getPosition().getY();
    BoundingBox boundingBox = new BoundingBox();
    boundingBox.setPosition(new Point(x, y, structuralState.getSpecies().getZ() + 1));
    Dimensions dimensions = new Dimensions();
    dimensions.setWidth(width);
    dimensions.setHeight(height);
    boundingBox.setDimensions(dimensions);
    return boundingBox;
  }

  public BoundingBox createBoundingBoxForStructuralStateText(StructuralState structuralState) {
    BoundingBox result = createBoundingBoxForStructuralState(structuralState);
    result.getPosition().setZ(result.getPosition().getZ() + 1);
    return result;
  }

}
