package lcsb.mapviewer.model.graphics;

import java.awt.Color;
import java.awt.geom.*;
import java.io.Serializable;
import java.util.*;

import javax.persistence.*;
import javax.persistence.Entity;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.annotations.*;
import org.hibernate.annotations.CascadeType;

import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.common.exception.NotImplementedException;
import lcsb.mapviewer.common.geometry.*;
import lcsb.mapviewer.model.map.Drawable;

/**
 * This class represents general line structure. Line can contain several
 * segments. Beginning and the end of the line could be finished with the arrow.
 * There are also color and line width associated with the object.
 * 
 * @author Piotr Gawron
 * 
 */
@Entity
public class PolylineData implements Serializable, Drawable {
  /**
   * 
   */
  private static final long serialVersionUID = 1L;

  /**
   * Class used to perform some operation on points.
   */
  private static PointTransformation pt = new PointTransformation();

  /**
   * Default class logger.
   */
  private static Logger logger = LogManager.getLogger();

  /**
   * Unique identifier in the database.
   */
  @Id
  @GeneratedValue(strategy = GenerationType.IDENTITY)
  private int id;

  /**
   * List of points that build the line.
   */
  @ElementCollection(fetch = FetchType.EAGER)
  @Fetch(FetchMode.SUBSELECT)
  @CollectionTable(name = "point_table", joinColumns = @JoinColumn(name = "id"))
  @OrderColumn(name = "idx")
  @Columns(columns = { @Column(name = "x1"), @Column(name = "y1"), @Column(name = "x2"), @Column(name = "y2") })
  @Type(type = "lcsb.mapviewer.persist.mapper.Line2DMapper")
  @Cascade({ CascadeType.ALL })
  private List<Line2D> segments = new ArrayList<>();

  /**
   * Arrow at the beginning of the line.
   */
  @Type(type = "lcsb.mapviewer.persist.mapper.ArrowTypeDataMapper")
  private ArrowTypeData beginAtd = new ArrowTypeData();

  /**
   * Arrow at the end of the line.
   */
  @Type(type = "lcsb.mapviewer.persist.mapper.ArrowTypeDataMapper")
  private ArrowTypeData endAtd = new ArrowTypeData();

  /**
   * Width of the line.
   */
  private double width = 1;

  /**
   * Color of the line.
   */
  @Type(type = "lcsb.mapviewer.persist.mapper.ColorMapper")
  @Column(nullable = false)
  private Color color = Color.BLACK;

  /**
   * Type of the line (pattern used to draw it).
   */
  @Enumerated(EnumType.STRING)
  private LineType type = LineType.SOLID;

  /**
   * Default constructor for simple to points line.
   * 
   * @param startPoint
   *          start point of the line
   * @param endPoint
   *          end point of the line
   */
  public PolylineData(Point2D startPoint, Point2D endPoint) {
    addLine(new Line2D.Double(startPoint, endPoint));
  }

  /**
   * Default constructor.
   */
  public PolylineData() {
  }

  /**
   * Constructor that creates object initialized by the parameter object.
   * 
   * @param param
   *          parameter used for data initialization
   */
  public PolylineData(PolylineData param) {
    segments = new ArrayList<>();
    for (Line2D line : param.getLines()) {
      addLine(new Line2D.Double(line.getP1(), line.getP2()));
    }
    beginAtd = new ArrowTypeData(param.getBeginAtd());
    endAtd = new ArrowTypeData(param.getEndAtd());
    width = param.getWidth();
    color = param.getColor();
    type = param.getType();
  }

  /**
   * Creates line from the list of points.
   * 
   * @param pointsList
   *          list of points that represent the line
   */
  public PolylineData(List<Point2D> pointsList) {
    if (pointsList.size() == 1) {
      throw new InvalidArgumentException("At least two points are required");
    }
    for (int i = 1; i < pointsList.size(); i++) {
      addLine(pointsList.get(i - 1), pointsList.get(i));
    }
  }

  public void addLine(Point2D p1, Point2D p2) {
    addLine(new Line2D.Double(p1, p2));
  }

  /**
   * Returns all segment lines as a list of {@link Line2D} objects.
   * 
   * @return list of segment lines
   */
  public List<Line2D> getLines() {
    List<Line2D> data = new ArrayList<>();
    for (Line2D line2d : segments) {
      data.add(new ImmutableLine2D(line2d));
    }
    return Collections.unmodifiableList(data);
  }

  /**
   * Returns coordinates of the last point in the line.
   *
   * @return coordinates of the last point in the line
   */
  public Point2D getEndPoint() {
    return new ImmutablePoint2D(segments.get(segments.size() - 1).getP2());
  }

  /**
   * Transforms line representation into {@link GeneralPath} class.
   * 
   * @return {@link GeneralPath} representation of line
   */
  public GeneralPath toGeneralPath() {
    GeneralPath path = new GeneralPath();
    Point2D p = segments.get(0).getP1();
    path.moveTo(p.getX(), p.getY());
    for (Line2D line : segments) {
      p = line.getP2();
      path.lineTo(p.getX(), p.getY());
    }
    return path;
  }

  /**
   * Trims the end of the line.
   * 
   * @param distToTrim
   *          distance by which end of line should be trimmed
   */
  public double trimEnd(double distToTrim) {
    Line2D line = segments.get(segments.size() - 1);
    Point2D last = line.getP2();
    Point2D last2 = line.getP1();
    double oldDist = last.distance(last2);
    if (distToTrim < 0 && oldDist <= Configuration.EPSILON) {
      logger.debug("Cannot extend empty line");
      line.setLine(last2, last2);
      return 0;
    } else if (oldDist <= distToTrim) {
      line.setLine(last2, last2);
      return oldDist;
    } else {
      double newDistr = oldDist - distToTrim;
      double ratio = newDistr / oldDist;
      double dx = last.getX() - last2.getX();
      double dy = last.getY() - last2.getY();
      dx *= ratio;
      dy *= ratio;
      line.setLine(last2.getX(), last2.getY(), last2.getX() + dx, last2.getY() + dy);
      return distToTrim;
    }
  }

  /**
   * Trims the beginning of the line.
   * 
   * @param distToTrim
   *          distance by which beginning of line should be trimmed
   */
  public double trimBegin(double distToTrim) {
    Line2D line = segments.get(0);
    Point2D last = line.getP1();
    Point2D last2 = line.getP2();
    double oldDist = last.distance(last2);
    if (oldDist <= distToTrim) {
      line.setLine(last2, last2);
      return oldDist;
    } else {
      double newDistr = oldDist - distToTrim;
      double ratio = newDistr / oldDist;
      double dx = last.getX() - last2.getX();
      double dy = last.getY() - last2.getY();
      dx *= ratio;
      dy *= ratio;
      line.setLine(last2.getX() + dx, last2.getY() + dy, last2.getX(), last2.getY());
      return distToTrim;
    }
  }

  /**
   * Returns the object with reversed order of line.
   * 
   * @return reversed line
   */
  public PolylineData reverse() {
    PolylineData result = new PolylineData(this);
    result.segments = new ArrayList<>();
    for (Line2D line : getLines()) {
      result.addLine(new Line2D.Double(line.getP2(), line.getP1()));
    }
    Collections.reverse(result.segments);
    result.setBeginAtd(getEndAtd());
    result.setEndAtd(getBeginAtd());
    return result;
  }

  public void addLine(Line2D line) {
    if (!pt.isValidPoint(line.getP1())) {
      throw new InvalidArgumentException("Invalid coordinates: " + line.getP1());
    }
    if (!pt.isValidPoint(line.getP2())) {
      throw new InvalidArgumentException("Invalid coordinates: " + line.getP2());
    }
    segments.add(new Line2D.Double(line.getX1(), line.getY1(), line.getX2(), line.getY2()));
  }

  /**
   * Returns length of the whole line.
   * 
   * @return length of the line
   */
  public double length() {
    double dist = 0;
    for (Line2D line : segments) {
      dist += line.getP1().distance(line.getP2());
    }
    return dist;
  }

  /**
   * Sets first point in the line.
   * 
   * @param point
   *          new first point in the line
   */
  public void setStartPoint(Point2D point) {
    if (!pt.isValidPoint(point)) {
      throw new InvalidArgumentException("Invalid coordinates: " + point);
    }
    Line2D line = segments.get(0);
    line.setLine(point, line.getP2());
  }

  public void setEndPoint(Point2D point) {
    if (!pt.isValidPoint(point)) {
      throw new InvalidArgumentException("Invalid coordinates: " + point);
    }
    Line2D line = segments.get(segments.size() - 1);
    line.setLine(line.getP1(), point);
  }

  /**
   * Returns first point in the line.
   * 
   * @return first point in the line
   */
  public Point2D getStartPoint() {
    return new ImmutablePoint2D(segments.get(0).getP1());
  }

  /**
   * @return the beginAtd
   * @see #beginAtd
   */
  public ArrowTypeData getBeginAtd() {
    return beginAtd;
  }

  /**
   * @param beginAtd
   *          the beginAtd to set
   * @see #beginAtd
   */
  public void setBeginAtd(ArrowTypeData beginAtd) {
    this.beginAtd = beginAtd;
  }

  /**
   * @return the endAtd
   * @see #endAtd
   */
  public ArrowTypeData getEndAtd() {
    return endAtd;
  }

  /**
   * @param endAtd
   *          the endAtd to set
   * @see #endAtd
   */
  public void setEndAtd(ArrowTypeData endAtd) {
    this.endAtd = endAtd;
  }

  /**
   * @return the width
   * @see #width
   */
  public double getWidth() {
    return width;
  }

  /**
   * @param width
   *          the width to set
   * @see #width
   */
  public void setWidth(double width) {
    this.width = width;
  }

  /**
   * @param string
   *          the width to set
   * @see #width
   */
  public void setWidth(String string) {
    setWidth(Double.parseDouble(string));
  }

  /**
   * @return the color
   * @see #color
   */
  public Color getColor() {
    return color;
  }

  /**
   * @param color
   *          the color to set
   * @see #color
   */
  public void setColor(Color color) {
    this.color = color;
  }

  /**
   * @return the type
   * @see #type
   */
  public LineType getType() {
    return type;
  }

  /**
   * @param type
   *          the type to set
   * @see #type
   */
  public void setType(LineType type) {
    this.type = type;
  }

  /**
   * @return the id
   * @see #id
   */
  public int getId() {
    return id;
  }

  /**
   * @param id
   *          the id to set
   * @see #id
   */
  public void setId(int id) {
    this.id = id;
  }

  /**
   * Prepares a copy of the object.
   * 
   * @return copy of {@link PolylineData}
   */
  public PolylineData copy() {
    if (this.getClass().equals(PolylineData.class)) {
      return new PolylineData(this);
    } else {
      throw new NotImplementedException("Copy method for " + this.getClass() + " class not implemented.");
    }
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("[" + this.getClass().getSimpleName() + "]: ");
    for (Line2D point : segments) {
      sb.append(point.getX1() + ", " + point.getY1() + "-" + point.getX2() + ", " + point.getY2() + "; ");
    }
    return sb.toString();
  }

  @Override
  public Integer getZ() {
    return 0;
  }

  @Override
  public void setZ(Integer z) {
    logger.warn("Not implemented");
  }

  @Override
  public String getElementId() {
    return toString();
  }

  @Override
  public double getSize() {
    return 0;
  }

  public void addLines(List<Line2D> lines) {
    for (Line2D line : lines) {
      addLine(line);
    }
  }

  public void setStartPoint(double x, double y) {
    setStartPoint(new Point2D.Double(x, y));
  }

  public void addLine(double x1, double y1, double x2, double y2) {
    addLine(new Line2D.Double(x1, y1, x2, y2));
  }

  public void setEndPoint(double x, double y) {
    setEndPoint(new Point2D.Double(x, y));
  }

  public void removeLine(int index) {
    segments.remove(index);
  }

  public void setLine(int index, Point2D p1, Point2D p2) {
    segments.get(index).setLine(p1, p2);
  }

  public void setLine(int index, double x1, double y1, double x2, double y2) {
    segments.get(index).setLine(x1, y1, x2, y2);
  }

  public void removeLines() {
    segments.clear();
  }
}
