package lcsb.mapviewer.model.map.layout;

import java.io.Serializable;
import java.util.*;

import javax.persistence.*;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.annotations.Cascade;
import org.hibernate.annotations.CascadeType;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;

import lcsb.mapviewer.common.exception.NotImplementedException;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.modelutils.serializer.ProjectAsIdSerializer;
import lcsb.mapviewer.modelutils.serializer.UserAsIdSerializer;

/**
 * This object represents type of visualization for the model.
 * 
 * @author Piotr Gawron
 * 
 */
@Entity
public class ProjectBackground implements Serializable {
  public static final Comparator<ProjectBackground> ID_COMPARATOR = new Comparator<ProjectBackground>() {

    @Override
    public int compare(ProjectBackground o1, ProjectBackground o2) {
      return o1.getId() - o2.getId();
    }
  };
  public static final Comparator<? super ProjectBackground> ORDER_COMPARATOR = new Comparator<ProjectBackground>() {

    @Override
    public int compare(ProjectBackground o1, ProjectBackground o2) {
      return o1.getOrderIndex() - o2.getOrderIndex();
    }
  };
  /**
   *
   */
  private static final long serialVersionUID = 1L;
  /**
   * Default class logger.
   */
  @SuppressWarnings("unused")
  private static Logger logger = LogManager.getLogger();

  /**
   * Unique database identifier.
   */
  @Id
  @GeneratedValue(strategy = GenerationType.IDENTITY)
  private int id;

  /**
   * Title of this layout.
   */
  @Column(length = 255)
  private String name;

  /**
   * Is the overlay considered as default (should it be open on startup).
   */
  private boolean defaultOverlay = false;

  /**
   * Does the layout present data in hierarchical view.
   */
  @JsonIgnore
  private boolean hierarchicalView = false;

  /**
   * If overlay contain hierarchical view then it might be fixed on some specific
   * level. This parameter defines the level at which it's fixed or contains null
   * if it's general hierarchical view.
   */
  @JsonIgnore
  private Integer hierarchyViewLevel;

  /**
   * Project to which this data overlay is assigned.
   */
  @ManyToOne(fetch = FetchType.LAZY, optional = false)
  @JsonSerialize(using = ProjectAsIdSerializer.class)
  private Project project;

  @ManyToOne(fetch = FetchType.LAZY, optional = false)
  @JsonSerialize(using = UserAsIdSerializer.class)
  private User creator;

  /**
   * Describes progress of layout uploading process.
   */
  @Enumerated(EnumType.STRING)
  private ProjectBackgroundStatus status = ProjectBackgroundStatus.UNKNOWN;

  /**
   * What is the progress.
   */
  private double progress = 0.0;

  private int orderIndex = 0;

  /**
   * List of image folders for generated background images.
   */
  @Cascade({ CascadeType.SAVE_UPDATE })
  @OneToMany(fetch = FetchType.LAZY, mappedBy = "projectBackground", orphanRemoval = true)
  private Set<ProjectBackgroundImageLayer> backgrounds = new HashSet<>();

  /**
   * Short description used for this layout.
   */
  @Column(columnDefinition = "TEXT")
  private String description;

  /**
   * Default constructor.
   */
  public ProjectBackground() {

  }

  /**
   * Constructor that initializes object with basic parameters.
   *
   * @param title
   *          title of the layout
   */
  public ProjectBackground(String title) {
    this.name = title;
  }

  /**
   * Constructor that initializes object with the copy of the data from parameter.
   *
   * @param layout
   *          from this object the data will be initialized
   */
  public ProjectBackground(ProjectBackground layout) {
    this.description = layout.getDescription();
    this.name = layout.getName();
    this.status = layout.status;
    this.progress = layout.progress;
    this.hierarchicalView = layout.hierarchicalView;
    this.hierarchyViewLevel = layout.hierarchyViewLevel;
    this.orderIndex = layout.orderIndex;
    for (ProjectBackgroundImageLayer dataOverlayImageLayer : layout.getProjectBackgroundImageLayer()) {
      ProjectBackgroundImageLayer dataOverlayImageLayerCopy = new ProjectBackgroundImageLayer(
          dataOverlayImageLayer.getModel(),
          dataOverlayImageLayer.getDirectory());
      this.addProjectBackgroundImageLayer(dataOverlayImageLayerCopy);
    }
  }

  public void addProjectBackgroundImageLayer(ProjectBackgroundImageLayer projectBackgroundImageLayer) {
    this.backgrounds.add(projectBackgroundImageLayer);
    projectBackgroundImageLayer.setProjectBackground(this);
  }

  /**
   * Prepares a copy of layout.
   *
   * @return copy of layout
   */
  public ProjectBackground copy() {
    if (this.getClass() == ProjectBackground.class) {
      return new ProjectBackground(this);
    } else {
      throw new NotImplementedException("Method copy() should be overriden in class " + this.getClass());
    }
  }

  /**
   * @return the hierarchicalView
   * @see #hierarchicalView
   */
  public boolean isHierarchicalView() {
    return hierarchicalView;
  }

  /**
   * @param hierarchicalView
   *          the hierarchicalView to set
   * @see #hierarchicalView
   */
  public void setHierarchicalView(boolean hierarchicalView) {
    this.hierarchicalView = hierarchicalView;
  }

  /**
   * @return the id
   * @see #id
   */
  public int getId() {
    return id;
  }

  /**
   * @param id
   *          the id to set
   * @see #id
   */
  public void setId(int id) {
    this.id = id;
  }

  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
  }

  /**
   * @return the creator
   * @see #creator
   */
  public User getCreator() {
    return creator;
  }

  /**
   * @param creator
   *          the creator to set
   * @see #creator
   */
  public void setCreator(User creator) {
    this.creator = creator;
  }

  /**
   * @return the status
   * @see #status
   */
  public ProjectBackgroundStatus getStatus() {
    return status;
  }

  /**
   * @param status
   *          the status to set
   * @see #status
   */
  public void setStatus(ProjectBackgroundStatus status) {
    this.status = status;
  }

  /**
   * @return the progress
   * @see #progress
   */
  public double getProgress() {
    return progress;
  }

  /**
   * @param progress
   *          the progress to set
   * @see #progress
   */
  public void setProgress(double progress) {
    this.progress = progress;
  }

  /**
   * @return the description
   * @see #description
   */
  public String getDescription() {
    return description;
  }

  /**
   * @param description
   *          the description to set
   * @see #description
   */
  public void setDescription(String description) {
    this.description = description;
  }

  /**
   * @return the hierarchyViewLevel
   * @see #hierarchyViewLevel
   */
  public Integer getHierarchyViewLevel() {
    return hierarchyViewLevel;
  }

  /**
   * @param hierarchyViewLevel
   *          the hierarchyViewLevel to set
   * @see #hierarchyViewLevel
   */
  public void setHierarchyViewLevel(Integer hierarchyViewLevel) {
    this.hierarchyViewLevel = hierarchyViewLevel;
  }

  public boolean isDefaultOverlay() {
    return defaultOverlay;
  }

  public void setDefaultOverlay(boolean defaultOverlay) {
    this.defaultOverlay = defaultOverlay;
  }

  @JsonProperty("order")
  public int getOrderIndex() {
    return orderIndex;
  }

  public void setOrderIndex(int orderIndex) {
    this.orderIndex = orderIndex;
  }

  public Project getProject() {
    return project;
  }

  public void setProject(Project project) {
    Project oldProject = this.project;
    this.project = project;
    if (project == null && oldProject != null) {
      oldProject.removeProjectBackground(this);
    }
  }

  @JsonProperty("images")
  public Set<ProjectBackgroundImageLayer> getProjectBackgroundImageLayer() {
    return backgrounds;
  }

  public void setProjectBackgroundImageLayer(Set<ProjectBackgroundImageLayer> dataOverlayImageLayers) {
    this.backgrounds = dataOverlayImageLayers;
  }
}
