package lcsb.mapviewer.persist.dao.map.species;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.query.Query;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Repository;

import lcsb.mapviewer.common.exception.NotImplementedException;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.persist.dao.BaseDao;

/**
 * Data Access Object class for {@link Element} class.
 * 
 * @author Piotr Gawron
 * 
 */
@Repository
public class ElementDao extends BaseDao<Element> {

  Logger logger = LogManager.getLogger();

  public ElementDao() {
    super(Element.class);
  }

  public List<Element> getByCoordinates(ModelData map, Point2D point, int numberOfElements,
      List<Class<? extends BioEntity>> classes) {
    String typeClause;
    if (classes.size() > 0) {
      typeClause = " type(element) in :classes ";
    } else {
      typeClause = " not element_type_db like '%COMPARTMENT' and not element_type_db like '%PATHWAY' ";
    }
    String order = "z desc, width*height";
    org.hibernate.query.Query<?> query = getSession()
        .createQuery(" from " + this.getClazz().getSimpleName() + " element where "
            + "model = :map and "
            + "x <= :x and "
            + "y <= :y and "
            + "x + width >= :x and "
            + "y + height >= :y and "
            + typeClause
            + removableAndStatemant()
            + " order by " + order)
        .setParameter("map", map)
        .setParameter("x", point.getX())
        .setParameter("y", point.getY())
        .setMaxResults(numberOfElements);

    if (classes.size() > 0) {
      query.setParameter("classes", classes);
    }

    List<?> list = query.list();

    List<Element> result = genericListToElementList(list);

    if (numberOfElements > result.size()) {
      // first fetch the ids, because otherwise subselect does not work (there
      // is a
      // bug in hibernate that ads named parameters for subselects from order
      // clause
      // that are not used there)
      query = getSession()
          .createQuery("select element.id from " + this.getClazz().getSimpleName() + " element where "
              + "model = :map and"
              + typeClause
              + removableAndStatemant()
              + " order by "
              + "least("
              + "function('dist_to_segment', :x, :y, x, y, x, y+height),"
              + "function('dist_to_segment', :x, :y, x, y+height, x+width, y+height),"
              + "function('dist_to_segment', :x, :y, x+width, y+height, x+width, y),"
              + "function('dist_to_segment', :x, :y, x+width, y, x, y)"
              + " ), " + order)
          .setParameter("map", map)
          .setParameter("x", point.getX())
          .setParameter("y", point.getY())
          .setMaxResults(numberOfElements);

      if (classes.size() > 0) {
        query.setParameter("classes", classes);
      }

      List<?> orderedList = query.list();
      if (orderedList.size() > 0) {
        Map<Integer, Integer> orderId = new HashMap<>();
        for (int i = 0; i < orderedList.size(); i++) {
          orderId.put((Integer) orderedList.get(i), i);
        }

        // and now fetch the elements
        list = getSession()
            .createQuery("from " + this.getClazz().getSimpleName() + " element where element.id in :ids")
            .setParameter("ids", orderedList).list();
        List<Element> elements = genericListToElementList(list);
        elements.sort(new Comparator<Element>() {
          @Override
          public int compare(Element o1, Element o2) {
            return orderId.get(o1.getId()) - orderId.get(o2.getId());
          }
        });
        for (Element element : elements) {
          if (!result.contains(element) && numberOfElements > result.size()) {
            result.add(element);
          }
        }
      }
    }

    return result;
  }

  public Page<Element> getByFilter(Pageable pageable, Map<ElementProperty, List<? extends Object>> filter) {
    String queryString = "select element from " + this.getClazz().getSimpleName() + " element "
        + getJoinsForFilter(filter)
        + getWhereByFilter(filter);
    queryString += " group by element ";

    Query<?> query = getSession().createQuery(queryString);

    if (pageable.isPaged()) {
      query.setFirstResult(pageable.getPageNumber() * pageable.getPageSize());
      query.setMaxResults(pageable.getPageSize());
    }

    assignFilterOptions(query, filter);
    List<Element> list = genericListToElementList(query.list());
    if (pageable.getSort().isUnsorted()) {
      list.sort(new Comparator<Element>() {
        @Override
        public int compare(Element arg0, Element arg1) {
          return Double.compare(getSearchPriority(arg0, filter), getSearchPriority(arg1, filter));
        }
      });
    } else {
      throw new NotImplementedException("Sorting is not implemented");
    }

    return new PageImpl<Element>(list, pageable, getCountByFilter(filter));
  }

  private double getSearchPriority(Element element, Map<ElementProperty, List<? extends Object>> filter) {
    if (filter.get(ElementProperty.TEXT) != null) {
      String textQuery = stripString((String) filter.get(ElementProperty.TEXT).get(0));
      if (stripString(element.getName()).contains(textQuery)) {
        return 1 - (textQuery.length() / (double) stripString(element.getName()).length() + 1);
      }
      for (String synonym : element.getSynonyms()) {
        if (stripString(synonym).contains(textQuery)) {
          return 2 - (textQuery.length() / (double) stripString(synonym).length() + 1);
        }
      }
      if (stripString(element.getFullName()).contains(textQuery)) {
        return 3 - (textQuery.length() / (double) stripString(element.getFullName()).length() + 1);
      }
      for (String formerSymbol : element.getFormerSymbols()) {
        if (stripString(formerSymbol).contains(textQuery)) {
          return 4 - (textQuery.length() / (double) stripString(formerSymbol).length() + 1);
        }
      }
      if (stripString(element.getElementId()).contains(textQuery)) {
        return 5 - (textQuery.length() / (double) stripString(element.getElementId()).length() + 1);
      }
      return 6;
    } else if (filter.get(ElementProperty.PERFECT_TEXT) != null) {
      String textQuery = stripString((String) filter.get(ElementProperty.PERFECT_TEXT).get(0));
      if (stripString(element.getName()).equals(textQuery)) {
        return 1;
      }
      for (String synonym : element.getSynonyms()) {
        if (stripString(synonym).equals(textQuery)) {
          return 2;
        }
      }
      if (stripString(element.getFullName()).equals(textQuery)) {
        return 3;
      }
      for (String formerSymbol : element.getFormerSymbols()) {
        if (stripString(formerSymbol).equals(textQuery)) {
          return 4;
        }
      }
      if (stripString(element.getElementId()).equals(textQuery)) {
        return 5;
      }
      return 6;
    } else {
      return element.getId();
    }
  }

  private String stripString(String name) {
    if (name == null) {
      return "";
    }
    return name.toLowerCase().replaceAll("[^a-z0-9]", "");
  }

  private void assignFilterOptions(Query<?> query, Map<ElementProperty, List<? extends Object>> filter) {
    for (ElementProperty property : filter.keySet()) {
      List<? extends Object> list = filter.get(property);
      if (list != null && list.size() > 0) {
        switch (property) {
          case ANNOTATION:
            for (int i = 0; i < list.size(); i++) {
              MiriamData annotation = (MiriamData) list.get(i);
              query.setParameter("type" + i, annotation.getDataType());
              query.setParameter("resource" + i, annotation.getResource().toLowerCase());
            }
            break;
          case CLASS:
            query.setParameter("classes", list);
            break;
          case ID:
            query.setParameter("ids", list);
            break;
          case MAP:
            query.setParameter("maps", list);
            break;
          case NAME:
            List<String> names = new ArrayList<>();
            for (Object object : list) {
              names.add(((String) object).toLowerCase());
            }
            query.setParameter("names", names);
            break;
          case PROJECT:
            query.setParameter("projects", list);
            break;
          case SUBMAP_CONNECTION:
            break;
          case TEXT: {
            String text = ((String) list.get(0)).toLowerCase().trim().replaceAll("[^a-z0-9]", "");
            query.setParameter("query", "%" + text + "%");
            break;
          }
          case PERFECT_TEXT: {
            String text = ((String) list.get(0)).toLowerCase().trim().replaceAll("[^a-z0-9]", "");
            query.setParameter("query", text);
            if (query.getParameterMetadata().getNamedParameterNames().contains("query_integer")) {
              query.setParameter("query_integer", Integer.valueOf(text));
            }
            break;
          }
          default:
            throw new NotImplementedException("Search by " + property + " not implemented");
        }
      }
    }
  }

  private String getWhereByFilter(Map<ElementProperty, List<? extends Object>> filter) {
    List<String> statements = new ArrayList<>();

    for (ElementProperty property : filter.keySet()) {
      List<? extends Object> list = filter.get(property);
      if (list != null && list.size() > 0) {
        switch (property) {
          case ANNOTATION:
            StringBuilder statement = new StringBuilder("(");
            for (int i = 0; i < list.size(); i++) {
              if (i > 0) {
                statement.append(" or ");
              }
              statement.append("md.dataType = :type" + i + " and lower(md.resource) = :resource" + i + " ");
            }
            statement.append(")");
            statements.add(statement.toString());
            break;
          case CLASS:
            statements.add(" type(element) in :classes ");
            break;
          case ID:
            statements.add(" id in :ids ");
            break;
          case MAP:
            statements.add(" element.model in :maps ");
            break;
          case NAME:
            statements.add(" lower(element.name) in :names ");
            break;
          case PROJECT:
            statements.add(" element.model.project in :projects ");
            break;
          case SUBMAP_CONNECTION:
            if (list.size() != 1) {
              throw new NotImplementedException("Only single true false value was expected");
            }
            if (list.contains(Boolean.TRUE)) {
              statements.add(" element.submodel is not null");
            } else if (list.contains(Boolean.FALSE)) {
              statements.add(" element.submodel is null");
            } else {
              throw new NotImplementedException("Only single true false value was expected");
            }
            break;
          case TEXT: {
            String searchClause = "(" +
                " REGEXP_REPLACE(lower(element.elementId),'[^a-z0-9]', '','g') like  :query " +
                " OR REGEXP_REPLACE(lower(element.name),'[^a-z0-9]', '','g') like  :query " +
                " OR REGEXP_REPLACE(lower(element.fullName),'[^a-z0-9]', '','g') like  :query " +
                " OR REGEXP_REPLACE(lower(formerSymbols),'[^a-z0-9]', '','g') like  :query " +
                " OR REGEXP_REPLACE(lower(synonyms),'[^a-z0-9]', '','g') like  :query " +
                ")";
            statements.add(searchClause);
            break;
          }
          case PERFECT_TEXT: {
            String searchClause = "(" +
                " REGEXP_REPLACE(lower(element.elementId),'[^a-z0-9]', '','g') = :query " +
                " OR REGEXP_REPLACE(lower(element.name),'[^a-z0-9]', '','g') = :query " +
                " OR REGEXP_REPLACE(lower(element.fullName),'[^a-z0-9]', '','g') = :query " +
                " OR REGEXP_REPLACE(lower(formerSymbols),'[^a-z0-9]', '','g') = :query " +
                " OR REGEXP_REPLACE(lower(synonyms),'[^a-z0-9]', '','g') = :query ";
            String indexQuery = (String) list.get(0);
            if (StringUtils.isNumeric(indexQuery)) {
              searchClause += " or element.id = :query_integer ";
            }
            searchClause += ")";
            statements.add(searchClause);
            break;
          }
          default: {

            throw new NotImplementedException("Search by " + property + " not implemented");
          }
        }
      }
    }

    if (statements.size() > 0) {
      return " where " + StringUtils.join(statements, " AND ");
    } else {
      return "";
    }
  }

  private String getJoinsForFilter(Map<ElementProperty, List<? extends Object>> filter) {
    String result = "";
    for (ElementProperty property : filter.keySet()) {
      List<? extends Object> list = filter.get(property);
      if (list != null && list.size() > 0) {
        switch (property) {
          case ANNOTATION:
            result += " inner join element.miriamData md ";
            break;
          case CLASS:
            break;
          case ID:
            break;
          case MAP:
            break;
          case NAME:
            break;
          case PROJECT:
            result += " inner join element.model map ";
            break;
          case SUBMAP_CONNECTION:
            break;
          case TEXT:
            result += " left join element.synonyms synonyms ";
            result += " left join element.formerSymbols formerSymbols ";
            break;
          case PERFECT_TEXT:
            result += " left join element.synonyms synonyms ";
            result += " left join element.formerSymbols formerSymbols ";
            break;
          default:
            throw new NotImplementedException("Search by " + property + " not implemented");
        }
      }
    }
    return result;
  }

  private List<Element> genericListToElementList(List<?> list) {
    List<Element> result = new ArrayList<>();

    for (Object object : list) {
      result.add((Element) object);
    }
    return result;
  }

  public List<String> getSuggestedQueries(List<ModelData> maps) {
    Set<String> entries = new HashSet<>();
    List<?> list = getSession()
        .createQuery(
            "select distinct(trim(both ' ' from lower(name))) from " + this.getClazz().getSimpleName() + " element "
                + " where element.model in :maps")
        .setParameter("maps", maps).list();
    for (Object object : list) {
      if (object != null) {
        entries.add((String) object);
      }
    }

    list = getSession()
        .createQuery(
            "select distinct(trim(both ' ' from lower(fullName))) from " + this.getClazz().getSimpleName() + " element "
                + " where element.model in :maps")
        .setParameter("maps", maps).list();
    for (Object object : list) {
      if (object != null) {
        entries.add((String) object);
      }
    }

    list = getSession()
        .createQuery(
            "select distinct(trim(both ' ' from lower(synonyms))) from " + this.getClazz().getSimpleName() + " element "
                + " inner join element.synonyms synonyms"
                + " where element.model in :maps")
        .setParameter("maps", maps).list();
    for (Object object : list) {
      if (object != null) {
        entries.add((String) object);
      }
    }

    list = getSession()
        .createQuery("select distinct(trim(both ' ' from lower(formerSymbols))) from " + this.getClazz().getSimpleName()
            + " element "
            + " inner join element.formerSymbols formerSymbols"
            + " where element.model in :maps")
        .setParameter("maps", maps).list();
    for (Object object : list) {
      if (object != null) {
        entries.add((String) object);
      }
    }

    List<String> result = new ArrayList<>(entries);
    Collections.sort(result);
    return result;

  }

  public Map<MiriamType, Integer> getAnnotationStatistics(List<ModelData> maps) {
    Map<MiriamType, Integer> result = new HashMap<>();
    List<?> list = getSession()
        .createQuery(
            "select md.dataType, count(*) as counter from " + getClazz().getSimpleName()
                + " element_t join element_t.miriamData md "
                + " where element_t.model in :maps "
                + " group by md.dataType")
        .setParameter("maps", maps).list();
    for (Object object : list) {
      Object[] row = (Object[]) object;
      result.put((MiriamType) row[0], ((Long) row[1]).intValue());
    }
    return result;
  }

  public long getCountByFilter(Map<ElementProperty, List<? extends Object>> filter) {
    String countQueryString = "select count(distinct element) from " + this.getClazz().getSimpleName() + " element "
        + getJoinsForFilter(filter)
        + getWhereByFilter(filter);
    Query<?> countQuery = getSession().createQuery(countQueryString);
    assignFilterOptions(countQuery, filter);
    return (Long) countQuery.list().get(0);
  }

}
