package lcsb.mapviewer.persist.dao.map.layout.alias;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.io.IOException;
import java.util.*;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.data.domain.PageRequest;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.model.*;
import lcsb.mapviewer.model.map.species.*;
import lcsb.mapviewer.modelutils.map.ElementUtils;
import lcsb.mapviewer.persist.*;
import lcsb.mapviewer.persist.dao.map.ModelDao;
import lcsb.mapviewer.persist.dao.map.species.ElementDao;
import lcsb.mapviewer.persist.dao.map.species.ElementProperty;

@RunWith(Parameterized.class)
public class ElementDaoGetByQueryTest extends PersistTestFunctions {

  static ApplicationContext applicationContext;

  @SuppressWarnings("unused")
  private Logger logger = LogManager.getLogger();

  private ElementDao elementDao;

  private List<ModelData> maps;
  
  private static List<ModelData> testMaps;

  private String query;

  private boolean perfect;

  private List<BioEntity> expectedResult;

  private List<Class<? extends BioEntity>> types;

  public ElementDaoGetByQueryTest(String testName, Model map, String query, boolean perfect, Object result) {
    maps = Arrays.asList(map.getModelData());
    this.query = query;
    this.perfect = perfect;
    this.expectedResult = new ArrayList<>();
    if (result == null) {
      this.types = new ElementUtils().getClassesByStringTypes(new ElementUtils().getBioEntityStringTypes());
    } else if (result instanceof BioEntity) {
      BioEntity bioEntity = (BioEntity) result;
      this.types = Arrays.asList(bioEntity.getClass());
      this.expectedResult.add(bioEntity);
    } else {
      this.types = new ArrayList<>();
      List<?> bioEntities = (List<?>) result;
      for (Object object : bioEntities) {
        BioEntity bioEntity = (BioEntity) object;
        expectedResult.add(bioEntity);
        types.add(bioEntity.getClass());
      }
    }

    elementDao = applicationContext.getBean(ElementDao.class);
  }

  @Parameters(name = "{index} : {0}")
  public static Collection<Object[]> data() throws IOException {
    Model map = new ModelFullIndexed(null);
    Protein protein = createProtein();
    protein.setName("SNN-1");
    protein.setFullName("FULL BTR-5");
    protein.addSynonym("fancy synonym");
    protein.addSynonym("Fantastic ");
    protein.addFormerSymbol("former symbol2");
    map.addElement(protein);

    Ion ion = createIon("ion1");
    ion.setName("X");
    map.addElement(ion);

    Protein falseIon = createProtein("ion2");
    falseIon.setName("Ion");
    map.addElement(falseIon);

    Protein falseIon2 = createProtein("ion");
    falseIon2.setFullName("Ion");
    map.addElement(falseIon2);

    applicationContext = new AnnotationConfigApplicationContext(SpringPersistTestConfig.class);

    Collection<Object[]> data = new ArrayList<Object[]>();
    data.add(new Object[] { "existing name perfect match", map, protein.getName(), true, protein });
    data.add(new Object[] { "existing name not-perfect match", map, "snn", false, protein });
    data.add(new Object[] { "not existing name perfect match", map, "snn", true, null });
    data.add(new Object[] { "not existing name not-perfect match", map, "bla", false, null });

    data.add(new Object[] { "existing full name perfect match", map, protein.getFullName(), true, protein });
    data.add(new Object[] { "existing full name not-perfect match", map, "BTR5", false, protein });
    data.add(new Object[] { "not existing full name perfect match", map, "BTR5", true, null });

    data.add(new Object[] { "existing element id perfect match", map, protein.getElementId(), true, protein });
    data.add(new Object[] { "existing element id not-perfect match", map, protein.getElementId().substring(1), false,
        protein });

    data.add(new Object[] { "existing synonym perfect match", map, "fancy synonym", true, protein });
    data.add(new Object[] { "existing synonym not-perfect match", map, "synonym", false, protein });
    data.add(new Object[] { "not existing synonym perfect match", map, "synonym", true, null });

    data.add(new Object[] { "existing former symbol perfect match", map, "former symbol2", true, protein });
    data.add(new Object[] { "existing former symbol not-perfect match", map, "symbol2", false, protein });
    data.add(new Object[] { "not existing former symbol perfect match", map, "symbol2", true, null });

    data.add(new Object[] { "element without synonym and former symbol ", map, ion.getElementId(), false, ion });

    data.add(new Object[] { "type selection", map, "ion", false, ion });

    data.add(new Object[] { "sort order", map, "ion", false, Arrays.asList(falseIon, falseIon2, ion) });

    testMaps = Arrays.asList(map.getModelData());
    return data;
  }

  @BeforeClass
  public static void initialize() {
    DbUtils dbUtils = applicationContext.getBean(DbUtils.class);
    dbUtils.createSessionForCurrentThread();

    ModelDao modelDao = applicationContext.getBean(ModelDao.class);
    for (ModelData map : testMaps) {
      modelDao.add(map);
    }
  }

  @AfterClass
  public static void shutdown() throws Exception {
    DbUtils dbUtils = applicationContext.getBean(DbUtils.class);
    dbUtils.getSessionForCurrentThread().getTransaction().setRollbackOnly();
    dbUtils.closeSessionForCurrentThread();
  }

  @Test
  @Transactional
  public void test() throws Exception {
    Map<ElementProperty, List<? extends Object>> filter = new HashMap<>();
    filter.put(ElementProperty.MAP, maps);
    if (perfect) {
      filter.put(ElementProperty.PERFECT_TEXT, Arrays.asList(query));
    } else {
      filter.put(ElementProperty.TEXT, Arrays.asList(query));
    }
    filter.put(ElementProperty.CLASS, types);
    List<Element> elements = elementDao.getByFilter(PageRequest.of(0, 10), filter).getContent();

    assertNotNull(elements);
    assertEquals(expectedResult, elements);
  }

}
