package lcsb.mapviewer.api.configuration;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

import javax.servlet.ServletContext;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.web.bind.annotation.*;

import lcsb.mapviewer.annotation.services.dapi.DapiConnectionException;
import lcsb.mapviewer.annotation.services.dapi.dto.UserDto;
import lcsb.mapviewer.api.BaseController;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IConfigurationService;

@RestController
@RequestMapping(value = "/configuration", produces = MediaType.APPLICATION_JSON_VALUE)
public class ConfigurationController extends BaseController {

  Logger logger = LogManager.getLogger();

  private ConfigurationRestImpl configurationRestImpl;
  private IConfigurationService configurationService;
  private ServletContext context;

  @Autowired
  public ConfigurationController(ConfigurationRestImpl configurationController,
      IConfigurationService configurationService,
      ServletContext context) {
    this.configurationRestImpl = configurationController;
    this.configurationService = configurationService;
    this.context = context;
  }

  @GetMapping(value = "/")
  public Map<String, Object> getConfiguration(Authentication authentication) {
    Map<String, Object> result = new TreeMap<>();
    result.put("options", getOptions(authentication));
    result.put("imageFormats", configurationRestImpl.getImageFormats());
    result.put("modelFormats", configurationRestImpl.getModelFormats());
    result.put("overlayTypes", configurationRestImpl.getOverlayTypes());
    result.put("elementTypes", configurationRestImpl.getElementTypes());
    result.put("reactionTypes", configurationRestImpl.getReactionTypes());
    result.put("miriamTypes", configurationRestImpl.getMiriamTypes());
    result.put("mapTypes", configurationRestImpl.getMapTypes());
    result.put("mapCanvasTypes", configurationRestImpl.getMapCanvasTypes());
    result.put("unitTypes", configurationRestImpl.getUnitTypes());
    result.put("modificationStateTypes", configurationRestImpl.getModificationStateTypes());
    result.put("privilegeTypes", configurationRestImpl.getPrivilegeTypes());
    result.put("version", configurationService.getSystemSvnVersion(context.getRealPath("/") + "META-INF/"));
    result.put("buildDate", configurationService.getSystemBuild(context.getRealPath("/") + "META-INF/"));
    result.put("gitHash", configurationService.getSystemGitVersion(context.getRealPath("/") + "META-INF/"));
    result.put("annotators", configurationRestImpl.getAnnotators());
    result.put("bioEntityFields", configurationRestImpl.getBioEntityFields());
    return result;
  }

  @GetMapping(value = "/options/")
  public List<Map<String, Object>> getOptions(Authentication authentication) {
    boolean isAdmin = authentication.getAuthorities()
        .contains(new SimpleGrantedAuthority(PrivilegeType.IS_ADMIN.toString()));
    return configurationRestImpl.getAllValues().stream()
        .filter(option -> !((Boolean) option.get("isServerSide")) || isAdmin)
        .collect(Collectors.toList());
  }

  @GetMapping(value = "/dapi/")
  public Map<String, Object> getDapi() {
    return configurationRestImpl.getDapiConfig();
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')")
  @PostMapping(value = "/dapi:registerUser")
  public void registerDapiUser(@RequestBody UserDto user) throws DapiConnectionException, QueryException {
    configurationRestImpl.registerDapiUser(user);
  }

  @GetMapping(value = "/dapi/database/")
  public List<Map<String, Object>> getDapiDatabases() {
    return configurationRestImpl.getDapiDatabases();
  }

  @GetMapping(value = "/dapi/database/{database}/release/")
  public List<Map<String, Object>> getDapiReleases(@PathVariable(value = "database") String database) {
    return configurationRestImpl.getDapiDatabaseReleases(database);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')")
  @PostMapping(value = "/dapi/database/{database}/release/{release:.+}:acceptLicense")
  public void acceptReleasesLicense(@PathVariable(value = "database") String database,
      @PathVariable(value = "release") String release) throws DapiConnectionException {
    configurationRestImpl.acceptDapiRelease(database, release);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')")
  @PatchMapping(value = "/options/{option}")
  public Map<String, Object> updateOption(@RequestBody String body, @PathVariable(value = "option") String option)
      throws IOException, QueryException {
    Map<String, Object> node = parseBody(body);
    Map<String, Object> data = getData(node, "option");
    return configurationRestImpl.updateOption(option, data);
  }
}