package lcsb.mapviewer.api.projects.chemicals;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.data.Chemical;
import lcsb.mapviewer.annotation.data.MeSH;
import lcsb.mapviewer.annotation.services.MeSHParser;
import lcsb.mapviewer.annotation.services.TaxonomyBackend;
import lcsb.mapviewer.annotation.services.dapi.ChemicalSearchException;
import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.api.ElementIdentifierType;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IElementService;
import lcsb.mapviewer.services.search.DbSearchCriteria;
import lcsb.mapviewer.services.search.chemical.IChemicalService;

@Transactional
@Service
public class ChemicalRestImpl extends BaseRestImpl {

  /**
   * Default class logger.
   */
  private Logger logger = LogManager.getLogger();

  private IChemicalService chemicalService;

  private IElementService elementService;

  private MeSHParser meSHParser;

  @Autowired
  public ChemicalRestImpl(final IChemicalService chemicalService, final MeSHParser meSHParser, final IElementService elementService) {
    this.chemicalService = chemicalService;
    this.meSHParser = meSHParser;
    this.elementService = elementService;
  }

  public List<Map<String, Object>> getChemicalsByQuery(final String projectId, final String columns, final String query)
      throws QueryException {
    Project project = getProjectService().getProjectByProjectId(projectId);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }
    if (project.getDisease() == null) {
      throw new QueryException("Project doesn't have disease associated to it");
    }

    Set<String> columnSet = createChemicalColumnSet(columns);

    List<Map<String, Object>> result = new ArrayList<>();

    MiriamData organism = project.getOrganism();
    if (organism == null) {
      organism = TaxonomyBackend.HUMAN_TAXONOMY;
    }
    Chemical chemical = chemicalService.getByName(query,
        new DbSearchCriteria().project(project).organisms(organism).colorSet(0).disease(project.getDisease()));
    if (chemical != null) {
      result.add(prepareChemical(chemical, columnSet, project));
    }

    return result;
  }

  protected Map<String, Object> prepareChemical(final Chemical chemical, final Set<String> columnsSet, final Project project) {
    Map<String, Object> result = new TreeMap<>();

    String description = "Mesh term not available";
    List<String> synonyms = new ArrayList<>();

    MeSH mesh = meSHParser.getMeSH(chemical.getChemicalId());
    if (mesh != null) {
      description = mesh.getDescription();
      synonyms = mesh.getSynonyms();
    } else {
      logger.warn("Mesh used by chemical is invalid: " + chemical.getChemicalId());
    }

    for (String string : columnsSet) {
      String column = string.toLowerCase();
      Object value = null;
      switch (column) {
        case "id":
        case "idobject":
          value = chemical.getChemicalId();
          break;
        case "name":
          value = chemical.getChemicalName();
          break;
        case "references":
          List<Map<String, Object>> references = new ArrayList<>();
          references.add(createAnnotation(chemical.getChemicalId()));
          if (chemical.getCasID() != null) {
            references.add(createAnnotation(chemical.getCasID()));
          }
          value = references;
          break;
        case "directevidencereferences":
          value = createAnnotations(chemical.getDirectEvidencePublication());
          break;
        case "description":
          value = description;
          break;
        case "directevidence":
          if (chemical.getDirectEvidence() != null) {
            value = chemical.getDirectEvidence().getValue();
          }
          break;
        case "synonyms":
          value = synonyms;
          break;
        case "targets":
          value = prepareTargets(chemical.getInferenceNetwork(), project);
          break;
        default:
          value = "Unknown column";
          break;
      }
      result.put(string, value);
    }
    return result;
  }

  protected Set<String> createChemicalColumnSet(final String columns) {
    Set<String> columnsSet = new LinkedHashSet<>();
    if (columns.equals("")) {
      columnsSet.addAll(getAvailableChemicalColumns());
    } else {
      columnsSet.addAll(Arrays.asList(columns.split(",")));
    }
    return columnsSet;
  }

  public List<Map<String, Object>> getChemicalsByTarget(final String projectId, final String targetType,
      final String targetId, final String columns) throws QueryException {
    Project project = getProjectService().getProjectByProjectId(projectId);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }

    if (project.getDisease() == null) {
      throw new QueryException("Project doesn't have disease associated to it");
    }

    Integer dbId = Integer.valueOf(targetId);
    List<Element> targets = new ArrayList<>();
    if (targetType.equals(ElementIdentifierType.ALIAS.getJsName())) {
      targets.add(elementService.getElementById(project.getProjectId(), dbId));
    } else {
      throw new QueryException("Targeting for the type not implemented");
    }
    MiriamData organism = project.getOrganism();
    if (organism == null) {
      organism = TaxonomyBackend.HUMAN_TAXONOMY;
    }

    Set<String> columnSet = createChemicalColumnSet(columns);

    List<Chemical> chemicals = chemicalService.getForTargets(targets,
        new DbSearchCriteria().project(project).organisms(organism).disease(project.getDisease()));

    List<Map<String, Object>> result = new ArrayList<>();

    for (Chemical chemical : chemicals) {
      result.add(prepareChemical(chemical, columnSet, project));
    }

    return result;
  }

  public List<String> getSuggestedQueryList(final String projectId) throws ChemicalSearchException, ObjectNotFoundException {
    Project project = getProjectService().getProjectByProjectId(projectId);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist: " + projectId);
    }
    return chemicalService.getSuggestedQueryList(project, project.getDisease());
  }

  public List<String> getAvailableChemicalColumns() {
    List<String> result = new ArrayList<>();
    result.add("name");
    result.add("references");
    result.add("description");
    result.add("synonyms");
    result.add("id");
    result.add("directEvidenceReferences");
    result.add("directEvidence");
    result.add("targets");
    return result;
  }
}
