package lcsb.mapviewer.api.projects.models.bioEntities.elements;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.TreeMap;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.converter.model.celldesigner.structure.fields.ModificationType;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.layout.graphics.Glyph;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.species.AntisenseRna;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.map.species.field.AbstractSiteModification;
import lcsb.mapviewer.model.map.species.field.BindingRegion;
import lcsb.mapviewer.model.map.species.field.CodingRegion;
import lcsb.mapviewer.model.map.species.field.ModificationResidue;
import lcsb.mapviewer.model.map.species.field.ModificationSite;
import lcsb.mapviewer.model.map.species.field.ProteinBindingDomain;
import lcsb.mapviewer.model.map.species.field.RegulatoryRegion;
import lcsb.mapviewer.model.map.species.field.Residue;
import lcsb.mapviewer.model.map.species.field.StructuralState;
import lcsb.mapviewer.model.map.species.field.Structure;
import lcsb.mapviewer.model.map.species.field.TranscriptionSite;
import lcsb.mapviewer.model.map.species.field.UniprotRecord;
import lcsb.mapviewer.persist.dao.map.species.ElementProperty;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IElementService;
import lcsb.mapviewer.services.interfaces.IModelService;

@Transactional
@Service
public class ElementsRestImpl extends BaseRestImpl {

  private IElementService elementService;
  private IModelService modelService;

  @SuppressWarnings("unused")
  private Logger logger = LogManager.getLogger();

  @Autowired
  public ElementsRestImpl(final IElementService elementService, final IModelService modelService) {
    this.elementService = elementService;
    this.modelService = modelService;
  }

  public List<Map<String, Object>> getElements(final String projectId, final String id, final String columns, final String mapId,
      final String type, final String includedCompartmentIds, final String excludedCompartmentIds)
      throws QueryException {
    Set<Integer> ids = new LinkedHashSet<>();
    if (!id.equals("")) {
      for (String str : id.split(",")) {
        ids.add(Integer.valueOf(str));
      }
    }
    Set<String> types = new LinkedHashSet<>();
    if (!type.isEmpty()) {
      for (String str : type.split(",")) {
        types.add(str.toLowerCase());
      }
    }

    List<ModelData> maps = modelService.getModelsByMapId(projectId, mapId);

    Map<ElementProperty, List<? extends Object>> properties = new HashMap<>();
    properties.put(ElementProperty.MAP, maps);
    properties.put(ElementProperty.ID, new ArrayList<>(ids));

    List<Element> allElements = elementService.getElementsByFilter(properties);

    Set<Compartment> includedCompartments = getCompartments(includedCompartmentIds, allElements);
    Set<Compartment> excludedCompartments = getCompartments(excludedCompartmentIds, allElements);
    Set<String> columnsSet = createElementColumnSet(columns);

    List<Element> elements = new ArrayList<>();

    for (Element element : allElements) {
      if (ids.size() == 0 || ids.contains(element.getId())) {
        if (types.size() == 0 || types.contains(element.getStringType().toLowerCase())) {
          if (matchIncludedExcludedCompartments(element, includedCompartments, excludedCompartments)) {
            elements.add(element);
          }
        }
      }
    }

    elements.sort(BioEntity.ID_COMPARATOR);
    List<Map<String, Object>> result = new ArrayList<>();
    for (Element element : elements) {
      result.add(preparedElement(element, columnsSet));

    }
    return result;
  }

  private boolean matchIncludedExcludedCompartments(final Element element, final Set<Compartment> includedCompartments,
      final Set<Compartment> excludedCompartments) {
    boolean matchIncluded = true;
    boolean matchExcluded = false;

    if (includedCompartments.size() > 0) {
      matchIncluded = matchCompartments(element, includedCompartments);
    }
    if (excludedCompartments.size() > 0) {
      matchExcluded = matchCompartments(element, excludedCompartments);
    }

    return matchIncluded && !matchExcluded;
  }

  private boolean matchCompartments(final Element element, final Set<Compartment> compartmentList) {
    boolean isInside = false;
    if (element != null) {
      for (Compartment compartment : compartmentList) {
        if (compartment.contains(element) && Objects.equals(element.getModelData().getId(), compartment.getModelData().getId())) {
          isInside = true;
        }
      }
    }
    return isInside;
  }

  private Set<Compartment> getCompartments(final String includedCompartmentIds, final List<Element> elements) {
    Set<Compartment> includedCompartments = new LinkedHashSet<>();
    int count = 0;
    if (!includedCompartmentIds.isEmpty()) {
      for (String compartmentId : includedCompartmentIds.split(",")) {
        Integer integerId = Integer.valueOf(compartmentId);
        for (Element element : elements) {
          if (element.getId() == integerId && element instanceof Compartment) {
            includedCompartments.add((Compartment) element);
          }
        }
        count++;
      }
      if (count > 0 && includedCompartments.size() == 0) {
        includedCompartments.add(new Compartment("empty_comp_to_reject_all_elements"));
      }
    }
    return includedCompartments;
  }

  private Map<String, Object> preparedElement(final Element element, final Set<String> columnsSet) {
    Map<String, Object> result = new TreeMap<>();
    for (String string : columnsSet) {
      String column = string.toLowerCase();
      Object value = null;
      switch (column) {
        case "id":
        case "idobject":
          value = element.getId();
          break;
        case "modelid":
          value = element.getModelData().getId();
          break;
        case "elementid":
          value = element.getElementId();
          break;
        case "name":
          value = element.getName();
          break;
        case "type":
          value = element.getStringType();
          break;
        case "symbol":
          value = element.getSymbol();
          break;
        case "fullname":
          value = element.getFullName();
          break;
        case "abbreviation":
          value = element.getAbbreviation();
          break;
        case "compartmentid":
          if (element.getCompartment() != null) {
            value = element.getCompartment().getId();
          }
          break;
        case "complexid":
          if (element instanceof Species) {
            if (((Species) element).getComplex() != null) {
              value = ((Species) element).getComplex().getId();
            }
          }
          break;
        case "initialconcentration":
          if (element instanceof Species) {
            value = ((Species) element).getInitialConcentration();
          }
          break;
        case "initialamount":
          if (element instanceof Species) {
            value = ((Species) element).getInitialAmount();
          }
          break;
        case "boundarycondition":
          if (element instanceof Species) {
            value = ((Species) element).isBoundaryCondition();
          }
          break;
        case "constant":
          if (element instanceof Species) {
            value = ((Species) element).isConstant();
          }
          break;
        case "hypothetical":
          if (element instanceof Species) {
            value = ((Species) element).isHypothetical();
          }
          break;
        case "activity":
          if (element instanceof Species) {
            value = ((Species) element).getActivity();
          }
          break;
        case "references":
          value = createAnnotations(element.getMiriamData());
          break;
        case "synonyms":
          value = element.getSynonyms();
          break;
        case "homomultimer":
          if (element instanceof Species) {
            value = ((Species) element).getHomodimer();
          } else {
            value = null;
          }
          break;
        case "formula":
          value = element.getFormula();
          break;
        case "notes":
          value = element.getNotes();
          break;
        case "other":
          value = getOthersForElement(element);
          break;
        case "formersymbols":
          value = element.getFormerSymbols();
          break;
        case "hierarchyvisibilitylevel":
          value = element.getVisibilityLevel();
          break;
        case "transparencylevel":
          value = element.getTransparencyLevel();
          break;
        case "linkedsubmodel":
          if (element.getSubmodel() != null) {
            value = element.getSubmodel().getSubmodel().getId();
          }
          break;
        case "bounds":
          value = createBounds(element.getX(), element.getY(), element.getZ(), element.getWidth(), element.getHeight());
          break;
        case "glyph":
          value = createGlyph(element.getGlyph());
          break;
        default:
          value = "Unknown column";
          break;
      }
      result.put(string, value);
    }
    return result;
  }

  private Map<String, Object> createGlyph(final Glyph glyph) {
    if (glyph == null) {
      return null;
    } else {
      Map<String, Object> result = new HashMap<>();
      result.put("fileId", glyph.getFile().getId());
      return result;
    }
  }

  protected Map<String, Object> getOthersForElement(final Element element) {
    Map<String, Object> result = new TreeMap<>();
    List<Map<String, Object>> modifications = new ArrayList<>();
    StructuralState structuralState = null;
    Map<String, Object> structures = new TreeMap<>();
    if (element instanceof Protein) {
      Protein protein = ((Protein) element);
      modifications = getModifications(protein.getModificationResidues());
    } else if (element instanceof Rna) {
      Rna rna = ((Rna) element);
      modifications = getModifications(rna.getModificationResidues());
    } else if (element instanceof AntisenseRna) {
      AntisenseRna antisenseRna = ((AntisenseRna) element);
      modifications = getModifications(antisenseRna.getModificationResidues());
    } else if (element instanceof Gene) {
      Gene gene = ((Gene) element);
      modifications = getModifications(gene.getModificationResidues());
    }
    if (element instanceof Species) {
      Species species = (Species) element;
      structures = getStructures(species.getUniprots());
      structuralState = species.getStructuralState();
    }
    result.put("modifications", modifications);
    if (structuralState != null) {
      result.put("structuralState", structuralState.getValue());
    } else {
      result.put("structuralState", null);
    }
    result.put("structures", structures);

    return result;
  }

  private List<Map<String, Object>> getModifications(final List<? extends ModificationResidue> elements) {
    List<Map<String, Object>> result = new ArrayList<>();
    for (ModificationResidue region : elements) {
      Map<String, Object> row = new TreeMap<>();
      row.put("name", region.getName());
      row.put("modificationId", region.getIdModificationResidue());
      if (region instanceof AbstractSiteModification) {
        AbstractSiteModification siteModification = ((AbstractSiteModification) region);
        if (siteModification.getState() != null) {
          row.put("state", siteModification.getState().name());
        }
      }
      String type;
      if (region instanceof Residue) {
        type = ModificationType.RESIDUE.name();
      } else if (region instanceof BindingRegion) {
        type = ModificationType.BINDING_REGION.name();
      } else if (region instanceof CodingRegion) {
        type = ModificationType.CODING_REGION.name();
      } else if (region instanceof ProteinBindingDomain) {
        type = ModificationType.PROTEIN_BINDING_DOMAIN.name();
      } else if (region instanceof RegulatoryRegion) {
        type = ModificationType.REGULATORY_REGION.name();
      } else if (region instanceof TranscriptionSite) {
        if (((TranscriptionSite) region).getDirection().equals("LEFT")) {
          type = ModificationType.TRANSCRIPTION_SITE_LEFT.name();
        } else {
          type = ModificationType.TRANSCRIPTION_SITE_RIGHT.name();
        }
      } else if (region instanceof ModificationSite) {
        type = ModificationType.MODIFICATION_SITE.name();
      } else {
        throw new InvalidArgumentException("Unknown class: " + region.getClass());
      }
      row.put("type", type);
      result.add(row);
    }
    return result;
  }

  private Map<String, Object> getStructures(final Set<UniprotRecord> uniprots) {
    Map<String, Object> result = new TreeMap<>();
    for (UniprotRecord uniprotRec : uniprots) {
      Set<Object> structs = new LinkedHashSet<>();
      for (Structure struct : uniprotRec.getStructures()) {
        structs.add(struct.toMap());
      }
      result.put(uniprotRec.getUniprotId(), structs);
    }
    return result;
  }

  private Map<String, Object> createBounds(final Double x, final Double y, final Integer z, final Double width, final Double height) {
    Map<String, Object> result = new TreeMap<>();
    result.put("x", x);
    result.put("y", y);
    result.put("z", z);
    result.put("width", width);
    result.put("height", height);
    return result;
  }

  public List<String> getAvailableElementColumns() {
    List<String> result = new ArrayList<>();
    result.add("id");
    result.add("elementId");
    result.add("modelId");
    result.add("name");
    result.add("type");
    result.add("notes");
    result.add("symbol");
    result.add("complexId");
    result.add("compartmentId");
    result.add("fullName");
    result.add("abbreviation");
    result.add("formula");
    result.add("synonyms");
    result.add("formerSymbols");
    result.add("references");
    result.add("bounds");
    result.add("hierarchyVisibilityLevel");
    result.add("transparencyLevel");
    result.add("linkedSubmodel");
    result.add("other");
    result.add("initialConcentration");
    result.add("boundaryCondition");
    result.add("constant");
    result.add("hypothetical");
    result.add("activity");
    result.add("initialAmount");
    result.add("glyph");
    result.add("homomultimer");
    return result;
  }

  private Set<String> createElementColumnSet(final String columns) {
    Set<String> columnsSet = new LinkedHashSet<>();
    if (columns.equals("")) {
      columnsSet.addAll(getAvailableElementColumns());
    } else {
      columnsSet.addAll(Arrays.asList(columns.split(",")));
    }
    return columnsSet;
  }

}
