package lcsb.mapviewer.api.projects.models.publications;

import java.util.*;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.*;
import org.springframework.data.domain.Sort.Direction;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.services.PubmedParser;
import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.model.Article;
import lcsb.mapviewer.model.map.*;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.persist.dao.ArticleProperty;
import lcsb.mapviewer.persist.dao.map.ReactionProperty;
import lcsb.mapviewer.persist.dao.map.species.ElementProperty;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.*;

@Transactional
@Service
public class PublicationsRestImpl extends BaseRestImpl {

  @SuppressWarnings("unused")
  private Logger logger = LogManager.getLogger();

  private PubmedParser pubmedParser;

  private IPublicationService publicationService;

  private IElementService elementService;;
  private IReactionService reactionService;

  @Autowired
  public PublicationsRestImpl(PubmedParser pubmedParser, IPublicationService publicationService,
      IElementService elementService,
      IReactionService reactionService) {
    this.pubmedParser = pubmedParser;
    this.publicationService = publicationService;
    this.elementService = elementService;
    this.reactionService = reactionService;
  }

  public SortedMap<MiriamData, List<BioEntity>> getPublications(Collection<Model> models) {
    SortedMap<MiriamData, List<BioEntity>> publications = new TreeMap<>();

    List<BioEntity> bioEntities = new ArrayList<>();
    for (Model modelData : models) {
      bioEntities.addAll(modelData.getBioEntities());
    }
    bioEntities.sort(BioEntity.ID_COMPARATOR);
    for (BioEntity bioEntity : bioEntities) {
      for (MiriamData md : bioEntity.getMiriamData()) {
        if (md.getDataType().equals(MiriamType.PUBMED)) {
          List<BioEntity> list = publications.computeIfAbsent(md, k -> new ArrayList<>());
          list.add(bioEntity);
        }
      }

    }
    return publications;
  }

  public Map<String, Object> getPublications(
      String projectId, String modelId, String pageString, Integer length,
      String sortColumn, String sortOrder, String search) throws QueryException {

    int page = parseInteger(pageString, "page");
    if (length <= 0) {
      throw new QueryException("Page length must not be less than one");
    }
    ArticleProperty sortProperty = extractSortColumn(sortColumn);
    Pageable pageable;
    if (sortProperty != null) {
      pageable = PageRequest.of(page, length, Sort.by(Direction.fromString(sortOrder), sortProperty.name()));
    } else {
      pageable = PageRequest.of(page, length);

    }
    Map<ArticleProperty, Object> filterOptions = new HashMap<>();

    List<ModelData> models = getModelService().getModelsByMapId(projectId, modelId);
    filterOptions.put(ArticleProperty.MODEL, models);
    long count = publicationService.getCountByFilter(filterOptions);

    if (search != null && !search.trim().equals("")) {
      filterOptions.put(ArticleProperty.TEXT, search);
    }
    Page<Article> articles = publicationService.getByFilter(pageable, filterOptions);

    List<MiriamData> mds = new ArrayList<>();
    for (Article article : articles) {
      mds.add(new MiriamData(MiriamType.PUBMED, article.getPubmedId()));
    }

    Map<ElementProperty, List<? extends Object>> elementSearch = new HashMap<>();
    elementSearch.put(ElementProperty.ANNOTATION, mds);
    elementSearch.put(ElementProperty.MAP, models);

    List<BioEntity> bioEntities = new ArrayList<>();
    bioEntities.addAll(elementService.getElementsByFilter(elementSearch));

    Map<ReactionProperty, List<? extends Object>> reactionSearch = new HashMap<>();
    reactionSearch.put(ReactionProperty.ANNOTATION, mds);
    reactionSearch.put(ReactionProperty.MAP, models);

    bioEntities.addAll(reactionService.getReactionsByFilter(reactionSearch).getContent());

    List<Map<String, Object>> resultList = new ArrayList<>();

    Map<MiriamData, List<BioEntity>> publications = new HashMap<>();

    for (BioEntity bioEntity : bioEntities) {
      for (MiriamData md : bioEntity.getMiriamData()) {
        if (md.getDataType().equals(MiriamType.PUBMED)) {
          List<BioEntity> list = publications.computeIfAbsent(md, k -> new ArrayList<>());
          list.add(bioEntity);
        }
      }

    }

    for (Article article : articles) {
      List<Object> elements = new ArrayList<>();
      for (BioEntity bioEntity : publications.computeIfAbsent(new MiriamData(MiriamType.PUBMED, article.getPubmedId()),
          k -> new ArrayList<>())) {
        elements.add(createMinifiedSearchResult(bioEntity));
      }

      Map<String, Object> row = new TreeMap<>();
      row.put("elements", elements);
      Map<String, Object> articleEntry = new HashMap<>();
      articleEntry.put("article", article);
      row.put("publication", articleEntry);
      resultList.add(row);
    }

    Map<String, Object> result = new TreeMap<>();
    result.put("data", resultList);
    result.put("totalSize", (int) count);
    result.put("filteredSize", articles.getTotalElements());
    result.put("page", page);
    result.put("length", resultList.size());
    return result;
  }

  private ArticleProperty extractSortColumn(String sortColumn) throws QueryException {
    if (sortColumn == null) {
      return null;
    }
    switch (sortColumn.toLowerCase().trim()) {
    case "":
      return null;
    case "pubmedid":
      return ArticleProperty.PUBMED_ID;
    case "year":
      return ArticleProperty.YEAR;
    case "journal":
      return ArticleProperty.JOURNAL;
    case "title":
      return ArticleProperty.TITLE;
    default:
      throw new QueryException("Unknown sort order: " + sortColumn);
    }
  }

  /**
   * @return the pubmedParser
   * @see #pubmedParser
   */
  public PubmedParser getPubmedParser() {
    return pubmedParser;
  }

  /**
   * @param pubmedParser
   *          the pubmedParser to set
   * @see #pubmedParser
   */
  public void setPubmedParser(PubmedParser pubmedParser) {
    this.pubmedParser = pubmedParser;
  }

}
