package lcsb.mapviewer.api.projects.overlays;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PostAuthorize;
import org.springframework.security.access.prepost.PostFilter;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PatchMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import lcsb.mapviewer.api.BaseController;
import lcsb.mapviewer.api.SerializationUtils;
import lcsb.mapviewer.model.cache.FileEntry;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IUserService;

@RestController
@RequestMapping(value = "/projects/{projectId}/overlays", produces = MediaType.APPLICATION_JSON_VALUE)
public class OverlayController extends BaseController {

  Logger logger = LogManager.getLogger();

  private OverlayRestImpl overlayRestImp;
  private IUserService userService;

  public OverlayController(final OverlayRestImpl overlayRestImp,
      final IUserService userService) {
    this.overlayRestImp = overlayRestImp;
    this.userService = userService;
  }

  @PreAuthorize("hasAnyAuthority('IS_ADMIN', 'READ_PROJECT:' + #projectId) "
      + "or not @projectService.projectExists(#projectId)")
  @PostFilter("hasAuthority('IS_ADMIN')" +
      " or hasAuthority('IS_CURATOR') and hasAuthority('READ_PROJECT:' + #projectId)" +
      " or hasAuthority('READ_PROJECT:' + #projectId) and (filterObject['creator'] == authentication.name or filterObject['publicOverlay'])")
  @GetMapping(value = "/")
  public List<Map<String, Object>> getOverlayList(
      @PathVariable(value = "projectId") final String projectId,
      @RequestParam(value = "creator", defaultValue = "") final String creator,
      @RequestParam(value = "publicOverlay", defaultValue = "false") final boolean publicOverlay)
      throws lcsb.mapviewer.services.ObjectNotFoundException {
    return overlayRestImp.getOverlayList(projectId).stream()
        .filter(overlay -> !publicOverlay || (Boolean) overlay.get("publicOverlay"))
        .filter(
            overlay -> creator.isEmpty() || (overlay.get("creator") != null && overlay.get("creator").equals(creator)))
        .collect(Collectors.toList());
  }

  @PostAuthorize("hasAuthority('IS_ADMIN')" +
      " or hasAuthority('IS_CURATOR') and hasAuthority('READ_PROJECT:' + #projectId)" +
      " or hasAuthority('READ_PROJECT:' + #projectId) and (returnObject['creator'] == authentication.name or returnObject['publicOverlay'])")
  @GetMapping(value = "/{overlayId}/")
  public Map<String, Object> getOverlayById(
      @PathVariable(value = "projectId") final String projectId,
      @PathVariable(value = "overlayId") final Integer overlayId) throws QueryException {
    return overlayRestImp.getOverlayById(projectId, overlayId);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')" +
      " or hasAuthority('IS_CURATOR') and hasAuthority('READ_PROJECT:' + #projectId)" +
      " or hasAuthority('READ_PROJECT:' + #projectId) and " +
      "   (@dataOverlayService.getOverlayCreator(this.getId(#overlayId))?.login == authentication.name " +
      "    or @dataOverlayService.getDataOverlayById(this.getId(#overlayId))?.public == true)")
  @GetMapping(value = "/{overlayId}/models/{modelId}/bioEntities/")
  public List<Map<String, Object>> getOverlayElements(
      @PathVariable(value = "projectId") final String projectId,
      @PathVariable(value = "modelId") final String modelId,
      @PathVariable(value = "overlayId") final String overlayId,
      @RequestParam(value = "columns", defaultValue = "") final String columns) throws QueryException {

    return overlayRestImp.getOverlayElements(projectId, modelId, SerializationUtils.parseInteger(overlayId, "overlayId"),
        columns);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')" +
      " or hasAuthority('IS_CURATOR') and hasAuthority('READ_PROJECT:' + #projectId)" +
      " or hasAuthority('READ_PROJECT:' + #projectId) and " +
      "   (@dataOverlayService.getOverlayCreator(#overlayId)?.login == authentication.name or @dataOverlayService.getDataOverlayById(#overlayId)?.public == true)")
  @GetMapping(value = "/{overlayId}/models/{modelId}/bioEntities/reactions/{reactionId}/")
  public List<Map<String, Object>> getFullReaction(
      @PathVariable(value = "projectId") final String projectId,
      @PathVariable(value = "modelId") final String modelId,
      @PathVariable(value = "overlayId") final Integer overlayId,
      @PathVariable(value = "reactionId") final String reactionId,
      @RequestParam(value = "columns", defaultValue = "") final String columns)
      throws QueryException, NumberFormatException, ObjectNotFoundException {

    return overlayRestImp.getOverlayElement(projectId,
        SerializationUtils.parseInteger(modelId, "modelId"),
        SerializationUtils.parseInteger(overlayId, "overlayId"),
        SerializationUtils.parseInteger(reactionId, "reactionId"),
        "REACTION", columns);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')" +
      " or hasAuthority('IS_CURATOR') and hasAuthority('READ_PROJECT:' + #projectId)" +
      " or hasAuthority('READ_PROJECT:' + #projectId) and " +
      "   (@dataOverlayService.getOverlayCreator(#overlayId)?.login == authentication.name or @dataOverlayService.getDataOverlayById(#overlayId)?.public == true)")
  @GetMapping(value = "/{overlayId}/models/{modelId}/bioEntities/elements/{elementId}/")
  public List<Map<String, Object>> getFullSpecies(
      @PathVariable(value = "projectId") final String projectId,
      @PathVariable(value = "modelId") final String modelId,
      @PathVariable(value = "overlayId") final Integer overlayId,
      @PathVariable(value = "elementId") final String reactionId,
      @RequestParam(value = "columns", defaultValue = "") final String columns)
      throws QueryException, NumberFormatException, ObjectNotFoundException {
    return overlayRestImp.getOverlayElement(projectId,
        SerializationUtils.parseInteger(modelId, "modelId"),
        SerializationUtils.parseInteger(overlayId, "overlayId"),
        SerializationUtils.parseInteger(reactionId, "reactionId"),
        "ALIAS", columns);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')" +
      " or (hasAuthority('IS_CURATOR') and hasAuthority('WRITE_PROJECT:' + #projectId))" +
      " or (hasAuthority('IS_CURATOR') and hasAuthority('READ_PROJECT:' + #projectId))" +
      " or (hasAuthority('READ_PROJECT:' + #projectId) and hasAuthority('CAN_CREATE_OVERLAYS'))")
  @PostMapping(value = "/")
  public Map<String, Object> addOverlay(
      final Authentication authentication,
      @PathVariable(value = "projectId") final String projectId,
      @RequestParam(value = "name") final String name,
      @RequestParam(value = "description") final String description,
      @RequestParam(value = "content", defaultValue = "") final String content,
      @RequestParam(value = "fileId", defaultValue = "") final String fileId,
      @RequestParam(value = "filename") final String filename,
      @RequestParam(value = "googleLicenseConsent") final String googleLicenseConsent,
      @RequestParam(value = "type", defaultValue = "") final String type)
      throws QueryException, IOException, SecurityException {
    User user = userService.getUserByLogin(authentication.getName());
    return overlayRestImp.addOverlay(
        projectId, name, description, content, fileId, filename, type, googleLicenseConsent, user);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')" +
      "or hasAuthority('IS_CURATOR') and hasAuthority('WRITE_PROJECT:' + #projectId)" +
      "or hasAuthority('READ_PROJECT:' + #projectId) and @dataOverlayService.getOverlayCreator(#overlayId)?.login == authentication.name")
  @DeleteMapping(value = "/{overlayId}")
  public Map<String, Object> removeOverlay(
      @PathVariable(value = "projectId") final String projectId,
      @PathVariable(value = "overlayId") final Integer overlayId) throws QueryException, IOException {
    return overlayRestImp.removeOverlay(projectId, overlayId);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')" +
      " or hasAuthority('IS_CURATOR') and hasAuthority('WRITE_PROJECT:' + #projectId)" +
      " or hasAuthority('READ_PROJECT:' + #projectId) and @dataOverlayService.getOverlayCreator(#overlayId)?.login == authentication.name")
  @PatchMapping(value = "/{overlayId}")
  public Map<String, Object> updateOverlay(
      @RequestBody final String body,
      @PathVariable(value = "overlayId") final Integer overlayId,
      @PathVariable(value = "projectId") final String projectId)
      throws QueryException, IOException {
    Map<String, Object> node = parseBody(body);
    Map<String, Object> data = getData(node, "overlay");
    return overlayRestImp.updateOverlay(projectId, overlayId, data);
  }

  @PreAuthorize("hasAuthority('IS_ADMIN')" +
      " or hasAuthority('IS_CURATOR') and hasAuthority('READ_PROJECT:' + #projectId)" +
      " or hasAuthority('READ_PROJECT:' + #projectId) and " +
      "   (@dataOverlayService.getOverlayCreator(#overlayId)?.login == authentication.name or @dataOverlayService.getDataOverlayById(#overlayId)?.public == true)")
  @GetMapping(value = "/{overlayId}:downloadSource")
  public ResponseEntity<byte[]> getOverlaySource(
      @PathVariable(value = "projectId") final String projectId,
      @PathVariable(value = "overlayId") final Integer overlayId)
      throws QueryException {

    FileEntry file = overlayRestImp.getOverlaySource(projectId, overlayId);
    MediaType type = MediaType.TEXT_PLAIN;
    String filename = file.getOriginalFileName();
    if (filename != null) {
      if (file.getOriginalFileName().endsWith("xml")) {
        type = MediaType.APPLICATION_XML;
      }
    } else {
      filename = overlayId + ".txt";
    }
    return ResponseEntity.ok().contentLength(file.getFileContent().length).contentType(type)
        .header("Content-Disposition", "attachment; filename=" + filename).body(file.getFileContent());
  }
}