package lcsb.mapviewer.api.projects.overlays;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.stream.Collectors;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.Hibernate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.api.ElementIdentifierType;
import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.common.geometry.ColorParser;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.cache.FileEntry;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.overlay.DataOverlayEntry;
import lcsb.mapviewer.model.overlay.DataOverlayType;
import lcsb.mapviewer.model.overlay.GeneVariantDataOverlayEntry;
import lcsb.mapviewer.model.overlay.InvalidDataOverlayException;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.dao.cache.UploadedFileEntryDao;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IDataOverlayService;
import lcsb.mapviewer.services.interfaces.IDataOverlayService.CreateDataOverlayParams;
import lcsb.mapviewer.services.interfaces.IModelService;

@Transactional
@Service
public class OverlayRestImpl extends BaseRestImpl {

  @SuppressWarnings("unused")
  private Logger logger = LogManager.getLogger();

  private IDataOverlayService dataOverlayService;

  private IModelService modelService;

  private UploadedFileEntryDao uploadedFileEntryDao;

  private ColorParser colorParser = new ColorParser();

  @Autowired
  public OverlayRestImpl(final UploadedFileEntryDao uploadedFileEntryDao,
      final IDataOverlayService dataOverlayService,
      final IModelService modelService) {
    this.uploadedFileEntryDao = uploadedFileEntryDao;
    this.dataOverlayService = dataOverlayService;
    this.modelService = modelService;
  }

  public List<Map<String, Object>> getOverlayList(final String projectId) throws ObjectNotFoundException {
    Project project = getProjectService().getProjectByProjectId(projectId);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }
    List<DataOverlay> dataOverlays = dataOverlayService.getDataOverlaysByProject(project);

    List<Map<String, Object>> result = new ArrayList<>();
    result.addAll(overlaysToMap(dataOverlays));

    return result;
  }

  private List<Map<String, Object>> overlaysToMap(final List<DataOverlay> overlays) {
    List<Map<String, Object>> result = new ArrayList<>();
    for (DataOverlay overlay : overlays) {
      result.add(overlayToMap(overlay));
    }
    return result;
  }

  public Map<String, Object> overlayToMap(final DataOverlay overlay) {

    Map<String, Object> result = new TreeMap<>();
    result.put("idObject", overlay.getId());
    result.put("name", overlay.getName());
    result.put("order", overlay.getOrderIndex());
    result.put("description", overlay.getDescription());
    result.put("publicOverlay", overlay.isPublic());

    result.put("type", overlay.getColorSchemaType());
    result.put("genomeType", overlay.getGenomeType());
    result.put("genomeVersion", overlay.getGenomeVersion());

    result.put("googleLicenseConsent", overlay.isGoogleLicenseConsent());
    if (overlay.getCreator() != null) {
      result.put("creator", overlay.getCreator().getLogin());
    }
    return result;
  }

  public List<Map<String, Object>> getOverlayElements(final String projectId, final String modelId, final int overlayId, final String columns)
      throws QueryException {
    dataOverlayService.getDataOverlayById(projectId, overlayId);

    List<ModelData> models = modelService.getModelsByMapId(projectId, modelId);

    String[] columnSet;
    if (columns != null && !columns.trim().isEmpty()) {
      columnSet = columns.split(",");
    } else {
      columnSet = new String[] { "modelId", "idObject", "value", "color", "uniqueId", "width" };
    }
    List<Map<String, Object>> result = new ArrayList<>();

    for (ModelData model : models) {
      List<Pair<Element, DataOverlayEntry>> speciesList = dataOverlayService.getElementReferencesForDataOverlay(overlayId, Arrays.asList(model));
      for (Pair<Element, DataOverlayEntry> elementDataOverlay : speciesList) {
        Map<String, Object> element = new TreeMap<>();
        element.put("type", ElementIdentifierType.ALIAS);
        element.put("overlayContent", overlayContentToMap(elementDataOverlay, columnSet, model));
        result.add(element);
      }

      List<Pair<Reaction, DataOverlayEntry>> reactions = dataOverlayService.getReactionReferencesForDataOverlay(overlayId, Arrays.asList(model));
      for (Pair<Reaction, DataOverlayEntry> reactionDataOverlay : reactions) {
        Map<String, Object> element = new TreeMap<>();
        element.put("type", ElementIdentifierType.REACTION);
        element.put("overlayContent", overlayContentToMap(reactionDataOverlay, columnSet, model));
        result.add(element);
      }
    }
    return result;
  }

  public Map<String, Object> getOverlayById(final String projectId, final Integer overlayId)
      throws QueryException {
    return overlayToMap(dataOverlayService.getDataOverlayById(projectId, overlayId));
  }

  public FileEntry getOverlaySource(final String projectId, final Integer overlayId)
      throws QueryException {
    DataOverlay overlay = dataOverlayService.getDataOverlayById(projectId, overlayId);
    // lazy initialization issue
    overlay.getInputData().getFileContent();
    return overlay.getInputData();
  }

  public Map<String, Object> updateOverlay(final String projectId, final Integer overlayId, final Map<String, Object> overlayData)
      throws QueryException {
    if (overlayData == null) {
      throw new QueryException("overlay field cannot be undefined");
    }
    List<User> reorderUsers = new ArrayList<>();
    try {
      DataOverlay layout = dataOverlayService.getDataOverlayById(projectId, overlayId);
      for (String key : overlayData.keySet()) {
        Object value = overlayData.get(key);
        if (key.equalsIgnoreCase("description")) {
          layout.setDescription((String) value);
        } else if (key.equalsIgnoreCase("name")) {
          if (value == null || ((String) value).trim().isEmpty()) {
            throw new QueryException("name cannot be empty");
          }
          if (((String) value).length() > 255) {
            throw new QueryException("name too long");
          }
          layout.setName((String) value);
        } else if (key.equalsIgnoreCase("order")) {
          layout.setOrderIndex(parseInteger(value));
        } else if (key.equalsIgnoreCase("type")) {
          layout.setColorSchemaType(DataOverlayType.valueOf((String) value));
        } else if (key.equalsIgnoreCase("publicOverlay")) {
          layout.setPublic(parseBoolean(value));
        } else if (key.equalsIgnoreCase("googleLicenseConsent")) {
          layout.setGoogleLicenseConsent((Boolean) overlayData.get("googleLicenseConsent"));
        } else if (key.equalsIgnoreCase("creator")) {
          if ("".equals(value)) {
            throw new QueryException("overlay creator must be defined");
          } else {
            if (layout.getCreator() == null) {
              reorderUsers.add(getUserService().getUserByLogin((String) value));
            } else if (!layout.getCreator().getLogin().equals(value)) {
              reorderUsers.add(getUserService().getUserByLogin((String) value));
              reorderUsers.add(layout.getCreator());
            }
            layout.setCreator(getUserService().getUserByLogin((String) value));
          }
        } else {
          throw new QueryException("Unknown parameter: " + key);
        }
      }
      dataOverlayService.updateDataOverlay(layout);
      for (User user : reorderUsers) {
        reorderOverlays(user, layout.getProject());
      }
      return getOverlayById(layout.getProject().getProjectId(), overlayId);
    } catch (NumberFormatException e) {
      throw new QueryException("invliad overlay id", e);
    }
  }

  public Map<String, Object> removeOverlay(final String projectId, final Integer overlayId)
      throws QueryException, IOException {
    DataOverlay layout = dataOverlayService.getDataOverlayById(projectId, overlayId);
    User owner = layout.getCreator();
    Project project = layout.getProject();
    dataOverlayService.removeDataOverlay(layout);

    reorderOverlays(owner, project);
    return okStatus();
  }

  private void reorderOverlays(final User owner, final Project project) {
    List<DataOverlay> overlays = dataOverlayService.getDataOverlaysByProject(project).stream()
        .filter(lay -> lay.getCreator() == owner).collect(Collectors.toList());

    overlays.sort(DataOverlay.ORDER_COMPARATOR);
    for (int i = 0; i < overlays.size(); i++) {
      DataOverlay overlay = overlays.get(i);
      if (overlay.getOrderIndex() != i + 1) {
        overlay.setOrderIndex(i + 1);
        dataOverlayService.updateDataOverlay(overlay);
      }
    }
  }

  public Map<String, Object> addOverlay(final String projectId, final String name, final String description, final String content,
      final String fileId, final String filename, final String type, final String googleLicenseConsent, final User user)
      throws QueryException, IOException, SecurityException {
    Project project = getProjectService().getProjectByProjectId(projectId);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }
    DataOverlayType colorSchemaType = DataOverlayType.GENERIC;
    if (type != null && !type.equals("")) {
      try {
        colorSchemaType = DataOverlayType.valueOf(type);
      } catch (IllegalArgumentException e) {
        throw new QueryException("Invalid type of overlay: " + type, e);
      }
    }
    if (content.isEmpty() && fileId.isEmpty()) {
      throw new QueryException("Either content or fileId must be provided");
    }

    if (name == null || name.trim().isEmpty()) {
      throw new QueryException("Name cannot be empty");
    }
    try {
      InputStream stream;
      if (!content.isEmpty()) {
        stream = new ByteArrayInputStream(content.getBytes(StandardCharsets.UTF_8));
      } else {
        try {
          int id = Integer.valueOf(fileId);
          UploadedFileEntry file = uploadedFileEntryDao.getById(id);
          if (file == null) {
            throw new QueryException("Invalid file id: " + fileId);
          }
          if (file.getOwner() == null || !file.getOwner().getLogin().equals(user.getLogin())) {
            throw new SecurityException("Access denied to source file");
          }
          stream = new ByteArrayInputStream(file.getFileContent());
        } catch (NumberFormatException e) {
          throw new QueryException("Invalid fileId: " + fileId);
        }
      }

      DataOverlay layout = dataOverlayService
          .createDataOverlay(new CreateDataOverlayParams()
              .colorInputStream(stream)
              .description(description)
              .fileName(filename)
              .name(name).user(user)
              .colorSchemaType(colorSchemaType)
              .project(project)
              .googleLicenseConsent(googleLicenseConsent.equalsIgnoreCase("true")));

      int count = (int) dataOverlayService.getDataOverlaysByProject(project).stream()
          .filter(overlay -> (overlay.getCreator().getLogin().equals(user.getLogin()) && !overlay.isPublic()))
          .count();
      layout.setOrderIndex(count);
      dataOverlayService.updateDataOverlay(layout);

      return getOverlayById(projectId, layout.getId());
    } catch (InvalidDataOverlayException e) {
      throw new QueryException(e.getMessage(), e);
    }

  }

  public List<Map<String, Object>> getOverlayElement(final String projectId, final Integer modelId, final Integer overlayId,
      final Integer elementId, final String elementType, final String columns) throws QueryException {
    dataOverlayService.getDataOverlayById(projectId, overlayId);
    String[] columnSet;
    if (columns != null && !columns.trim().isEmpty()) {
      columnSet = columns.split(",");
    } else {
      columnSet = new String[] { "modelId", "idObject", "value", "color",
          "description", "type", "geneVariations",
          "uniqueId" };
    }
    List<Map<String, Object>> result = new ArrayList<>();
    if (ElementIdentifierType.ALIAS.getJsName().equals(elementType)) {
      List<Pair<Element, DataOverlayEntry>> elementDataOverlay = dataOverlayService
          .getFullElementForDataOverlay(
              projectId,
              modelId,
              elementId,
              overlayId);
      if (elementDataOverlay.size() == 0) {
        throw new ObjectNotFoundException("Element data cannot be found");
      }
      for (Pair<Element, DataOverlayEntry> pair : elementDataOverlay) {
        Map<String, Object> entry = new HashMap<>();
        entry.put("type", ElementIdentifierType.ALIAS);
        entry.put("overlayContent", overlayContentToMap(pair, columnSet));
        result.add(entry);
      }
      return result;
    } else if (ElementIdentifierType.REACTION.getJsName().equals(elementType)) {
      List<Pair<Reaction, DataOverlayEntry>> reactionDataOverlay = dataOverlayService
          .getFullReactionForDataOverlay(projectId,
              modelId,
              elementId, overlayId);
      if (reactionDataOverlay.size() == 0) {
        throw new ObjectNotFoundException("Reaction data cannot be found");
      }
      for (Pair<Reaction, DataOverlayEntry> pair : reactionDataOverlay) {
        Map<String, Object> entry = new HashMap<>();
        entry.put("type", ElementIdentifierType.REACTION);
        entry.put("overlayContent", overlayContentToMap(pair, columnSet));
        result.add(entry);
      }
      return result;
    } else {
      throw new QueryException("Unknown element type: " + elementType);
    }
  }

  private Map<String, Object> overlayContentToMap(final Pair<? extends BioEntity, DataOverlayEntry> pair, final String[] columns) {
    return overlayContentToMap(pair, columns, pair.getLeft().getModelData());
  }

  private Map<String, Object> overlayContentToMap(final Pair<? extends BioEntity, DataOverlayEntry> bioEntityDataOverlay,
      final String[] columns, final ModelData model) {
    Map<String, Object> result = new LinkedHashMap<>();
    BioEntity bioEntity = bioEntityDataOverlay.getLeft();
    DataOverlayEntry overlayEntry = bioEntityDataOverlay.getRight();
    for (String string : columns) {
      String column = string.toLowerCase();
      Object value = null;
      switch (column) {
        case "id":
        case "idobject":
          // casting to string is only to provide the same results as before
          // refactoring
          value = bioEntity.getId() + "";
          break;
        case "uniqueid":
          value = overlayEntry.getId();
          break;
        case "modelid":
          value = model.getId();
          break;
        case "value":
          value = overlayEntry.getValue();
          break;
        case "color":
          if (overlayEntry.getColor() != null) {
            value = colorParser.colorToMap(overlayEntry.getColor());
          }
          break;
        case "description":
          value = overlayEntry.getDescription();
          break;
        case "width":
          if (bioEntity instanceof Reaction) {
            value = overlayEntry.getLineWidth();
          } else {
            continue;
          }
          break;
        case "type":
          if (overlayEntry instanceof GeneVariantDataOverlayEntry) {
            value = DataOverlayType.GENETIC_VARIANT;
          } else {
            value = DataOverlayType.GENERIC;
          }
          break;
        case "genevariations":
          if (overlayEntry instanceof GeneVariantDataOverlayEntry) {
            value = ((GeneVariantDataOverlayEntry) overlayEntry).getGeneVariants();
            Hibernate.initialize(((GeneVariantDataOverlayEntry) overlayEntry).getGeneVariants());
          } else {
            value = new Object[] {};
          }
          break;
        default:
          value = "Unknown column";
          break;
      }
      result.put(string, value);
    }
    return result;
  }

}
