package lcsb.mapviewer.api.projects;

import static org.junit.Assert.*;

import java.util.*;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;

import com.google.gson.Gson;

import lcsb.mapviewer.api.RestTestFunctions;
import lcsb.mapviewer.converter.zip.*;
import lcsb.mapviewer.model.*;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.persist.dao.ProjectDao;
import lcsb.mapviewer.persist.dao.user.UserDao;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IProjectService;

public class ProjectRestImplTest extends RestTestFunctions {
  Logger logger = LogManager.getLogger();

  @Autowired
  ProjectRestImpl _projectRest;

  ProjectRestImpl projectRest;

  @Autowired
  IModelService modelService;

  @Autowired
  IProjectService projectService;

  @Autowired
  ProjectDao projectDao;

  @Autowired
  UserDao userDao;

  @Before
  public void before() {
    projectRest = _projectRest;
  }

  @Test(expected = ObjectNotFoundException.class)
  public void testRemoveNonExistentProject() throws Exception {
    projectRest.removeProject("blabla", null);
  }

  @Test
  public void testRemoveProject() throws Exception {
    String projectId = "test";
    Project project = new Project();
    project.setProjectId(projectId);
    project.setOwner(userDao.getUserByLogin(ADMIN_BUILT_IN_LOGIN));
    projectDao.add(project);
    projectRest.removeProject(projectId, null);
  }

  @Test(expected = ObjectNotFoundException.class)
  public void testGetInvalidMetaData() throws Exception {
    projectRest.getProject("unknown_model_id");
  }

  @Test
  public void testGetProject() throws Exception {
    createMockProjectRest("testFiles/model/sample.xml");
    Map<String, Object> result = projectRest.getProject("sample");
    Gson gson = new Gson();
    assertNotNull(gson.toJson(result));
  }

  @Test
  public void testUpdateProject() throws Exception {
    createMockProjectRest("testFiles/model/sample.xml");
    Map<String, String> disease = new HashMap<>();
    disease.put("type", MiriamType.MESH_2012.name());
    disease.put("resource", "D010300");
    Map<String, Object> data = new HashMap<>();
    data.put("version", "1");
    data.put("name", "test");
    data.put("organism", null);
    data.put("disease", disease);
    data.put("projectId", "sample");
    projectRest.updateProject("sample", data);
  }

  @Test
  public void testComputePathForProject() throws Exception {
    String projectId = "Some_id";
    String directory1 = projectRest.computePathForProject(projectId, ".");
    Project project = new Project(projectId);
    project.setOwner(userDao.getUserByLogin(ADMIN_BUILT_IN_LOGIN));
    projectDao.add(project);
    projectDao.delete(project);
    String directory2 = projectRest.computePathForProject(projectId, ".");

    assertFalse(directory1.equals(directory2));
  }

  @Test
  public void testGetProjects() throws Exception {
    createMockProjectRest("testFiles/model/sample.xml");
    List<Map<String, Object>> result = projectRest.getProjects();
    Gson gson = new Gson();
    assertNotNull(gson.toJson(result));
  }

  @Test
  public void testGetStatistics() throws Exception {
    createMockProjectRest("testFiles/model/sample.xml");
    Map<String, Object> result = projectRest.getStatistics("sample");
    Gson gson = new Gson();
    assertNotNull(gson.toJson(result));

    Map<?, ?> elementAnnotations = (Map<?, ?>) result.get("elementAnnotations");
    assertEquals(elementAnnotations.get(MiriamType.CAS), 0);
    assertTrue((Integer) elementAnnotations.get(MiriamType.ENTREZ) > 0);

    Map<?, ?> reactionAnnotations = (Map<?, ?>) result.get("reactionAnnotations");
    assertEquals(reactionAnnotations.get(MiriamType.ENTREZ), 0);
    assertEquals(reactionAnnotations.get(MiriamType.PUBMED), 1);
  }

  @Test
  public void testGetMetaDataForComplexWithImages() throws Exception {
    createMockProjectRest("testFiles/model/complex_model_with_submaps.zip");
    Map<String, Object> result = projectRest.getProject("main");
    Gson gson = new Gson();
    assertNotNull(gson.toJson(result));
  }

  @Test
  public void testExtractZipEntries() throws Exception {
    MultiValueMap<String, Object> data = new LinkedMultiValueMap<>();
    data.put("zip-entries[0][_type]", createLinkedList("MAP"));
    data.put("zip-entries[0][_filename]", createLinkedList("main.xml"));
    data.put("zip-entries[0][_data][root]", createLinkedList("true"));
    data.put("zip-entries[0][_data][name]", createLinkedList("main"));
    data.put("zip-entries[0][_data][type][id]", createLinkedList("UNKNOWN"));
    data.put("zip-entries[0][_data][type][name]", createLinkedList("Unknown"));
    data.put("zip-entries[1][_type]", createLinkedList("OVERLAY"));
    data.put("zip-entries[1][_filename]", createLinkedList("layouts/goodschema.txt"));
    data.put("zip-entries[1][_data][name]", createLinkedList("example name"));
    data.put("zip-entries[1][_data][description]", createLinkedList("layout description"));
    data.put("zip-entries[2][_type]", createLinkedList("IMAGE"));
    data.put("zip-entries[2][_filename]", createLinkedList("images/test.png"));
    List<ZipEntryFile> result = projectRest.extractZipEntries(data);
    assertNotNull(result);
    assertEquals(3, result.size());
    assertTrue(result.get(0) instanceof ModelZipEntryFile);
    assertTrue(result.get(1) instanceof LayoutZipEntryFile);
    assertTrue(result.get(2) instanceof ImageZipEntryFile);

  }

  @Test(expected = QueryException.class)
  public void testExtractZipEntriesWithInvalidType() throws Exception {
    MultiValueMap<String, Object> data = new LinkedMultiValueMap<>();
    data.put("zip-entries[0][_type]", createLinkedList("MAPX"));
    data.put("zip-entries[0][_filename]", createLinkedList("main.xml"));
    projectRest.extractZipEntries(data);
  }

  @Test
  public void testExtractZipEntriesWithNoMapType() throws Exception {
    MultiValueMap<String, Object> data = new LinkedMultiValueMap<>();
    data.put("zip-entries[0][_type]", createLinkedList("MAP"));
    data.put("zip-entries[0][_filename]", createLinkedList("main.xml"));
    data.put("zip-entries[0][_data][root]", createLinkedList("true"));
    data.put("zip-entries[0][_data][name]", createLinkedList("main"));
    List<ZipEntryFile> result = projectRest.extractZipEntries(data);
    assertNotNull(result);
    assertEquals(1, result.size());
    assertTrue(result.get(0) instanceof ModelZipEntryFile);
  }

  private LinkedList<Object> createLinkedList(Object string) {
    LinkedList<Object> result = new LinkedList<>();
    result.add(string);
    return result;
  }

  private void createMockProjectRest(String string) throws Exception {
    Model model = null;
    Project project = null;
    model = super.getModelForFile(string, false);
    project = new Project(model.getName());
    project.addModel(model);
    project.setOwner(userDao.getUserByLogin(ADMIN_BUILT_IN_LOGIN));
    project.setProjectId(model.getName());
    ProjectLogEntry entry = new ProjectLogEntry();
    entry.setContent("content");
    entry.setType(ProjectLogEntryType.OTHER);
    entry.setSeverity("DEBUG");
    project.addLogEntry(entry);
    projectDao.add(project);
    projectRest = Mockito.spy(_projectRest);
  }

  @Test
  public void testGetLogs() throws Exception {
    createMockProjectRest("testFiles/model/sample.xml");
    Map<String, Object> result = projectRest.getLogs("sample", "", "0", 10, "id", "asc", "");
    assertNotNull(result);
    Gson gson = new Gson();
    assertNotNull(gson.toJson(result));
  }

}
