package lcsb.mapviewer.api.projects.models.bioEntities.elements;

import static org.junit.Assert.*;

import java.awt.geom.Rectangle2D;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.springframework.beans.factory.annotation.Autowired;

import com.fasterxml.jackson.databind.ObjectMapper;

import lcsb.mapviewer.annotation.services.annotators.PdbAnnotator;
import lcsb.mapviewer.api.RestTestFunctions;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.graphics.HorizontalAlign;
import lcsb.mapviewer.model.graphics.VerticalAlign;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.species.*;
import lcsb.mapviewer.model.map.species.field.*;
import lcsb.mapviewer.persist.dao.ProjectDao;

public class ElementRestImplTest extends RestTestFunctions {
  Logger logger = LogManager.getLogger();

  @Autowired
  ElementsRestImpl _elementsRestImpl;

  @Autowired
  ProjectDao projectDao;

  @Autowired
  PdbAnnotator pdbAnnotator;

  ObjectMapper mapper = new ObjectMapper();

  @AfterClass
  public static void tearDownAfterClass() throws Exception {
  }

  @Before
  public void setUp() throws Exception {
  }

  @After
  public void tearDown() throws Exception {
  }

  @Test
  public void testGetElementsProcessAllColumns() throws Exception {
    createMockElementRest("testFiles/model/sample.xml", true);
    List<Map<String, Object>> result = _elementsRestImpl.getElements("sample", "", "", "*", "", "", "");
    for (Map<String, Object> element : result) {
      for (String paramName : element.keySet()) {
        Object val = element.get(paramName);
        String paramValue = "";
        if (val != null) {
          paramValue = val.toString();
        }
        assertFalse("Improper handler for column name: " + paramName, paramValue.contains("Unknown column"));
      }
    }
    String json = mapper.writeValueAsString(result);
    assertNotNull(json);
  }

  @Test
  public void testGetElementsByType() throws Exception {
    String proteinType = new GenericProtein("1").getStringType();
    createMockElementRest("testFiles/model/sample.xml", false);
    List<Map<String, Object>> result = _elementsRestImpl.getElements("sample", "", "", "*", proteinType, "", "");
    assertEquals(12, result.size());
  }

  @Test
  public void testGetElementsVisibility() throws Exception {
    createMockElementRest("testFiles/model/sample.xml");
    List<Map<String, Object>> result = _elementsRestImpl.getElements("sample", "", "", "*", "", "", "");
    for (Map<String, Object> map : result) {
      assertTrue(map.get("hierarchyVisibilityLevel") instanceof String);
    }
  }

  @SuppressWarnings("unchecked")
  @Test
  public void testGetModificationsForProtein() throws Exception {
    StructuralState state = new StructuralState();
    state.setValue("asd");
    GenericProtein protein = new GenericProtein("s1");
    Residue mr = new Residue();
    mr.setState(ModificationState.ACETYLATED);
    mr.setName("S250");
    protein.addResidue(mr);
    protein.setStructuralState(state);
    Map<String, Object> result = _elementsRestImpl.getOthersForElement(protein);
    assertNotNull(result.get("modifications"));
    assertEquals(state.getValue(), result.get("structuralState"));
    List<Map<String, Object>> modifications = (List<Map<String, Object>>) result.get("modifications");
    assertEquals(1, modifications.size());
  }

  @SuppressWarnings("unchecked")
  @Test
  public void testGetModificationsForRna() throws Exception {
    Rna rna = new Rna("s1");
    ModificationSite mr = new ModificationSite();
    mr.setState(ModificationState.ACETYLATED);
    mr.setName("S250");
    rna.addModificationSite(mr);
    Map<String, Object> result = _elementsRestImpl.getOthersForElement(rna);
    assertNotNull(result.get("modifications"));
    assertEquals(null, result.get("structuralState"));
    List<Map<String, Object>> modifications = (List<Map<String, Object>>) result.get("modifications");
    assertEquals(1, modifications.size());
  }

  @SuppressWarnings("unchecked")
  @Test
  public void testGetModificationsForAntisenseRna() throws Exception {
    AntisenseRna antisenseRna = new AntisenseRna("s1");
    ModificationSite mr = new ModificationSite();
    mr.setName("S250");
    antisenseRna.addModificationSite(mr);
    Map<String, Object> result = _elementsRestImpl.getOthersForElement(antisenseRna);
    assertNotNull(result.get("modifications"));
    assertEquals(null, result.get("structuralState"));
    List<Map<String, Object>> modifications = (List<Map<String, Object>>) result.get("modifications");
    assertEquals(1, modifications.size());
  }

  private void createMockElementRest(String string) throws Exception {
     createMockElementRest(string, false);
  }

  private void createMockElementRest(String string, Boolean annotate) throws Exception {
    Model model = super.getModelForFile(string, false);
    if (annotate) {
      try {
        Protein protein = new GenericProtein("SNCA");
        protein.setElementId("SNCA");
        protein.setBorder(new Rectangle2D.Double(10, 20, 30, 40));
        protein.setZ(0);
        protein.setNameBorder(new Rectangle2D.Double(10, 20, 30, 40));
        protein.setNameHorizontalAlign(HorizontalAlign.CENTER);
        protein.setNameVerticalAlign(VerticalAlign.MIDDLE);
        pdbAnnotator.annotateElement(protein);
        model.addElement(protein);
      } catch (Exception e) {
      }
    }
    Project project = new Project(model.getName());
    project.setOwner(userService.getUserByLogin(ADMIN_BUILT_IN_LOGIN));
    project.addModel(model);
    projectDao.add(project);
  }
}
