package lcsb.mapviewer.services.impl;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.mail.MessagingException;

import org.apache.commons.io.IOUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.Hibernate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.commands.ColorModelCommand;
import lcsb.mapviewer.commands.CopyCommand;
import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.common.TextFileUtils;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.converter.ColorSchemaColumn;
import lcsb.mapviewer.converter.ColorSchemaReader;
import lcsb.mapviewer.converter.zip.ZipEntryFileFactory;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.layout.ReferenceGenomeType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.overlay.DataOverlayEntry;
import lcsb.mapviewer.model.overlay.DataOverlayType;
import lcsb.mapviewer.model.overlay.GeneVariant;
import lcsb.mapviewer.model.overlay.GeneVariantDataOverlayEntry;
import lcsb.mapviewer.model.overlay.GenericDataOverlayEntry;
import lcsb.mapviewer.model.overlay.InvalidDataOverlayException;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.DbUtils;
import lcsb.mapviewer.persist.dao.map.DataOverlayDao;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IConfigurationService;
import lcsb.mapviewer.services.interfaces.IDataOverlayService;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.utils.EmailSender;

@Transactional
@Service
public class DataOverlayService implements IDataOverlayService {

  private Logger logger = LogManager.getLogger();

  @Autowired
  private IUserService userService;

  @Autowired
  private IConfigurationService configurationService;

  @Autowired
  private DataOverlayDao dataOverlayDao;

  @Autowired
  private IModelService modelService;

  @Autowired
  private DbUtils dbUtils;

  /**
   * Object that sends emails.
   */
  private EmailSender emailSender;

  @Override
  public DataOverlay createDataOverlay(final CreateDataOverlayParams params)
      throws IOException, InvalidDataOverlayException {

    ColorSchemaReader reader = new ColorSchemaReader();
    Map<String, String> parameters = extractHeaderParameters(params.getColorInputStream(), params.getColorSchemaType());
    DataOverlayType colorSchemaType = DataOverlayType
        .valueOf(parameters.get(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_TYPE));
    final Collection<DataOverlayEntry> schemas = reader.readColorSchema(params.getColorInputStream(), parameters);

    dataOverlayDao.flush();

    DataOverlay overlay = new DataOverlay();
    overlay.setName(params.getName());
    overlay.setColorSchemaType(colorSchemaType);
    overlay.setPublic(false);
    overlay.setGoogleLicenseConsent(params.isGoogleLicenseConsent());
    UploadedFileEntry fileEntry = new UploadedFileEntry();
    fileEntry.setFileContent(IOUtils.toByteArray(params.getColorInputStream()));
    fileEntry.setOriginalFileName(params.getFileName());
    fileEntry.setLength(fileEntry.getFileContent().length);
    fileEntry.setOwner(params.getUser());
    overlay.setInputData(fileEntry);
    overlay.setDescription(params.getDescription());
    overlay.setProject(params.getProject());
    overlay.setCreator(params.getUser());
    overlay.addEntries(schemas);

    if (colorSchemaType == DataOverlayType.GENETIC_VARIANT) {
      String genomeVersion = parameters.get(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_GENOME_VERSION);
      ReferenceGenomeType genomeType = ReferenceGenomeType.UCSC;
      if (parameters.get(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_GENOME_TYPE) != null) {
        genomeType = ReferenceGenomeType
            .valueOf(parameters.get(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_GENOME_TYPE));
      }
      overlay.setGenomeType(genomeType);
      overlay.setGenomeVersion(genomeVersion);
    }
    dataOverlayDao.add(overlay);

    final String projectId = params.getProject().getProjectId();

    Thread sendEmailThread = new Thread(() -> {
      try {
        // open transaction for this thread
        dbUtils.createSessionForCurrentThread();

        final Map<DataOverlayEntry, Integer> matches = new HashMap<>();
        for (ModelData modelData : modelService.getModelsByProjectId(projectId)) {
          Model model = modelData.getModel();
          if (model == null) {
            model = new ModelFullIndexed(modelData);
          }
          Model copy = new CopyCommand(model).execute();
          ColorModelCommand command = new ColorModelCommand(copy, schemas,
              userService.getColorExtractorForUser(params.getUser()));
          command.execute();
          for (DataOverlayEntry entry : schemas) {
            if (matches.containsKey(entry)) {
              matches.put(entry, matches.get(entry) + command.getMatches().get(entry));
            } else {
              matches.put(entry, command.getMatches().get(entry));
            }
          }
        }

        sendSuccessfullGenerationEmail(params, matches);
      } catch (MessagingException e) {
        logger.error("Problem with sending email", e);
      } catch (ObjectNotFoundException e) {
        logger.error("Problem with sending email", e);
      } finally {
        dbUtils.closeSessionForCurrentThread();
      }
    });

    logger.info("Data overlay " + overlay.getId() + " created successfully");
    sendEmailThread.start();
    return overlay;
  }

  protected Map<String, String> extractHeaderParameters(final InputStream colorInputStream, DataOverlayType colorSchemaType)
      throws IOException {
    Map<String, String> parameters = TextFileUtils.getHeaderParametersFromFile(colorInputStream);
    if (colorSchemaType == null) {
      String type = parameters.get(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_TYPE);
      if (type != null) {
        colorSchemaType = DataOverlayType.valueOf(type);
      }
    }
    if (colorSchemaType == null) {
      colorSchemaType = DataOverlayType.GENERIC;
    }

    if (parameters.get(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_TYPE) == null) {
      parameters.put(ZipEntryFileFactory.LAYOUT_HEADER_PARAM_TYPE, colorSchemaType.name());
    }
    return parameters;
  }

  /**
   * Sends notification email that data overlay was generated.
   *
   * @param params
   *          list of {@link IDataOverlayService.CreateDataOverlayParams params}
   *          used for data overlay creation
   * @param matches
   *          set of schemas used in coloring
   * @throws MessagingException
   *           thrown when there is a problem with sending email
   */
  protected void sendSuccessfullGenerationEmail(final CreateDataOverlayParams params,
      final Map<DataOverlayEntry, Integer> matches)
      throws MessagingException {
    StringBuilder content = new StringBuilder("Project: " + params.getProject().getProjectId() + ".\n");
    content.append("Data overlay " + params.getName() + " generated successfully.\n");
    content.append("Result of coloring:<br/>");

    String table = prepareTableResult(matches, new ColorSchemaReader());
    content.append(table);
    String email = params.getProject().getNotifyEmail();
    getEmailSender().sendEmail("Minerva notification", content.toString(), params.getUser());

    // send email to the model owner
    if (email != null && !email.equals(params.getUser().getEmail())) {
      content = new StringBuilder();
      String username = params.getUser().getLogin();
      content.append("User ")
          .append(username)
          .append(" created data overlay in ")
          .append(params.getProject().getProjectId())
          .append(" map.");

      getEmailSender().sendEmail("Minerva notification", content.toString(), email);
    }
  }

  /**
   * Prepares table with statistics about coloring.
   *
   * @param matches
   *          schemas that were used for coloring
   * @param scr
   *          interface that returns list of columns that should be printed
   * @return table with statistics about coloring
   */
  protected String prepareTableResult(final Map<DataOverlayEntry, Integer> matches, final ColorSchemaReader scr) {
    StringBuilder sb = new StringBuilder();

    Collection<ColorSchemaColumn> columns = scr.getSetColorSchemaColumns(matches.keySet());

    for (ColorSchemaColumn column : ColorSchemaColumn.values()) {
      if (columns.contains(column)) {
        sb.append(column.getColumnName())
            .append("\t");
      }
    }
    sb.append("matches<br/>\n");

    for (DataOverlayEntry originalSchema : matches.keySet()) {
      for (DataOverlayEntry schema : splitColorSchema(originalSchema)) {
        sb.append(prepareTableRow(columns, schema, matches.get(originalSchema)));
      }
    }

    return sb.toString();
  }

  /**
   * {@link DataOverlayEntry} sometimes contains merged value from few rows.
   * This method split single {@link DataOverlayEntry} object to simple flat
   * objects that can be serialized in a simple tab separated file.
   *
   * @param originalSchema
   *          original {@link DataOverlayEntry} object that we want to split
   *          into flat objects
   * @return {@link List} of flat schema objects obtained from input
   */
  private List<DataOverlayEntry> splitColorSchema(final DataOverlayEntry originalSchema) {
    List<DataOverlayEntry> result = new ArrayList<>();
    if (originalSchema instanceof GenericDataOverlayEntry) {
      result.add(originalSchema);
    } else if (originalSchema instanceof GeneVariantDataOverlayEntry) {
      for (GeneVariant gv : ((GeneVariantDataOverlayEntry) originalSchema).getGeneVariants()) {
        GeneVariantDataOverlayEntry copy = (GeneVariantDataOverlayEntry) originalSchema.copy();
        copy.getGeneVariants().clear();
        copy.addGeneVariant(gv.copy());
        result.add(copy);
      }
    } else {
      throw new InvalidArgumentException("Unknown class type: " + originalSchema.getClass());
    }
    return result;
  }

  /**
   * Prepares tab separated {@link String} representation of
   * {@link DataOverlayEntry}.
   *
   * @param columns
   *          columns that should be output in the result String
   * @param schema
   *          input DataOverlayEntry that is going to be transformed into String
   *          representation
   * @return tab separated {@link String} representation of
   *         {@link DataOverlayEntry}
   */
  protected String prepareTableRow(final Collection<ColorSchemaColumn> columns, final DataOverlayEntry schema, final int match) {
    StringBuilder sb = new StringBuilder();
    for (ColorSchemaColumn column : ColorSchemaColumn.values()) {
      if (columns.contains(column)) {
        if (schema instanceof GenericDataOverlayEntry) {
          sb.append(prepareTableCellForGenericSchema((GenericDataOverlayEntry) schema, column));
        } else if (schema instanceof GeneVariantDataOverlayEntry) {
          sb.append(prepareTableCellForGeneVariationSchema((GeneVariantDataOverlayEntry) schema, column));
        } else {
          throw new InvalidArgumentException("Unknown schema type: " + schema.getClass());
        }
      }
    }
    sb.append(match + "").append("<br/>\n");
    return sb.toString();
  }

  /**
   * Returns String representing data of {@link GenericDataOverlayEntry} that
   * should appear in a given {@link ColorSchemaColumn}.
   *
   * @param schema
   *          object for which data will be returned
   * @param column
   *          column for which data should be returned
   * @return {@link String} representing data of {@link GenericDataOverlayEntry}
   *         that should appear in a given {@link ColorSchemaColumn}
   */
  protected String prepareTableCellForGenericSchema(final GenericDataOverlayEntry schema, final ColorSchemaColumn column) {
    StringBuilder sb = new StringBuilder();
    if (column.equals(ColorSchemaColumn.COLOR)) {
      if (schema.getColor() != null) {
        sb.append("#" + Integer.toHexString(schema.getColor().getRGB()).substring(2).toUpperCase());
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.NAME)) {
      sb.append(schema.getName() + "\t");
    } else if (column.equals(ColorSchemaColumn.MAP_NAME)) {
      sb.append(schema.getModelName() + "\t");
    } else if (column.equals(ColorSchemaColumn.VALUE)) {
      sb.append(schema.getValue() + "\t");
    } else if (column.equals(ColorSchemaColumn.COMPARTMENT)) {
      for (String str : schema.getCompartments()) {
        sb.append(str + ", ");
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.TYPE)) {
      for (Class<? extends BioEntity> str : schema.getTypes()) {
        sb.append(str.getSimpleName() + ", ");
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.IDENTIFIER)) {
      for (MiriamData md : schema.getMiriamData()) {
        sb.append(md.getDataType().getCommonName() + ": " + md.getResource() + ", ");
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.ELEMENT_IDENTIFIER)) {
      sb.append(schema.getElementId() + "\t");
    } else if (column.equals(ColorSchemaColumn.LINE_WIDTH)) {
      sb.append(schema.getLineWidth() + "\t");
    } else if (column.equals(ColorSchemaColumn.REVERSE_REACTION)) {
      sb.append(schema.getReverseReaction() + "\t");
    } else if (column.equals(ColorSchemaColumn.POSITION)) {
      sb.append(schema.getReverseReaction() + "\t");
    } else if (column.equals(ColorSchemaColumn.DESCRIPTION)) {
      sb.append(schema.getDescription() + "\t");
    } else {
      throw new InvalidArgumentException("Unknown column type: " + column + " for schema type: " + schema.getClass());
    }
    return sb.toString();
  }

  /**
   * Returns String representing data of {@link GeneVariantDataOverlayEntry}
   * that should appear in a given {@link ColorSchemaColumn}.
   *
   * @param schema
   *          object for which data will be returned
   * @param column
   *          column for which data should be returned
   * @return {@link String} representing data of
   *         {@link GeneVariantDataOverlayEntry} that should appear in a given
   *         {@link ColorSchemaColumn}
   */
  protected String prepareTableCellForGeneVariationSchema(final GeneVariantDataOverlayEntry schema,
      final ColorSchemaColumn column) {
    GeneVariant gv = schema.getGeneVariants().iterator().next();
    StringBuilder sb = new StringBuilder();
    if (column.equals(ColorSchemaColumn.COLOR)) {
      if (schema.getColor() != null) {
        sb.append("#" + Integer.toHexString(schema.getColor().getRGB()).substring(2).toUpperCase());
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.NAME)) {
      sb.append(schema.getName() + "\t");
    } else if (column.equals(ColorSchemaColumn.MAP_NAME)) {
      sb.append(schema.getModelName() + "\t");
    } else if (column.equals(ColorSchemaColumn.VALUE)) {
      sb.append(schema.getValue() + "\t");
    } else if (column.equals(ColorSchemaColumn.COMPARTMENT)) {
      for (String str : schema.getCompartments()) {
        sb.append(str + ", ");
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.TYPE)) {
      for (Class<? extends BioEntity> str : schema.getTypes()) {
        sb.append(str.getSimpleName() + ", ");
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.IDENTIFIER)) {
      for (MiriamData md : schema.getMiriamData()) {
        sb.append(md.getDataType().getCommonName() + ": " + md.getResource() + ", ");
      }
      sb.append("\t");
    } else if (column.equals(ColorSchemaColumn.ELEMENT_IDENTIFIER)) {
      sb.append(schema.getElementId() + "\t");
    } else if (column.equals(ColorSchemaColumn.LINE_WIDTH)) {
      sb.append(schema.getLineWidth() + "\t");
    } else if (column.equals(ColorSchemaColumn.REVERSE_REACTION)) {
      sb.append(schema.getReverseReaction() + "\t");
    } else if (column.equals(ColorSchemaColumn.POSITION)) {
      sb.append(gv.getPosition() + "\t");
    } else if (column.equals(ColorSchemaColumn.DESCRIPTION)) {
      sb.append(schema.getDescription() + "\t");
    } else if (column.equals(ColorSchemaColumn.ORIGINAL_DNA)) {
      sb.append(gv.getOriginalDna() + "\t");
    } else if (column.equals(ColorSchemaColumn.ALTERNATIVE_DNA)) {
      sb.append(gv.getModifiedDna() + "\t");
    } else if (column.equals(ColorSchemaColumn.CONTIG)) {
      sb.append(gv.getContig() + "\t");
    } else {
      throw new InvalidArgumentException("Unknown column type: " + column + " for schema type: " + schema.getClass());
    }
    return sb.toString();
  }

  public EmailSender getEmailSender() {
    if (emailSender == null) {
      emailSender = new EmailSender(configurationService);
    }
    return emailSender;
  }

  public void setEmailSender(final EmailSender emailSender) {
    this.emailSender = emailSender;
  }

  @Override
  public List<Pair<Element, DataOverlayEntry>> getElementReferencesForDataOverlay(final int overlayId, final List<ModelData> models) {
    return dataOverlayDao.getElementReferencesForModels(overlayId, models);
  }

  @Override
  public List<Pair<Reaction, DataOverlayEntry>> getReactionReferencesForDataOverlay(final int overlayId, final List<ModelData> models) {
    return dataOverlayDao.getReactionReferencesForModels(overlayId, models);
  }

  private void getOverlayByIdNotNull(final String projectId, final int overlayId) throws ObjectNotFoundException {
    DataOverlay overlay = dataOverlayDao.getById(overlayId);
    if (overlay == null || !overlay.getProject().getProjectId().equals(projectId)) {
      throw new ObjectNotFoundException("Overlay does not exist");
    }
  }

  @Override
  public List<Pair<? extends BioEntity, DataOverlayEntry>> getBioEntitiesForDataOverlay(final String projectId, final int mapId,
      final Integer overlayId) throws QueryException {
    ModelData map = modelService.getModelByMapId(projectId, mapId);
    getOverlayByIdNotNull(projectId, overlayId);

    List<Pair<? extends BioEntity, DataOverlayEntry>> result = new ArrayList<>();
    result.addAll(dataOverlayDao.getReactionsForModels(overlayId, Arrays.asList(map)));
    result.addAll(dataOverlayDao.getElementsForModels(overlayId, Arrays.asList(map)));
    return result;
  }

  @Override
  public List<Pair<Element, DataOverlayEntry>> getFullElementForDataOverlay(final String projectId, final int mapId, final Integer id,
      final int overlayId) throws QueryException {
    ModelData map = modelService.getModelByMapId(projectId, mapId);
    getOverlayByIdNotNull(projectId, overlayId);

    return dataOverlayDao.getElementsForModels(overlayId, Arrays.asList(map), Arrays.asList(id));
  }

  @Override
  public List<Pair<Reaction, DataOverlayEntry>> getFullReactionForDataOverlay(final String projectId, final int mapId, final Integer id,
      final int overlayId) throws QueryException {
    ModelData map = modelService.getModelByMapId(projectId, mapId);
    getOverlayByIdNotNull(projectId, overlayId);

    return dataOverlayDao.getReactionsForModels(overlayId, Arrays.asList(map), Arrays.asList(id));
  }

  @Override
  public DataOverlay getDataOverlayById(final Integer overlayId) {
    return dataOverlayDao.getById(overlayId);
  }

  @Override
  public List<DataOverlay> getDataOverlaysByProject(final Project project) {
    return dataOverlayDao.getByProject(project);
  }

  @Override
  public void removeDataOverlay(final DataOverlay dataOverlay) {
    dataOverlayDao.delete(dataOverlay);
  }

  @Override
  public void updateDataOverlay(final DataOverlay dataOverlay) {
    dataOverlayDao.update(dataOverlay);
  }

  @Override
  public DataOverlay getDataOverlayById(final String projectId, final Integer overlayId) throws ObjectNotFoundException {
    DataOverlay result = getDataOverlayById(overlayId);
    if (result != null && result.getProject().getProjectId().equals(projectId)) {
      return result;
    }
    throw new ObjectNotFoundException("Overlay with given id does not exist: " + overlayId);
  }

  @Override
  public User getOverlayCreator(final Integer overlayId) {
    DataOverlay overlay = dataOverlayDao.getById(overlayId);
    if (overlay != null) {
      Hibernate.initialize(overlay.getCreator());
      return overlay.getCreator();
    }
    return null;
  }

}
