package lcsb.mapviewer.services.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.Hibernate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.services.MiriamConnector;
import lcsb.mapviewer.annotation.services.PubmedParser;
import lcsb.mapviewer.annotation.services.PubmedSearchException;
import lcsb.mapviewer.common.IProgressUpdater;
import lcsb.mapviewer.model.Article;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.kinetics.SbmlParameter;
import lcsb.mapviewer.model.map.kinetics.SbmlUnit;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.persist.DbUtils;
import lcsb.mapviewer.persist.dao.ProjectDao;
import lcsb.mapviewer.persist.dao.map.ModelDao;
import lcsb.mapviewer.persist.dao.map.kinetics.SbmlParameterDao;
import lcsb.mapviewer.persist.dao.map.kinetics.SbmlUnitDao;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IModelService;

/**
 * Implementation of the service that manages models.
 * 
 * @author Piotr Gawron
 * 
 */
@Transactional
@Service
public class ModelService implements IModelService {

  /**
   * Default class logger.
   */
  private Logger logger = LogManager.getLogger();

  /**
   * Data access object for models.
   */
  private ModelDao modelDao;

  private ProjectDao projectDao;

  /**
   * Local backend to the pubmed data.
   */
  private PubmedParser backend;

  /**
   * Connector used for accessing data from miriam registry.
   */
  private MiriamConnector miriamConnector;

  private SbmlParameterDao sbmlParameterDao;
  private SbmlUnitDao sbmlUnitDao;

  private DbUtils dbUtils;

  @Autowired
  public ModelService(ModelDao modelDao,
      ProjectDao projectDao,
      PubmedParser backend,
      MiriamConnector miriamConnector,
      SbmlParameterDao sbmlParameterDao,
      SbmlUnitDao sbmlUnitDao,
      DbUtils dbUtils) {
    this.modelDao = modelDao;
    this.projectDao = projectDao;
    this.backend = backend;
    this.miriamConnector = miriamConnector;
    this.sbmlParameterDao = sbmlParameterDao;
    this.sbmlUnitDao = sbmlUnitDao;
    this.dbUtils = dbUtils;
  }

  @Override
  public void cacheAllPubmedIds(Model model, IProgressUpdater updater) {
    logger.debug("Caching pubmed ids...");
    if (model != null) {
      Set<Integer> pubmedIds = new HashSet<>();

      for (Element element : model.getElements()) {
        for (MiriamData md : element.getMiriamData()) {
          if (MiriamType.PUBMED.equals(md.getDataType())) {
            try {
              pubmedIds.add(Integer.parseInt(md.getResource()));
            } catch (NumberFormatException e) {
              logger.error("Problem with parsing: " + e.getMessage(), e);
            }
          }
        }
      }
      for (Reaction reaction : model.getReactions()) {
        for (MiriamData md : reaction.getMiriamData()) {
          if (MiriamType.PUBMED.equals(md.getDataType())) {
            try {
              pubmedIds.add(Integer.parseInt(md.getResource()));
            } catch (NumberFormatException e) {
              logger.error("Problem with parsing: " + e.getMessage(), e);
            }
          }
        }
      }
      double amount = pubmedIds.size();
      double counter = 0;
      for (Integer id : pubmedIds) {
        try {
          Article art = backend.getPubmedArticleById(id);
          if (art == null) {
            logger.warn("Cannot find pubmed article. Pubmed_id = " + id);
          }
        } catch (PubmedSearchException e) {
          logger.warn("Problem with accessing info about pubmed: " + id, e);
        }
        counter++;
        updater.setProgress(IProgressUpdater.MAX_PROGRESS * counter / amount);
      }
    }
    logger.debug("Caching finished");
  }

  @Override
  public void cacheAllMiriamLinks(Model model, IProgressUpdater updater) {
    logger.debug("Caching miriam ids...");
    if (model != null) {
      Set<MiriamData> pubmedIds = new HashSet<MiriamData>();

      for (Element element : model.getElements()) {
        pubmedIds.addAll(element.getMiriamData());
      }
      for (Reaction reaction : model.getReactions()) {
        pubmedIds.addAll(reaction.getMiriamData());
      }
      double amount = pubmedIds.size();
      double counter = 0;
      for (MiriamData md : pubmedIds) {
        miriamConnector.getUrlString(md);
        counter++;
        updater.setProgress(IProgressUpdater.MAX_PROGRESS * counter / amount);
      }
    }
    logger.debug("Caching finished");

  }

  @Override
  public void updateModel(ModelData model) {
    modelDao.update(model);

  }

  /**
   * @return the projectDao
   * @see #projectDao
   */
  public ProjectDao getProjectDao() {
    return projectDao;
  }

  /**
   * @param projectDao
   *          the projectDao to set
   * @see #projectDao
   */
  public void setProjectDao(ProjectDao projectDao) {
    this.projectDao = projectDao;
  }

  @Override
  public List<ModelData> getModelsByProjectId(String projectId) throws ObjectNotFoundException {
    if (projectDao.getProjectByProjectId(projectId) == null) {
      throw new ObjectNotFoundException("Project with given id does not exist: " + projectId);
    }
    return new ArrayList<>(modelDao.getByProjectId(projectId));
  }

  @Override
  public List<ModelData> getModelsByMapId(String projectId, String mapId) throws QueryException {
    if (Objects.equals("*", mapId)) {
      return getModelsByProjectId(projectId);
    }
    if (!StringUtils.isNumeric(mapId)) {
      throw new QueryException("Invalid map id: " + mapId);
    }
    ModelData map = getModelByMapId(projectId, Integer.valueOf(mapId));
    return Arrays.asList(map);
  }

  @Override
  public ModelData getModelByMapId(String projectId, int mapId) throws QueryException {
    ModelData map = modelDao.getById(projectId, mapId);
    if (map == null) {
      throw new ObjectNotFoundException("Map with given id does not exist: " + mapId);
    }
    return map;
  }

  @Override
  public Model getAndFetchModelByMapId(String projectId, int mapId) throws QueryException {
    ModelData map = getModelByMapId(projectId, mapId);

    Hibernate.initialize(map.getElements());
    Hibernate.initialize(map.getReactions());
    Hibernate.initialize(map.getProject());
    Hibernate.initialize(map.getFunctions());
    Hibernate.initialize(map.getUnits());
    Hibernate.initialize(map.getLayers());

    map.getSubmodels().clear();
    map.getParentModels().clear();

    for (Element element : map.getElements()) {
      if (element.getGlyph() != null) {
        Hibernate.initialize(element.getGlyph().getFile());
      }
      if (element instanceof Complex) {
        ((Complex) element).setElemets(new ArrayList<>());
      }
      if (element instanceof Compartment) {
        ((Compartment) element).setElements(new HashSet<>());
      }
      element.setSubmodel(null);
    }
    for (Element element : map.getElements()) {
      if (element instanceof Species) {
        if (((Species) element).getComplex() != null) {
          ((Species) element).getComplex().addSpecies((Species) element);
        }
      }
      if (element.getCompartment() != null) {
        element.getCompartment().addElement(element);
      }
    }
    dbUtils.getSessionForCurrentThread().clear();

    return new ModelFullIndexed(map);
  }

  @Override
  public List<SbmlParameter> getParametersForModel(String projectId, String modelId) throws QueryException {
    return sbmlParameterDao.getByMaps(getModelsByMapId(projectId, modelId));
  }

  @Override
  public int getPublicationCount(String projectId, String mapId) throws QueryException {
    return modelDao.getAnnotationCount(getModelsByMapId(projectId, mapId), MiriamType.PUBMED);
  }

  @Override
  public List<SbmlUnit> getUnitsForModel(String projectId, String modelId) throws QueryException {
    return sbmlUnitDao.getByMaps(getModelsByMapId(projectId, modelId));
  }

}
