package lcsb.mapviewer.services.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.mail.MessagingException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.Hibernate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.converter.graphics.MapGenerator;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.layout.ProjectBackground;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.DbUtils;
import lcsb.mapviewer.persist.dao.map.ProjectBackgroundDao;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.interfaces.*;
import lcsb.mapviewer.services.utils.EmailSender;

@Transactional
@Service
public class ProjectBackgroundService implements IProjectBackgroundService {

  private Logger logger = LogManager.getLogger();

  private ProjectBackgroundDao projectBackgroundDao;

  private IUserService userService;

  private IConfigurationService configurationService;

  private EmailSender emailSender;

  @Autowired
  public ProjectBackgroundService(ProjectBackgroundDao projectBackgroundDao,
      IUserService userService,
      IConfigurationService configurationService,
      DbUtils dbUtils) {
    this.projectBackgroundDao = projectBackgroundDao;
    this.userService = userService;
    this.configurationService = configurationService;
  }

  @Override
  public void removeProjectBackground(ProjectBackground background, final String homeDir) {
    final String dir;
    if (homeDir != null) {
      if (background.getProject().getDirectory() != null) {
        dir = homeDir + "/../map_images/" + background.getProject().getDirectory() + "/";
      } else {
        dir = homeDir + "/../map_images/";
      }
    } else {
      dir = null;
    }

    String projectId = background.getProject().getProjectId();
    final String email;
    User user = background.getCreator();
    if (user != null) {
      email = user.getEmail();
    } else {
      email = null;
    }

    background.getProject().removeProjectBackground(background);
    projectBackgroundDao.delete(background);

    logger.info("Data overlay " + background.getId() + " removed");

    if (email != null) {
      Thread sendEmailThread = new Thread(() -> {
        try {
          sendSuccesfullRemoveEmail(projectId, background.getName(), email);
        } catch (MessagingException e) {
          logger.error(e);
        }
      });
      sendEmailThread.start();
    }
    Thread removeFilesThread = new Thread(() -> {
      MapGenerator generator = new MapGenerator();
      try {
        generator.removeProjectBackground(background, dir);
      } catch (IOException e) {
        logger.error(e);
      }

    });
    removeFilesThread.start();
  }

  @Override
  public void updateProjectBackground(ProjectBackground projectBackground) {
    projectBackgroundDao.update(projectBackground);
  }

  @Override
  public EmailSender getEmailSender() {
    if (emailSender == null) {
      emailSender = new EmailSender(configurationService);
    }
    return emailSender;
  }

  @Override
  public void setEmailSender(EmailSender emailSender) {
    this.emailSender = emailSender;
  }

  @Override
  public List<ProjectBackground> getProjectBackgroundsByProject(Project project) {
    List<ProjectBackground> result = new ArrayList<>();
    List<ProjectBackground> backgrounds = projectBackgroundDao.getProjectBackgroundsByProject(project);
    for (ProjectBackground projectBackground : backgrounds) {
      Hibernate.initialize(projectBackground.getProjectBackgroundImageLayer());
    }
    result.addAll(backgrounds);
    result.sort(ProjectBackground.ID_COMPARATOR);
    return result;
  }

  @Override
  public ProjectBackground getProjectBackgroundById(Integer backgroundId) {
    return projectBackgroundDao.getById(backgroundId);
  }

  /**
   * @return the configurationService
   * @see #configurationService
   */
  public IConfigurationService getConfigurationService() {
    return configurationService;
  }

  /**
   * @param configurationService
   *          the configurationService to set
   * @see #configurationService
   */
  public void setConfigurationService(IConfigurationService configurationService) {
    this.configurationService = configurationService;
  }

  /**
   * @return the userService
   * @see #userService
   */
  public IUserService getUserService() {
    return userService;
  }

  /**
   * @param userService
   *          the userService to set
   * @see #userService
   */
  public void setUserService(IUserService userService) {
    this.userService = userService;
  }

  /**
   * Sends notification email that background was removed.
   *
   * @param projectId
   *          identifier of the project
   * @param backgroundName
   *          name of the background
   * @param email
   *          notification email
   * @throws MessagingException
   *           thrown when there is a problem with sending email
   */
  protected void sendSuccesfullRemoveEmail(String projectId, String backgroundName, String email)
      throws MessagingException {
    getEmailSender().sendEmail("Minerva notification", "Project background " + backgroundName +
        " in map " + projectId + " was successfully removed.<br/>", email);
  }

  @Override
  public List<ProjectBackground> getProjectBackgroundsByUser(User user) {
    return projectBackgroundDao.getProjectBackgroundsByUser(user);
  }

  @Override
  public ProjectBackground getProjectBackgroundById(String projectId, Integer backgroundId)
      throws ObjectNotFoundException {
    ProjectBackground result = getProjectBackgroundById(backgroundId);
    if (result != null && result.getProject().getProjectId().equals(projectId)) {
      return result;
    }
    throw new ObjectNotFoundException("Overlay with given id does not exist: " + backgroundId);
  }

  @Override
  public List<ProjectBackground> getProjectBackgroundsByProject(String projectId) {
    return projectBackgroundDao.getProjectBackgroundsByProject(projectId);
  }

}
