package lcsb.mapviewer.services.impl;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.hibernate.Hibernate;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.reaction.ReactionNode;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Degraded;
import lcsb.mapviewer.model.map.species.Drug;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Ion;
import lcsb.mapviewer.model.map.species.Phenotype;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.SimpleMolecule;
import lcsb.mapviewer.model.map.species.Unknown;
import lcsb.mapviewer.modelutils.map.ElementUtils;
import lcsb.mapviewer.persist.dao.map.ReactionDao;
import lcsb.mapviewer.persist.dao.map.ReactionProperty;
import lcsb.mapviewer.persist.dao.map.species.ElementDao;
import lcsb.mapviewer.persist.dao.map.species.ElementProperty;
import lcsb.mapviewer.services.QueryException;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.ISearchService;

/**
 * This is implementation of the service that returns information about species,
 * paths etc in the model.
 *
 *
 * @author Piotr Gawron
 *
 */
@Transactional
@Service
public class SearchService implements ISearchService {

  /**
   * Default class logger.
   */
  @SuppressWarnings("unused")
  private static Logger logger = LogManager.getLogger();

  private IModelService modelService;

  private ElementDao elementDao;

  private ReactionDao reactionDao;

  /**
   * This object maps short string prefixes into class of elements on the map
   * that will be searched. For instance prefix "complex" might refer to
   * {@link Complex} class and and example query would look like: "complex:alpha
   * subcomplex".
   */
  private Map<String, List<Class<? extends BioEntity>>> speciesSearchReversePrefix = new HashMap<>();

  /**
   * Default constructor that set list of prefixes used in the search queries.
   */
  public SearchService(IModelService modelService, ElementDao elementDao, ReactionDao reactionDao) {
    this.modelService = modelService;
    this.elementDao = elementDao;
    this.reactionDao = reactionDao;
    addSearchPrefix("complex", Arrays.asList(Complex.class));
    addSearchPrefix("degrded", Arrays.asList(Degraded.class));
    addSearchPrefix("drug", Arrays.asList(Drug.class));
    addSearchPrefix("gene", Arrays.asList(Gene.class));
    addSearchPrefix("ion", Arrays.asList(Ion.class));
    addSearchPrefix("phenotype", Arrays.asList(Phenotype.class));
    addSearchPrefix("protein", new ArrayList<>(new ElementUtils().getClassesByStringTypes("Protein")));
    addSearchPrefix("rna", Arrays.asList(Rna.class));
    addSearchPrefix("molecule", Arrays.asList(SimpleMolecule.class));
    addSearchPrefix("unknown", Arrays.asList(Unknown.class));
    addSearchPrefix("element", new ArrayList<>(new ElementUtils().getAvailableElementSubclasses()));
    addSearchPrefix("reaction", new ArrayList<>(new ElementUtils().getAvailableReactionSubclasses()));
  }

  /**
   * Adds search prefix for an element class.
   *
   * @param prefix
   *          string prefix used in search queries
   * @param clazz
   *          class for which this prefix stands
   */
  private void addSearchPrefix(String prefix, List<Class<? extends BioEntity>> clazz) {
    speciesSearchReversePrefix.put(prefix, clazz);
  };

  /**
   * Transform {@link Reaction} into set of result entries.
   *
   * @param reaction
   *          reaction to be transformed
   * @return set of result entries for all element of the {@link Reaction}
   */
  private List<BioEntity> reactionToResultList(Reaction reaction) {
    List<BioEntity> result = new ArrayList<>();
    result.add(reaction);
    for (ReactionNode node : reaction.getReactionNodes()) {
      result.add(node.getElement());
    }
    return result;
  }

  @Override
  public List<BioEntity> searchByQuery(String projectId, String mapId, String query, int limit, boolean perfectMatch,
      String ipAddress) throws QueryException {
    query = query.toLowerCase();
    if (query == null) {
      throw new InvalidArgumentException("Invalid query: null");
    }

    List<ModelData> maps = modelService.getModelsByMapId(projectId, mapId);

    MiriamData mt = getMiriamTypeForQuery(query);
    List<BioEntity> result = new ArrayList<>();

    Pageable pageable = PageRequest.of(0, limit);
    Map<ReactionProperty, List<? extends Object>> reactionFilter = new HashMap<>();
    reactionFilter.put(ReactionProperty.MAP, maps);

    Map<ElementProperty, List<? extends Object>> elementFilter = new HashMap<>();
    elementFilter.put(ElementProperty.MAP, maps);

    if (mt != null) {
      elementFilter.put(ElementProperty.ANNOTATION, Arrays.asList(mt));

      result.addAll(elementDao.getByFilter(pageable, elementFilter).getContent());

      reactionFilter.put(ReactionProperty.ANNOTATION, Arrays.asList(mt));

      for (Reaction reaction : reactionDao.getByFilter(pageable, reactionFilter)) {
        result.addAll(reactionToResultList(reaction));
      }
    } else {
      String indexQuery = getQueryStringForIndex(query);
      List<Class<? extends BioEntity>> type = getTypeForQuery(query);
      if (perfectMatch) {
        elementFilter.put(ElementProperty.PERFECT_TEXT, Arrays.asList(indexQuery));
        reactionFilter.put(ReactionProperty.PERFECT_TEXT, Arrays.asList(indexQuery));
      } else {
        elementFilter.put(ElementProperty.TEXT, Arrays.asList(indexQuery));
        reactionFilter.put(ReactionProperty.TEXT, Arrays.asList(indexQuery));
      }
      elementFilter.put(ElementProperty.CLASS, type);
      reactionFilter.put(ReactionProperty.CLASS, type);

      result.addAll(elementDao.getByFilter(pageable, elementFilter).getContent());

      for (Reaction reaction : reactionDao.getByFilter(pageable, reactionFilter)) {
        result.addAll(reactionToResultList(reaction));
      }
    }
    return result;
  }

  /**
   * This method transform a string into indexed query. AvailablePrefixes
   * contains a collection of all possible prefixes for query string that should
   * be omitted.
   * 
   * @param originalQuery
   *          query to be transformed
   * @param availablePrefixes
   *          collection of valid prefixes in the query
   * 
   * @return indexed string of the original query
   */
  private String getQueryStringForIndex(String originalQuery) {
    String result = originalQuery;
    for (String string : speciesSearchReversePrefix.keySet()) {
      if (result.startsWith(string + ":")) {
        result = result.replaceFirst(string + ":", "");
        break;
      }
    }

    return getIndexStringForString(result);
  }

  private String getIndexStringForString(String str) {
    return str.toLowerCase().replaceAll("[^a-z0-9]", "");
  }

  /**
   * Return type of elements that can be returned for the query.
   * 
   * @param string
   *          query string.
   * @param speciesSearchReversePrefix
   *          map containing information about possible prefixes used in the
   *          query
   * @return class of the element that might exist in the result
   */
  private List<Class<? extends BioEntity>> getTypeForQuery(String string) {
    String type = string.split(":")[0];
    List<Class<? extends BioEntity>> result = speciesSearchReversePrefix.get(type);
    if (result == null) {
      result = new ArrayList<>();
      result.addAll(new ElementUtils().getAvailableElementSubclasses());
    }
    return result;
  }

  @Override
  public List<BioEntity> searchByQuery(String projectId, String mapId, String query, int limit, boolean perfectMatch)
      throws QueryException {
    return searchByQuery(projectId, mapId, query, limit, perfectMatch, null);
  }

  @Override
  public List<BioEntity> getClosestElements(String projectId, int mapId, Point2D point, int numberOfElements,
      boolean perfectHit,
      Collection<String> types) throws QueryException {
    List<BioEntity> result = new ArrayList<>();

    List<DistanceToObject> tmpList = new ArrayList<>();
    ModelData map = modelService.getModelByMapId(projectId, mapId);
    List<Class<? extends BioEntity>> classes = new ElementUtils().getClassesByStringTypes(types);

    for (Pair<Reaction, Double> pair : reactionDao.getByCoordinatesWithDistance(map, point, numberOfElements, classes)) {
      tmpList.add(new DistanceToObject(pair));
    }
    for (Element element : elementDao.getByCoordinates(map, point, numberOfElements, classes)) {
      tmpList.add(new DistanceToObject(element, point));
    }
    Collections.sort(tmpList);
    int size = Math.min(tmpList.size(), numberOfElements);
    for (int i = 0; i < size; i++) {
      if (!perfectHit) {
        result.add(tmpList.get(i).getReference());
      } else if (tmpList.get(i).getDistance() < Configuration.EPSILON) {
        result.add(tmpList.get(i).getReference());
      }
    }
    for (BioEntity bioEntity: result) {
      Hibernate.initialize(bioEntity);
    }
    return result;
  }

  @Override
  public List<String> getSuggestedQueryList(String projectId) throws QueryException {
    return elementDao.getSuggestedQueries(modelService.getModelsByMapId(projectId, "*"));
  }

  /**
   * Tries to transform query into {@link MiriamData}.
   *
   * @param string
   *          query to transform
   * @return {@link MiriamData} that described query or null if query cannot be
   *         converted
   */
  protected MiriamData getMiriamTypeForQuery(String string) {
    for (MiriamType mt : MiriamType.values()) {
      if (string.startsWith(mt.toString().toLowerCase() + ":")) {
        return new MiriamData(mt, string.substring(mt.toString().length() + 1));
      }
    }
    return null;
  }

  /**
   * This class represents distance between object and some point. It's designed
   * to help sort objects by their distance to some point. It It contains two
   * fields: object reference and distance.
   *
   * @author Piotr Gawron
   *
   */
  private class DistanceToObject implements Comparable<DistanceToObject> {
    /**
     * Reference to the object.
     */
    private BioEntity reference;

    /**
     * Distance between the object and some point.
     */
    private double distance;

    /**
     * Constructor for alias objects.
     *
     * @param alias
     *          alias reference to store
     * @param point
     *          point from which the distance will be computed
     */
    DistanceToObject(Element alias, Point2D point) {
      reference = alias;
      distance = alias.getDistanceFromPoint(point);
    }

    public DistanceToObject(Pair<? extends BioEntity, Double> pair) {
      reference = pair.getLeft();
      distance = pair.getRight();
    }

    @Override
    public int compareTo(DistanceToObject arg0) {
      if (arg0.getDistance() < getDistance()) {
        return 1;
      } else if (arg0.getDistance() > getDistance()) {
        return -1;
      } else {
        if (arg0.getReference().getZ() > getReference().getZ()) {
          return 1;
        } else if (arg0.getReference().getZ() < getReference().getZ()) {
          return -1;
        } else {
          return 0;
        }
      }
    }

    /**
     * @return the reference
     * @see #reference
     */
    public BioEntity getReference() {
      return reference;
    }

    /**
     * @return the distance
     * @see #distance
     */
    public double getDistance() {
      return distance;
    }

  }

}