package lcsb.mapviewer.web;

import static org.junit.Assert.assertEquals;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.post;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.pathParameters;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.*;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import lcsb.mapviewer.api.convert.ConvertRestImpl;
import lcsb.mapviewer.converter.ConverterParams;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.species.Element;

@RunWith(SpringJUnit4ClassRunner.class)
public class ConvertControllerIntegrationTest extends ControllerIntegrationTest {

  @Autowired
  private ConvertRestImpl converterRestImpl;

  Logger logger = LogManager.getLogger();

  @Test
  public void testConvertInvalidFile() throws Exception {
    String body = "invalid content";
    RequestBuilder request = post("/convert/image/SBGN-ML:svg")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void convertWithBOM() throws Exception {
    FileReader fr = new FileReader(new File("src/test/resources/SBGN_with_BOM.sbgn"));
    BufferedReader br = new BufferedReader(fr);
    StringBuilder tmp = new StringBuilder();
    String line;
    while ((line = br.readLine()) != null) {
      tmp.append(line + System.getProperty("line.separator"));
    }
    RequestBuilder request = post("/convert/image/SBGN-ML:svg")
        .content(tmp.toString().getBytes("UTF-8")).characterEncoding("UTF-8");

    br.close();
    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testConvertInvalidMapFormat() throws Exception {
    String body = "invalid content";
    RequestBuilder request = post("/convert/image/unknown:svg")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testMergeCellDesignerMaps() throws Exception {
    byte[] body = Files.readAllBytes(new File("src/test/resources/convert/cd-maps.zip").toPath());
    RequestBuilder request = post("/convert/merge/{inputFormat}:{outputFormat}", "CellDesigner_SBML",
        "CellDesigner_SBML")
            .header("post-filename", "maps.zip")
            .content(body);

    byte[] content = mockMvc.perform(request)
        .andDo(document("converter/merge_maps",
            pathParameters(
                parameterWithName("inputFormat")
                    .description("input format, available options: " + getConverters()),
                parameterWithName("outputFormat")
                    .description("output format, available options: " + getConverters()))))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsByteArray();

    CellDesignerXmlParser parser = new CellDesignerXmlParser();
    Model model = parser.createModel(new ConverterParams().inputStream(new ByteArrayInputStream(content)));

    int s1 = 0;
    int s2 = 0;
    int s3 = 0;

    for (Element element : model.getElements()) {
      if (element.getName().equals("s1")) {
        s1++;
      }
      if (element.getName().equals("s2")) {
        s2++;
      }
      if (element.getName().equals("s3")) {
        s3++;
      }
    }
    assertEquals(1, s1);
    assertEquals(2, s2);
    assertEquals(1, s3);
  }

  @Test
  public void testListAvailableFormats() throws Exception {
    RequestBuilder request = get("/convert/");

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("converter/list_formats",
            responseFields(
                fieldWithPath("inputs")
                    .description("list of available input formats")
                    .type("array"),
                fieldWithPath("inputs[].available_names")
                    .description("list of names that could be used for specific format")
                    .type("array<string>"),
                fieldWithPath("outputs")
                    .description("list of available input formats")
                    .type("array"),
                fieldWithPath("outputs[].available_names")
                    .description("list of names that could be used for specific format")
                    .type("array<string>"))));
  }

  @Test
  public void convertCellDesignerToGpml() throws Exception {
    File file = new File("src/test/resources/convert/cd-sample.xml");
    String content = FileUtils.readFileToString(file, StandardCharsets.UTF_8);
    RequestBuilder request = post("/convert/{inputFormat}:{outputFormat}", "CellDesigner_SBML", "GPML")
        .header("post-filename", "input_file.xml")
        .content(content.getBytes("UTF-8"))
        .characterEncoding("UTF-8");

    mockMvc.perform(request)
        .andDo(document("converter/format_conversion",
            pathParameters(
                parameterWithName("inputFormat")
                    .description("input format, available options: " + getConverters()),
                parameterWithName("outputFormat")
                    .description("output format, available options: " + getConverters()))))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testListAvailableImageFormats() throws Exception {
    RequestBuilder request = get("/convert/image/");

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("converter/list_image_formats",
            responseFields(
                fieldWithPath("inputs")
                    .description("list of available input formats")
                    .type("array"),
                fieldWithPath("inputs[].available_names")
                    .description("list of names that could be used for specific format")
                    .type("array<string>"),
                fieldWithPath("outputs")
                    .description("list of available input formats")
                    .type("array"),
                fieldWithPath("outputs[].available_names")
                    .description("list of names that could be used for specific format")
                    .type("array<string>"))));
  }

  @Test
  public void convertCellDesignerToSvg() throws Exception {
    File file = new File("src/test/resources/convert/cd-sample.xml");
    String content = FileUtils.readFileToString(file, StandardCharsets.UTF_8);
    RequestBuilder request = post("/convert/image/{inputFormat}:{outputFormat}", "CellDesigner_SBML", "png")
        .header("post-filename", "input_file.xml")
        .content(content.getBytes("UTF-8"))
        .characterEncoding("UTF-8");

    mockMvc.perform(request)
        .andDo(document("converter/image_conversion",
            pathParameters(
                parameterWithName("inputFormat")
                    .description("input format, available options: " + getConverters()),
                parameterWithName("outputFormat")
                    .description("output format, available options: " + getImageConverters()))))
        .andExpect(status().is2xxSuccessful());
  }

  private String getConverters() {
    List<String> converters = new ArrayList<>();
    for (Map<String, List<String>> entry : converterRestImpl.getInformation().get("inputs")) {
      converters.addAll(entry.get("available_names"));
    }
    return StringUtils.join(converters, ", ");
  }

  private String getImageConverters() {
    List<String> converters = new ArrayList<>();
    for (Map<String, List<String>> entry : converterRestImpl.getInformation().get("outputs")) {
      converters.addAll(entry.get("available_names"));
    }
    return StringUtils.join(converters, ", ");
  }

}
