package lcsb.mapviewer.web;

import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.payload.PayloadDocumentation.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.junit.runner.RunWith;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.ResponseFieldsSnippet;
import org.springframework.test.annotation.Rollback;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.model.Project;

@RunWith(SpringJUnit4ClassRunner.class)
@Transactional
@Rollback
public class DrugControllerIntegrationTest extends ControllerIntegrationTest {

  private static String TEST_PROJECT = "TEST_PROJECT";

  Logger logger = LogManager.getLogger();

  @Before
  public void setup() {
    createProject(TEST_PROJECT);
  }

  @After
  public void tearDown() {
  }

  @Test
  public void testSearchDrugsInProjectUrl() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/{projectId}/drugs:search?query=aspirin", TEST_PROJECT)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("projects/project_drugs/search_by_query",
            getProjectPathParameters(),
            getChemicalFilter(),
            listOfDrugFields()));
  }

  @Test
  public void testSearchDrugsInUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/*/drugs:search?query=xyz")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testSearchDrugsByTargetInUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/*/drugs:search?target=ALIAS:123")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testSearchDrugsByTargetInSubmap() throws Exception {
    Project project = createProject("test");
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    int submapId = project.getTopModel().getSubmodels().iterator().next().getElements()
        .iterator().next().getId();
    RequestBuilder request = get("/projects/" + project.getProjectId() + "/drugs:search?target=ALIAS:" + submapId)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetSuggestedList() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/{projectId}/drugs/suggestedQueryList", TEST_PROJECT)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_drugs/suggested_query_list",
            getProjectPathParameters(),
            responseFields(
                fieldWithPath("[]")
                    .description("list of suggested drug queries")
                    .type("array<string>"))))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetSuggestedListWithUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/*/drugs/suggestedQueryList")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  private ResponseFieldsSnippet listOfDrugFields() {
    return responseFields(
        subsectionWithPath("[].id")
            .description("identifier of the chemical")
            .type("object"),
        fieldWithPath("[].name")
            .description("name")
            .type("string"),
        fieldWithPath("[].description")
            .description("description")
            .type("string"),
        fieldWithPath("[].synonyms")
            .description("list of synonyms")
            .type("array<string>"),
        fieldWithPath("[].brandNames")
            .description("list of brand names")
            .type("array<string>"),
        fieldWithPath("[].bloodBrainBarrier")
            .description("does drug cross blood brain barrier")
            .type("string"),
        subsectionWithPath("[].references")
            .description("list of references")
            .type("array<Reference>"),
        fieldWithPath("[].targets")
            .description("list of targets")
            .type("array<Target>"),
        fieldWithPath("[].targets[].name")
            .description("target name")
            .type("string"),
        subsectionWithPath("[].targets[].references")
            .description("list of target references")
            .type("array<Reference>"),
        subsectionWithPath("[].targets[].targetElements")
            .description("list of elements on the map associated with this target")
            .type("array<object>"),
        subsectionWithPath("[].targets[].targetParticipants")
            .description("list of identifiers associated with this target")
            .type("array<object>"));
  }

}
