package lcsb.mapviewer.web;

import static org.junit.Assert.assertNotNull;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.post;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.request.RequestDocumentation.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.Arrays;
import java.util.List;

import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import com.google.gson.JsonParser;

import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.persist.dao.cache.UploadedFileEntryDao;
import lcsb.mapviewer.services.interfaces.IFileService;
import lcsb.mapviewer.services.interfaces.IUserService;

@RunWith(SpringJUnit4ClassRunner.class)
public class FileControllerIntegrationTest extends ControllerIntegrationTest {

  @Autowired
  private UploadedFileEntryDao uploadedFileEntryDao;

  @Autowired
  private IUserService userService;
  
  @Autowired
  private IFileService fileService;

  @Before
  public void setup() {
  }

  @Test
  public void canCreateAndUploadFile() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("filename", "test_file"),
        new BasicNameValuePair("length", "999999"))));

    RequestBuilder request = post("/files/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    String response = mockMvc.perform(request)
        .andDo(document("file/create_file",
            requestParameters(
                parameterWithName("filename")
                    .description("original name of the file"),
                parameterWithName("length")
                    .description("length of the file (in bytes)")),
            responseFields(fileResponseFields())))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    String fileId = new JsonParser()
        .parse(response)
        .getAsJsonObject()
        .get("id")
        .getAsString();

    body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(

        new BasicNameValuePair("data", "test_content"))));

    request = post("/files/{fileId}:uploadContent", fileId)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("file/upload_content",
            pathParameters(parameterWithName("fileId")
                .description("file id")),
            requestParameters(
                parameterWithName("data")
                    .description("content to append")),
            responseFields(fileResponseFields())))
        .andExpect(status().is2xxSuccessful());

    assertNotNull(fileService.getById(Integer.valueOf(fileId)));
  }

  @Test
  public void tryToAppendToFileWithoutOwner() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    UploadedFileEntry file = new UploadedFileEntry();
    file.setFileContent(new byte[] {});
    file.setLength(99999);
    callInSeparateThread(() -> {
      try {
        uploadedFileEntryDao.add(file);
      } catch (Exception e) {
        e.printStackTrace();
      }
      return null;
    });

    try {
      String fileId = String.valueOf(file.getId());

      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("id", fileId),
          new BasicNameValuePair("data", "test_content"))));

      RequestBuilder request = post("/files/" + fileId + ":uploadContent")
          .content(body)
          .session(session);

      mockMvc.perform(request)
          .andExpect(status().isForbidden());
    } finally {
      removeFileInSeparateThread(file);
    }
  }

  @Test
  public void getInfoAboutFile() throws Exception {
    UploadedFileEntry file = createFileInSeparateThread("Hello world",
        userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    try {
      MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

      RequestBuilder request = get("/files/{fileId}", file.getId() + "")
          .session(session);

      mockMvc.perform(request)
          .andDo(document("file/get_file",
              pathParameters(parameterWithName("fileId")
                  .description("file id")),
              responseFields(fileResponseFields())))
          .andExpect(status().is2xxSuccessful());
    } finally {
      removeFileInSeparateThread(file);
    }
  }

  private List<FieldDescriptor> fileResponseFields() {
    return Arrays.asList(
        fieldWithPath("id")
            .description("unique id in the system")
            .type("number"),
        fieldWithPath("filename")
            .description("original name of the file")
            .type("number"),
        fieldWithPath("length")
            .description("file length")
            .type("number"),
        fieldWithPath("owner")
            .description("login of the file owner")
            .type("number"),
        fieldWithPath("uploadedDataLength")
            .description("total number of uploaded bytes")
            .type("number"));
  }

  @Test
  public void appendToNonExistingFile() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("id", "-1"),
        new BasicNameValuePair("data", "test_content"))));

    RequestBuilder request = post("/files/-1:uploadContent")
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is4xxClientError());
  }

  @Test
  public void createAndAppendToExistingFile() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("filename", "unknown.txt"),
        new BasicNameValuePair("length", "29"))));

    RequestBuilder request = post("/files/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful()).andReturn().getResponse().getContentAsString();
    int id = new JsonParser().parse(response).getAsJsonObject().get("id").getAsInt();

    String fileContent = "elementIdentifier\tvalue\nxx\t-1";

    RequestBuilder postContentRequest = post("/files/" + id + ":uploadContent")
        .content(fileContent)
        .session(session);

    mockMvc.perform(postContentRequest)
        .andExpect(status().is2xxSuccessful());
  }

}
