package lcsb.mapviewer.web;

import static org.junit.Assert.assertNotNull;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.pathParameters;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.Arrays;
import java.util.List;

import org.junit.*;
import org.junit.runner.RunWith;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.payload.ResponseFieldsSnippet;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.model.ModelData;

@RunWith(SpringJUnit4ClassRunner.class)
public class FunctionControllerIntegrationTest extends ControllerIntegrationTest {

  Project project;
  ModelData map;

  @Before
  public void setup() throws Exception {
    callInSeparateThread(() -> {
      project = createProject(TEST_PROJECT);
      map = project.getTopModel().getModelData();
      return null;
    });
    assertNotNull(project);
  }

  @After
  public void tearDown() throws Exception {
    removeProjectInSeparateThread(project);
  }

  @Test
  public void testGetFunctionListWithUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/*/models/*/functions/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetFunctionWithUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/*/models/1/functions/1")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetFunctions() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/{projectId}/models/{mapId}/functions/", TEST_PROJECT, map.getId())
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_maps/get_functions",
            pathParameters(parameterWithName("projectId").description("project identifier"),
                parameterWithName("mapId").description("map identifier")),
            listOfFunctionFields()))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetFunction() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/{projectId}/models/{mapId}/functions/{functionId}", TEST_PROJECT,
        map.getId(),
        map.getFunctions().iterator().next().getId())
            .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_maps/get_function",
            pathParameters(parameterWithName("projectId").description("project identifier"),
                parameterWithName("mapId").description("map identifier"),
                parameterWithName("functionId").description("function identifier")),
            responseFields(functionFields())))
        .andExpect(status().is2xxSuccessful());
  }

  private List<FieldDescriptor> functionFields() {
    return Arrays.asList(
        fieldWithPath("id")
            .description("unique function identifier")
            .type("number"),
        fieldWithPath("name")
            .description("name")
            .type("string"),
        fieldWithPath("definition")
            .description("definition")
            .type("string"),
        fieldWithPath("functionId")
            .description("function identifier taken from source file")
            .type("string"),
        fieldWithPath("arguments")
            .description("list of function parameters")
            .type("array<object>"));
  }

  private ResponseFieldsSnippet listOfFunctionFields() {
    return responseFields(
        fieldWithPath("[]").description("list of functions"))
            .andWithPrefix("[].", functionFields());
  }

}
