package lcsb.mapviewer.web;

import static org.junit.Assert.assertTrue;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.*;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.request.RequestDocumentation.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.*;

import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import com.google.gson.JsonParser;

import lcsb.mapviewer.model.cache.BigFileEntry;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.layout.*;
import lcsb.mapviewer.persist.dao.cache.BigFileEntryDao;
import lcsb.mapviewer.persist.dao.map.layout.ReferenceGenomeDao;
import lcsb.mapviewer.services.interfaces.IReferenceGenomeService;

@RunWith(SpringJUnit4ClassRunner.class)
@ActiveProfiles("webBigFileAlwaysExistsTestProfile")
public class GenomicsControllerIntegrationTest extends ControllerIntegrationTest {

  MiriamData ebolaId = new MiriamData(MiriamType.TAXONOMY, "1570291");
  String ebolaUrl = "ftp://hgdownload.cse.ucsc.edu/goldenPath/eboVir3/bigZips/eboVir3.2bit";

  Logger logger = LogManager.getLogger();

  @Autowired
  private ProjectSnippets snippets;

  @Autowired
  private ReferenceGenomeDao referenceGenomeDao;

  @Autowired
  private IReferenceGenomeService referenceGenomeService;

  @Autowired
  private BigFileEntryDao bigFileEntryDao;

  @Before
  public void setup() {
  }

  @After
  public void tearDown() {
  }

  @Test
  public void testGetTaxonomies() throws Exception {

    RequestBuilder request = get("/genomics/taxonomies/");

    String response = mockMvc.perform(request)
        .andDo(document("genomics/get_organisms",
            responseFields(fieldWithPath("[]")
                .description("list of organisms")
                .type("array")).andWithPrefix("[].", organismResponseFields())))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    int taxonomies = new JsonParser()
        .parse(response)
        .getAsJsonArray().size();

    assertTrue("list of taxonomies shouldn't be empty", taxonomies > 0);
  }

  private List<FieldDescriptor> organismResponseFields() {
    return Arrays.asList(
        fieldWithPath("resource")
            .description("taxonomy id of the organism")
            .type("string"),
        fieldWithPath("annotatorClassName").ignored(),
        fieldWithPath("id").ignored(),
        fieldWithPath("link").ignored(),
        fieldWithPath("type").ignored());
  }

  @Test
  public void testGetGenomeTypes() throws Exception {

    RequestBuilder request = get("/genomics/taxonomies/{taxonomyId}/genomeTypes/", "9606");

    mockMvc.perform(request)
        .andDo(document("genomics/get_genome_types",
            pathParameters(parameterWithName("taxonomyId").description("organism taxonomy id")),
            responseFields(
                fieldWithPath("[]")
                    .description("list of genome types")
                    .type("array"),
                fieldWithPath("[].type")
                    .description("type")
                    .type("string"))))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetAllGenomeVersions() throws Exception {

    RequestBuilder request = get("/genomics/taxonomies/{taxonomyId}/genomeTypes/{genomeType}/versions/", "9606",
        ReferenceGenomeType.UCSC.name());

    mockMvc.perform(request)
        .andDo(document("genomics/get_genome_versions",
            pathParameters(
                parameterWithName("taxonomyId").description("organism taxonomy id"),
                parameterWithName("genomeType").description(
                    "genome type, acceptable values: " + snippets.getOptionsAsString(ReferenceGenomeType.class))),
            responseFields(
                fieldWithPath("[]")
                    .description("list of genome versions")
                    .type("array"),
                fieldWithPath("[].version")
                    .description("version")
                    .type("string"))))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetRemoteUrl() throws Exception {

    RequestBuilder request = get(
        "/genomics/taxonomies/{taxonomyId}/genomeTypes/{genomeType}/versions/{version}:getAvailableRemoteUrls", "9606",
        ReferenceGenomeType.UCSC.name(), "hg38");

    mockMvc.perform(request)
        .andDo(document("genomics/get_genome_remote_url",
            pathParameters(
                parameterWithName("taxonomyId").description("organism taxonomy id"),
                parameterWithName("genomeType").description(
                    "genome type, acceptable values: " + snippets.getOptionsAsString(ReferenceGenomeType.class)),
                parameterWithName("version").description("genome version")),
            responseFields(
                fieldWithPath("[]")
                    .description("list of urls")
                    .type("array"),
                fieldWithPath("[].url")
                    .description("url")
                    .type("string"))))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testDownloadGenomeDataFromRemote() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("organismId", ebolaId.getResource()),
        new BasicNameValuePair("type", ReferenceGenomeType.UCSC.name()),
        new BasicNameValuePair("version", "eboVir3"),
        new BasicNameValuePair("url", "http://minerva-dev.lcsb.uni.lu/eboVir3.2bit"))));

    RequestBuilder request = post("/genomics/")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("genomics/download_genome",
            requestParameters(
                parameterWithName("organismId").description("organism taxonomy id"),
                parameterWithName("type").description(
                    "genome type, acceptable values: " + snippets.getOptionsAsString(ReferenceGenomeType.class)),
                parameterWithName("version").description("genome version"),
                parameterWithName("url").description("url address from where the file should be downloaded")),
            responseFields()))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

  }

  @Test
  public void testGetDownloadedGenomes() throws Exception {
    ReferenceGenome genome = createGenome();

    try {
      RequestBuilder request = get("/genomics/");

      mockMvc.perform(request)
          .andDo(document("genomics/get_genomes",
              responseFields(fieldWithPath("[]")
                  .description("list of downloaded genomes")
                  .type("array")).andWithPrefix("[].", genomeInformationResponseFields())))
          .andExpect(status().is2xxSuccessful())
          .andReturn().getResponse().getContentAsString();
    } finally {
      referenceGenomeService.removeGenome(genome);
    }

  }

  private ReferenceGenome createGenome() throws Exception {
    ReferenceGenome genome = new ReferenceGenome();
    genome.setOrganism(ebolaId);
    genome.setSourceUrl(ebolaUrl);
    genome.setDownloadProgress(100.0);
    genome.setType(ReferenceGenomeType.UCSC);
    genome.setVersion("eboVir3");
    return callInSeparateThread(() -> {
      referenceGenomeDao.add(genome);
      return genome;
    });

  }

  @Test
  public void testGetGenomeInformation() throws Exception {
    ReferenceGenome genome = createGenome();

    BigFileEntry entry = new BigFileEntry();
    entry.setDownloadProgress(100.0);
    entry.setDownloadDate(Calendar.getInstance());
    entry.setLocalPath("../");
    entry.setUrl(ebolaUrl);
    callInSeparateThread(() -> {
      bigFileEntryDao.add(entry);

      entry.setLocalPath(entry.getLocalPath() + "/" + entry.getId() + "/eboVir3.2bit");
      bigFileEntryDao.update(entry);
      return null;
    });

    try {
      RequestBuilder request = get(
          "/genomics/taxonomies/{taxonomyId}/genomeTypes/{genomeType}/versions/{version}/", ebolaId.getResource(),
          ReferenceGenomeType.UCSC.name(), "eboVir3");

      mockMvc.perform(request)
          .andDo(document("genomics/get_genome_information",
              pathParameters(
                  parameterWithName("taxonomyId").description("organism taxonomy id"),
                  parameterWithName("genomeType").description(
                      "genome type, acceptable values: " + snippets.getOptionsAsString(ReferenceGenomeType.class)),
                  parameterWithName("version").description("genome version")),
              responseFields(genomeInformationResponseFields())))
          .andExpect(status().is2xxSuccessful());
    } finally {
      referenceGenomeService.removeGenome(genome);
      callInSeparateThread(() -> {
        bigFileEntryDao.delete(entry);
        return null;
      });
    }
  }

  private List<FieldDescriptor> genomeInformationResponseFields() {
    return Arrays.asList(fieldWithPath("downloadProgress")
        .description("download progress (0-100%)")
        .type("double"),
        fieldWithPath("geneMapping")
            .description("list of available gene mappings for given genome")
            .type("array"),
        fieldWithPath("idObject")
            .description("unique id of genome in minerva")
            .type("number"),
        fieldWithPath("localUrl")
            .description("url on minerva where local copy of genome can be accessed")
            .type("string"),
        fieldWithPath("organism")
            .description("organism identifier")
            .type("object"),
        fieldWithPath("organism.resource")
            .description("organism identifier")
            .type("string"),
        fieldWithPath("organism.type")
            .description("organism identifier type (usually " + MiriamType.TAXONOMY + ")")
            .type("string"),
        fieldWithPath("organism.annotatorClassName")
            .ignored(),
        fieldWithPath("organism.id")
            .ignored(),
        fieldWithPath("organism.link")
            .ignored(),
        fieldWithPath("sourceUrl")
            .description("genome source url")
            .type("string"),
        fieldWithPath("type")
            .description("type of genome (database from which it was downloaded)")
            .type("string"),
        fieldWithPath("version")
            .description("genome version")
            .type("string"));
  }

  @Test
  public void deleteGenomeInformation() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    ReferenceGenome genome = createGenome();

    RequestBuilder request = delete("/genomics/{genomeId}/", genome.getId())
        .session(session);

    mockMvc.perform(request)
        .andDo(document("genomics/delete_genome_information",
            pathParameters(
                parameterWithName("genomeId").description("genome id")),
            responseFields()))
        .andExpect(status().is2xxSuccessful());

  }

  @Test
  public void deleteWithInvalidGenomeId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/genomics/ /").session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());

  }

  @Test
  public void deleteWithNonExistingGenomeId() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/genomics/-1/").session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());

  }

  @Test
  public void addGeneMappingInformation() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    ReferenceGenome genome = createGenome();
    try {
      String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
          new BasicNameValuePair("name", "mappingName"),
          new BasicNameValuePair("url", "https://minerva-dev.lcsb.uni.lu/tmp/refGene.bb"))));

      RequestBuilder request = post("/genomics/{genomeId}/geneMapping/", genome.getId())
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(session);

      mockMvc.perform(request)
          .andDo(document("genomics/add_gene_mapping",
              pathParameters(
                  parameterWithName("genomeId").description("genome id")),
              requestParameters(parameterWithName("name").description("name of the gene-genome mapping"),
                  parameterWithName("url").description("url where file with mapping is located")),
              responseFields()))
          .andExpect(status().is2xxSuccessful());
    } finally {
      referenceGenomeService.removeGenome(genome);
    }

  }

  @Test
  public void deleteGeneMappingInformation() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    ReferenceGenome genome = createGenome();

    ReferenceGenomeGeneMapping mapping = new ReferenceGenomeGeneMapping();
    mapping.setName("x");
    mapping.setSourceUrl("https://minerva-dev.lcsb.uni.lu/tmp/refGene.bb");
    genome.addReferenceGenomeGeneMapping(mapping);
    callInSeparateThread(() -> {
      referenceGenomeDao.update(genome);
      return genome;
    });
    try {
      RequestBuilder request = delete("/genomics/{genomeId}/geneMapping/{geneMappingId}/", genome.getId(),
          genome.getGeneMapping().get(0).getId())
              .session(session);

      mockMvc.perform(request)
          .andDo(document("genomics/delete_gene_mapping",
              pathParameters(
                  parameterWithName("genomeId").description("genome id"),
                  parameterWithName("geneMappingId").description("gene genome mapping id"))))
          .andExpect(status().is2xxSuccessful());
    } finally {
      referenceGenomeService.removeGenome(genome);
    }

  }

  @Test
  public void deleteInvalidGeneMappingInformation() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/genomics/ /geneMapping/ /").session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

}
