package lcsb.mapviewer.web;

import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.pathParameters;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.Arrays;
import java.util.List;

import org.junit.*;
import org.junit.runner.RunWith;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.payload.ResponseFieldsSnippet;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.model.ModelData;

@RunWith(SpringJUnit4ClassRunner.class)
public class ParameterControllerIntegrationTest extends ControllerIntegrationTest {

  private Project project;
  private ModelData map;

  @Before
  public void setup() {
    callInSeparateThreadNoReturn(() -> {
      project = createProject(TEST_PROJECT);
      map = project.getTopModel().getModelData();
    });
  }

  @After
  public void tearDown() throws Exception {
    removeProjectInSeparateThread(project);
  }

  @Test
  public void testGetParameterListWithUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/*/models/*/parameters/")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetParameterWithUndefinedProject() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/*/models/*/parameters/1")
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isNotFound());
  }

  @Test
  public void testGetParameters() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/{projectId}/models/{mapId}/parameters/", TEST_PROJECT, "*")
        .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_maps/get_parameters",
            pathParameters(parameterWithName("projectId").description("project identifier"),
                parameterWithName("mapId").description("map identifier")),
            listOfParameterFields()))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testGetParameter() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/projects/{projectId}/models/{mapId}/parameters/{parameterId}", TEST_PROJECT, "*",
        map.getParameters().iterator().next().getId())
            .session(session);

    mockMvc.perform(request)
        .andDo(document("projects/project_maps/get_parameter",
            pathParameters(parameterWithName("projectId").description("project identifier"),
                parameterWithName("mapId").description("map identifier"),
                parameterWithName("parameterId").description("parameter identifier")),
            responseFields(parameterFields())))
        .andExpect(status().is2xxSuccessful());
  }

  private List<FieldDescriptor> parameterFields() {
    return Arrays.asList(
        fieldWithPath("id")
            .description("unique parameter identifier")
            .type("number"),
        fieldWithPath("name")
            .description("name")
            .type("string"),
        fieldWithPath("value")
            .description("value")
            .type("number"),
        fieldWithPath("unitsId")
            .description("unit identifier")
            .type("number"),
        fieldWithPath("global")
            .description("is parameter global")
            .type("boolean"),
        fieldWithPath("parameterId")
            .description("parameter identifier taken from source file")
            .type("string"));
  }

  private ResponseFieldsSnippet listOfParameterFields() {
    return responseFields(
        fieldWithPath("[]").description("list of parameters"))
            .andWithPrefix("[].", parameterFields());
  }

}
