package lcsb.mapviewer.web;

import static org.junit.Assert.assertEquals;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.post;
import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.requestParameters;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import com.google.gson.JsonParser;

import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.services.interfaces.IUserService;

@RunWith(SpringJUnit4ClassRunner.class)
public class UserControllerAnonymousIntegrationTest extends ControllerIntegrationTest {

  @Autowired
  private IUserService userService;

  Logger logger = LogManager.getLogger();

  @Before
  public void setup() {
  }

  @Test
  public void grantPrivilegeForAnonymousUser() throws Exception {
    try {
      MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

      String body = "{\"privileges\":{\"IS_ADMIN\":true}}";

      RequestBuilder request = patch("/users/" + Configuration.ANONYMOUS_LOGIN + ":updatePrivileges")
          .contentType(MediaType.APPLICATION_FORM_URLENCODED)
          .content(body)
          .session(session);

      mockMvc.perform(request)
          .andExpect(status().is2xxSuccessful());

      RequestBuilder anonymousRequest = get("/users/");
      mockMvc.perform(anonymousRequest)
          .andExpect(status().is2xxSuccessful());
    } finally {
      userService.revokeUserPrivilege(userService.getUserByLogin(Configuration.ANONYMOUS_LOGIN),
          PrivilegeType.IS_ADMIN);
    }
  }

  @Test
  public void loginAsAnonymousUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "{\"user\":{\"password\":\"test\"}}";

    RequestBuilder request = patch("/users/" + Configuration.ANONYMOUS_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    RequestBuilder loginRequest = post("/api/doLogin")
        .param("login", Configuration.ANONYMOUS_LOGIN)
        .param("password", "test");
    mockMvc.perform(loginRequest)
        .andExpect(status().is4xxClientError());

  }

  @Test
  public void removeAnonymousUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = delete("/users/" + Configuration.ANONYMOUS_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is4xxClientError());
  }

  @Test
  public void checkIsAuthenticatedAdmin() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/users/isSessionValid")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("authentication/isSessionValid",
            responseFields(
                fieldWithPath("login")
                    .description("user login of the session owner"))))
        .andReturn().getResponse().getContentAsString();

    String login = new JsonParser()
        .parse(response)
        .getAsJsonObject()
        .get("login")
        .getAsString();
    assertEquals(BUILT_IN_TEST_ADMIN_LOGIN, login);
  }

  @Test
  public void checkIsAuthenticatedAnonymous() throws Exception {
    RequestBuilder request = get("/users/isSessionValid")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED);

    mockMvc.perform(request)
        .andExpect(status().is4xxClientError());
  }

  @Test
  public void testLogin() throws Exception {
    RequestBuilder request = post("/api/doLogin")
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content("login=" + BUILT_IN_TEST_ADMIN_LOGIN + "&password=" + BUILT_IN_TEST_ADMIN_PASSWORD);

    mockMvc.perform(request)
        .andDo(document("authentication/login", requestParameters(parameterWithName("login").description("login"),
            parameterWithName("password").description("password")),
            responseFields(
                fieldWithPath("info")
                    .description("status message"),
                fieldWithPath("login")
                    .description("user login"),
                fieldWithPath("token")
                    .description("session token"))))
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void testLogout() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = post("/api/doLogout")
        .session(session);

    mockMvc.perform(request)
        .andDo(document("authentication/logout",
            responseFields(
                fieldWithPath("status")
                    .description("status message"))))
        .andExpect(status().is2xxSuccessful());
  }

}
