package lcsb.mapviewer.web;

import static org.junit.Assert.*;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.*;
import static org.springframework.restdocs.payload.PayloadDocumentation.*;
import static org.springframework.restdocs.request.RequestDocumentation.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.*;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.junit.*;
import org.junit.runner.RunWith;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.restdocs.payload.*;
import org.springframework.restdocs.request.ParameterDescriptor;
import org.springframework.restdocs.request.RequestParametersSnippet;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import com.google.gson.*;

import lcsb.mapviewer.annotation.services.annotators.UniprotAnnotator;
import lcsb.mapviewer.api.users.UserRestImpl;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.species.*;
import lcsb.mapviewer.model.security.Privilege;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.*;
import lcsb.mapviewer.persist.dao.ProjectDao;
import lcsb.mapviewer.persist.dao.user.ResetPasswordTokenDao;
import lcsb.mapviewer.services.interfaces.*;
import lcsb.mapviewer.services.utils.EmailSender;

@RunWith(SpringJUnit4ClassRunner.class)
@ActiveProfiles("emailSenderProfile")
public class UserControllerIntegrationTest extends ControllerIntegrationTest {

  private static final String TEST_USER_PASSWORD = "test_pass";
  private static final String TEST_USER_LOGIN = "test_user";

  private static final String TEST_LOGIN = "xyz";

  Logger logger = LogManager.getLogger();
  private User user;

  @Autowired
  private IUserService userService;

  @Autowired
  private IConfigurationService configurationService;

  @Autowired
  private ProjectDao projectDao;

  @Autowired
  private ResetPasswordTokenDao resetPasswordTokenDao;

  @Autowired
  private EmailSender emailSender;

  @Autowired
  private IDataOverlayService dataOverlayService;

  @Autowired
  private ProjectSnippets projectSnippets;

  @Autowired
  private UserRestImpl userRestImpl;

  @Before
  public void setup() throws Exception {
    user = createUserInSeparateThread(TEST_USER_LOGIN, TEST_USER_PASSWORD);
  }

  @After
  public void tearDown() throws Exception {
    removeUserInSeparateThread(user);
    removeUserInSeparateThread(userService.getUserByLogin(TEST_LOGIN));
    removeProjectInSeparateThread(TEST_PROJECT);
  }

  @Test
  public void grantPrivilege() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "{\"privileges\":{\"IS_ADMIN\":true}}";

    RequestBuilder request = patch("/users/{login}:updatePrivileges", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/update_privilege_1",
            pathParameters(parameterWithName("login").description("user login")),
            updatePrivilegeRequest(),
            responseFields(getUserResponseFields())))
        .andExpect(status().is2xxSuccessful());
    callInSeparateThread(() -> {
      assertEquals(1, userService.getUserByLogin(TEST_USER_LOGIN).getPrivileges().size());
      return null;
    });
  }

  @Test
  public void grantPrivilegeForNonExistingUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "{\"privileges\":{\"IS_ADMIN\":true}}";

    RequestBuilder request = patch("/users/unkown_login:updatePrivileges")
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is4xxClientError());
  }

  @Test
  public void revokePrivilege() throws Exception {
    userService.grantUserPrivilege(user, PrivilegeType.IS_ADMIN);

    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    callInSeparateThread(() -> {
      assertEquals(1, userService.getUserByLogin(TEST_USER_LOGIN).getPrivileges().size());
      return null;
    });

    String body = "{\"privileges\":{\"IS_ADMIN\":false}}";

    RequestBuilder request = patch("/users/" + TEST_USER_LOGIN + ":updatePrivileges")
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      assertEquals("Privilege wasn't revoked", 0, user.getPrivileges().size());
      return null;
    });
  }

  @Test
  public void grantProjectPrivilege() throws Exception {
    Project project = new Project(TEST_PROJECT);
    project.setOwner(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    callInSeparateThreadNoReturn(() -> {
      projectDao.add(project);
    });

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "{\"privileges\":{\"READ_PROJECT:" + project.getProjectId() + "\":true}}";

    RequestBuilder request = patch("/users/{login}:updatePrivileges", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/update_privilege_2",
            pathParameters(parameterWithName("login").description("user login")),
            updatePrivilegeRequest(),
            responseFields(getUserResponseFields())))
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      User user = userService.getUserByLogin(TEST_USER_LOGIN);
      assertEquals(1, user.getPrivileges().size());
      return null;
    });

  }

  private RequestFieldsSnippet updatePrivilegeRequest() {
    return requestFields(
        fieldWithPath("privileges.*")
            .description("should the privilege defined as PRIVILEGE_TYPE:ID_OBJECT be granted or not")
            .type("boolean"));
  }

  @Test
  public void grantInvalidPrivilege() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "{\"privileges\":{\"XYZ\":true}}";

    RequestBuilder request = patch("/users/" + TEST_USER_LOGIN + ":updatePrivileges")
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void grantInvalidProjectPrivilege() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "{\"privileges\":{\"READ_PROJECT:-5:-1\":true}}";

    RequestBuilder request = patch("/users/" + TEST_USER_LOGIN + ":updatePrivileges")
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void updateInvalidPrivilege() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = "{\"privileges\":{\"IS_ADMIN\":\"surprise\"}}";

    RequestBuilder request = patch("/users/" + TEST_USER_LOGIN + ":updatePrivileges")
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void revokeProjectPrivilege() throws Exception {
    Project project = new Project(TEST_PROJECT);
    project.setOwner(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));
    callInSeparateThread(() -> {
      projectDao.add(project);
      return null;
    });

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String grantBody = "{\"privileges\":{\"READ_PROJECT:" + project.getId() + "\":true}}";
    String revokeBody = "{\"privileges\":{\"READ_PROJECT:" + project.getId() + "\":false}}";

    RequestBuilder grantRequest = patch("/users/" + TEST_USER_LOGIN + ":updatePrivileges")
        .content(grantBody)
        .session(session);
    RequestBuilder revokeRequest = patch("/users/" + TEST_USER_LOGIN + ":updatePrivileges")
        .content(revokeBody)
        .session(session);

    mockMvc.perform(grantRequest)
        .andExpect(status().is2xxSuccessful());
    mockMvc.perform(revokeRequest)
        .andExpect(status().is2xxSuccessful());
    assertEquals(0, user.getPrivileges().size());
  }

  @Test
  public void fetchUserPrivilege() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get("/users/{login}", TEST_USER_LOGIN)
        .session(session);

    String response = mockMvc.perform(request)
        .andDo(document("user/get_user",
            pathParameters(parameterWithName("login").description("user login")),
            userColumListRequestParam(),
            responseFields(getUserResponseFields())))
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    Map<?, ?> responseObject = new Gson().fromJson(response, Map.class);

    List<?> privileges = (List<?>) responseObject.get("privileges");
    assertEquals(0, privileges.size());
  }

  private RequestParametersSnippet userColumListRequestParam() {
    List<String> columns = new ArrayList<>(userRestImpl.createUserColumnSet(""));
    Collections.sort(columns);
    return requestParameters(parameterWithName("columns")
        .description("comma separated list of columns to be returned (accepted values: "
            + StringUtils.join(columns, ",") + ")")
        .optional());
  }

  @Test
  public void testUsersShouldBeFetchedInAlphabeticOrder() throws Exception {
    List<User> users = new ArrayList<>();
    try {
      for (int i = 10; i > 0; i--) {
        users.add(createUser("random_user_" + i, "passwd" + i));
      }

      MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

      RequestBuilder grantRequest = get("/users/?columns=login")
          .session(session);

      String response = mockMvc.perform(grantRequest)
          .andExpect(status().is2xxSuccessful())
          .andReturn().getResponse().getContentAsString();

      JsonArray result = new JsonParser()
          .parse(response)
          .getAsJsonArray();
      for (int i = 1; i < result.size(); i++) {
        String login1 = result.get(i - 1).getAsJsonObject().get("login").toString();
        String login2 = result.get(i).getAsJsonObject().get("login").toString();
        assertTrue("[sort issue]: " + login1 + " should be before " + login2, login1.compareTo(login2) > 0);
      }
    } finally {
      for (User u : users) {
        userService.deleteUser(u);
      }
    }
  }

  @Test
  public void testInformationAboutLdapForAllUsers() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder grantRequest = get("/users/?columns=login,ldapAccountAvailable,connectedToLdap")
        .session(session);

    String response = mockMvc.perform(grantRequest)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    JsonArray result = new JsonParser()
        .parse(response)
        .getAsJsonArray();
    for (int i = 0; i < result.size(); i++) {
      assertNotNull(result.get(i).getAsJsonObject().get("ldapAccountAvailable"));
      assertNotNull(result.get(i).getAsJsonObject().get("connectedToLdap"));
    }
  }

  @Test
  public void testInformationAboutLdapForSingleUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder grantRequest = get("/users/admin/?columns=login,ldapAccountAvailable,connectedToLdap")
        .session(session);

    String response = mockMvc.perform(grantRequest)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    JsonObject result = new JsonParser()
        .parse(response)
        .getAsJsonObject();
    assertNotNull(result.get("ldapAccountAvailable"));
    assertNotNull(result.get("connectedToLdap"));
  }

  @Test
  public void testAddUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("name", "FirstName"),
        new BasicNameValuePair("password", "FirstName"))));

    RequestBuilder grantRequest = post("/users/{login}", TEST_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(grantRequest)
        .andDo(document("user/create_user",
            pathParameters(parameterWithName("login").description("user login")),
            requestParameters(getUserRequestFieldsWithoutLogin()),
            responseFields(getUserResponseFields())))
        .andExpect(status().is2xxSuccessful());

    assertNotNull(userService.getUserByLogin(TEST_LOGIN));
  }

  @Test
  public void testAddUserWithoutPassword() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("name", "FirstName"))));

    RequestBuilder grantRequest = post("/users/" + TEST_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(grantRequest)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void testGrantDefaultPrivilegesWhenAddingUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    configurationService.setConfigurationValue(ConfigurationElementType.DEFAULT_WRITE_PROJECT, "true");
    configurationService.setConfigurationValue(ConfigurationElementType.DEFAULT_READ_PROJECT, "false");

    createProjectInSeparateThread(TEST_PROJECT);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("name", "FirstName"),
        new BasicNameValuePair("password", "FirstName"),
        new BasicNameValuePair("defaultPrivileges", "true"))));

    RequestBuilder grantRequest = post("/users/" + TEST_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(grantRequest)
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      User user = userService.getUserByLogin(TEST_LOGIN);

      assertTrue("Default privilege wasn't added after user was created",
          user.getPrivileges().contains(new Privilege(PrivilegeType.WRITE_PROJECT, "*")));
      assertTrue("Default privilege wasn't added after user was created",
          user.getPrivileges().contains(new Privilege(PrivilegeType.WRITE_PROJECT, TEST_PROJECT)));

      assertFalse("Default privilege shouldn't be added after user was created",
          user.getPrivileges().contains(new Privilege(PrivilegeType.READ_PROJECT, "*")));
      assertFalse("Default privilege shouldn't be added after user was created",
          user.getPrivileges().contains(new Privilege(PrivilegeType.READ_PROJECT, TEST_PROJECT)));
      return null;
    });

  }

  @Test
  public void testGrantDefaultGlobalPrivilegesWhenAddingUser() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);
    configurationService.setConfigurationValue(ConfigurationElementType.DEFAULT_CAN_CREATE_OVERLAYS, "true");

    createProjectInSeparateThread(TEST_PROJECT);

    String body = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("name", "FirstName"),
        new BasicNameValuePair("password", "FirstName"),
        new BasicNameValuePair("defaultPrivileges", "true"))));

    RequestBuilder grantRequest = post("/users/" + TEST_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(grantRequest)
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {

      User user = userService.getUserByLogin(TEST_LOGIN);

      assertTrue("Default privilege wasn't added after user was created",
          user.getPrivileges().contains(new Privilege(PrivilegeType.CAN_CREATE_OVERLAYS)));
      return null;
    });
  }

  @Test
  public void addUserWithoutBodyShouldThrowBadRequest() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder grantRequest = post("/users/" + TEST_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .session(session);

    mockMvc.perform(grantRequest)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void userUpdateOwnPassword() throws Exception {
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String newPassword = "new pass";
    String body = "{\"user\":{\"password\":\"" + newPassword + "\"}}";

    RequestBuilder grantRequest = patch("/users/{login}", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(grantRequest)
        .andDo(document("user/update_user",
            pathParameters(parameterWithName("login").description("user login")),
            requestParameters(getUserRequestFieldsWithoutLogin()),
            responseFields(getUserResponseFields())))
        .andExpect(status().is2xxSuccessful());

    MockHttpSession sessionWithNewPass = createSession(TEST_USER_LOGIN, newPassword);
    assertNotNull(sessionWithNewPass);
  }

  @Test
  public void userCannotUpdateOwnLdapConnection() throws Exception {
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String body = "{\"user\":{\"connectedtoldap\":" + (!user.isConnectedToLdap()) + "}}";

    RequestBuilder grantRequest = patch("/users/" + TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(grantRequest)
        .andExpect(status().isForbidden());
  }

  @Test
  public void userCannotUpdateOwnLdapConnectionButNoChangeShouldPass() throws Exception {
    MockHttpSession session = createSession(TEST_USER_LOGIN, TEST_USER_PASSWORD);

    String body = "{\"user\":{\"connectedtoldap\":" + user.isConnectedToLdap() + "}}";

    RequestBuilder grantRequest = patch("/users/" + TEST_USER_LOGIN)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED)
        .content(body)
        .session(session);

    mockMvc.perform(grantRequest)
        .andExpect(status().is2xxSuccessful());
  }

  @Test
  public void removeUserWithLayouts() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    Project project = new Project(TEST_PROJECT);
    project.setOwner(userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN));

    callInSeparateThread(() -> {
      projectDao.add(project);
      return null;
    });

    createOverlayInSeparateThread(TEST_PROJECT, user);

    RequestBuilder request = delete("/users/{login}", user.getLogin())
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/delete_user",
            pathParameters(parameterWithName("login").description("user login"))))
        .andExpect(status().is2xxSuccessful());

    assertEquals(0, dataOverlayService.getDataOverlaysByProject(project).size());
  }

  @Test
  public void requestResetPassword() throws Exception {
    long count = callInSeparateThread(() -> {
      return resetPasswordTokenDao.getCount();
    });
    configureServerForResetPasswordRequest();

    RequestBuilder grantRequest = post("/users/{login}:requestResetPassword", TEST_USER_LOGIN);

    mockMvc.perform(grantRequest)
        .andDo(document("user/request_reset_password",
            pathParameters(parameterWithName("login").description("user login"))))
        .andExpect(status().is2xxSuccessful());
    callInSeparateThread(() -> {
      assertEquals(count + 1, resetPasswordTokenDao.getCount());
      return null;
    });

  }

  @Test
  public void requestResetPasswordForLdapUser() throws Exception {
    configureServerForResetPasswordRequest();

    User user = userService.getUserByLogin(TEST_USER_LOGIN);
    user.setConnectedToLdap(true);
    userService.updateUser(user);

    RequestBuilder grantRequest = post("/users/" + TEST_USER_LOGIN + ":requestResetPassword");

    mockMvc.perform(grantRequest)
        .andExpect(status().isBadRequest());
  }

  @Test
  public void requestResetPasswordWhenMinervaRootNotConfigured() throws Exception {
    configureServerForResetPasswordRequest();
    configurationService.setConfigurationValue(ConfigurationElementType.MINERVA_ROOT, "");

    RequestBuilder grantRequest = post("/users/" + TEST_USER_LOGIN + ":requestResetPassword");

    mockMvc.perform(grantRequest)
        .andExpect(status().is5xxServerError());

  }

  @Test
  public void requestResetPasswordWhenMinervaCannotSendMails() throws Exception {
    try {
      configureServerForResetPasswordRequest();

      RequestBuilder grantRequest = post("/users/" + TEST_USER_LOGIN + ":requestResetPassword");

      Mockito.doReturn(false).when(emailSender).canSendEmails();
      mockMvc.perform(grantRequest)
          .andExpect(status().is5xxServerError());
    } finally {
      Mockito.doReturn(true).when(emailSender).canSendEmails();
    }

  }

  @Test
  public void requestResetPasswordWhenUserWithoutEmail() throws Exception {
    configureServerForResetPasswordRequest();

    User user = userService.getUserByLogin(TEST_USER_LOGIN);
    user.setEmail("");
    userService.updateUser(user);

    RequestBuilder grantRequest = post("/users/" + TEST_USER_LOGIN + ":requestResetPassword");

    mockMvc.perform(grantRequest)
        .andExpect(status().is4xxClientError());

  }

  @Test
  public void requestResetPasswordForUnknownUser() throws Exception {
    configureServerForResetPasswordRequest();

    RequestBuilder grantRequest = post("/users/blabla:requestResetPassword");

    mockMvc.perform(grantRequest)
        .andExpect(status().is4xxClientError());

  }

  @Test
  public void resetPassword() throws Exception {
    configureServerForResetPasswordRequest();

    RequestBuilder request = post("/users/" + TEST_USER_LOGIN + ":requestResetPassword");

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    String newPassword = "pass2";

    ResetPasswordToken token = callInSeparateThread(() -> {
      return resetPasswordTokenDao.getAll().iterator().next();
    });
    String content = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("token", token.getToken()),
        new BasicNameValuePair("password", newPassword))));

    RequestBuilder resetRequest = post("/users:resetPassword")
        .content(content)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED);

    mockMvc.perform(resetRequest)
        .andDo(document("user/reset_password",
            requestParameters(
                parameterWithName("password")
                    .optional()
                    .description("new password"),
                parameterWithName("token")
                    .optional()
                    .description("reset poassowrd token obtained using email"))))
        .andExpect(status().is2xxSuccessful());

    createSession(TEST_USER_LOGIN, newPassword);

  }

  @Test
  public void resetPasswordUsingTheSameTokenTwice() throws Exception {
    configureServerForResetPasswordRequest();

    RequestBuilder request = post("/users/" + TEST_USER_LOGIN + ":requestResetPassword");

    mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful());

    String newPassword = "pass2";

    ResetPasswordToken token = callInSeparateThread(() -> {
      return resetPasswordTokenDao.getAll().iterator().next();
    });
    String content = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("token", token.getToken()),
        new BasicNameValuePair("password", newPassword))));

    RequestBuilder resetRequest = post("/users:resetPassword")
        .content(content)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED);

    mockMvc.perform(resetRequest)
        .andExpect(status().is2xxSuccessful());

    content = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("token", token.getToken()),
        new BasicNameValuePair("password", newPassword + "xx"))));

    resetRequest = post("/users:resetPassword")
        .content(content)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED);

    mockMvc.perform(resetRequest)
        .andExpect(status().is4xxClientError());

    createSession(TEST_USER_LOGIN, newPassword);

  }

  @Test
  public void resetPasswordWithExpredToken() throws Exception {
    configureServerForResetPasswordRequest();

    Calendar expires = Calendar.getInstance();
    expires.add(Calendar.DATE, -1);
    ResetPasswordToken token = new ResetPasswordToken(userService.getUserByLogin(TEST_USER_LOGIN),
        "bla", expires);
    callInSeparateThread(() -> {
      resetPasswordTokenDao.add(token);
      return null;
    });

    String content = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("token", token.getToken()),
        new BasicNameValuePair("password", "blah"))));

    RequestBuilder resetRequest = post("/users:resetPassword")
        .content(content)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED);

    mockMvc.perform(resetRequest)
        .andExpect(status().isNotFound());
  }

  @Test
  public void resetPasswordWithInvalidToken() throws Exception {
    configureServerForResetPasswordRequest();

    String content = EntityUtils.toString(new UrlEncodedFormEntity(Arrays.asList(
        new BasicNameValuePair("token", "blah"),
        new BasicNameValuePair("password", "???"))));

    RequestBuilder resetRequest = post("/users:resetPassword")
        .content(content)
        .contentType(MediaType.APPLICATION_FORM_URLENCODED);

    mockMvc.perform(resetRequest)
        .andExpect(status().isNotFound());
  }

  private void configureServerForResetPasswordRequest() {
    User user = userService.getUserByLogin(TEST_USER_LOGIN);
    user.setEmail("test@test.xyz");
    userService.updateUser(user);

    configurationService.setConfigurationValue(ConfigurationElementType.MINERVA_ROOT, "http://localhost:8080/minerva/");
  }

  @Test
  public void testGetUsers() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder grantRequest = get("/users/")
        .session(session);

    mockMvc.perform(grantRequest)
        .andDo(document("user/list_users",
            userColumListRequestParam(),
            responseFields(fieldWithPath("[]")
                .description("list of available input formats")
                .type("array")).andWithPrefix("[].", getUserResponseFields())))
        .andExpect(status().is2xxSuccessful());
  }

  private List<FieldDescriptor> getUserResponseFields() {
    return Arrays.asList(
        fieldWithPath("login")
            .description("user login")
            .optional()
            .type("string"),
        fieldWithPath("connectedToLdap")
            .description("is user account connected to ldap")
            .optional()
            .type("boolean"),
        fieldWithPath("ldapAccountAvailable")
            .description("does is account exist in ldap")
            .optional()
            .type("boolean"),
        fieldWithPath("email")
            .description("email address")
            .optional()
            .type("string"),
        fieldWithPath("id")
            .description("user unique id")
            .optional()
            .type("number"),
        fieldWithPath("maxColor")
            .description("color used for drawing data overlays with max value")
            .optional()
            .type("color"),
        fieldWithPath("minColor")
            .description("color used for drawing data overlays with min value")
            .optional()
            .type("color"),
        fieldWithPath("neutralColor")
            .description("color used for drawing data overlays with 0 value")
            .optional()
            .type("color"),
        fieldWithPath("simpleColor")
            .description("color used for drawing data overlays without value")
            .optional()
            .type("color"),
        fieldWithPath("name")
            .description("first name")
            .optional()
            .type("string"),
        fieldWithPath("surname")
            .description("last name")
            .optional()
            .type("string"),
        fieldWithPath("removed")
            .description("is the account removed")
            .optional()
            .type("boolean"),
        fieldWithPath("termsOfUseConsent")
            .description("did user agree to terms of use")
            .optional()
            .type("boolean"),
        fieldWithPath("privileges")
            .description("list of user privileges")
            .optional()
            .type("array"),
        fieldWithPath("privileges[].objectId")
            .description("object id to which project has privilege")
            .optional()
            .type("string"),
        fieldWithPath("privileges[].privilegeType")
            .description(
                "type of privilege, available values: " + projectSnippets.getOptionsAsString(PrivilegeType.class))
            .optional()
            .type("string"));
  }

  private List<ParameterDescriptor> getUserRequestFieldsWithoutLogin() {
    return Arrays.asList(
        parameterWithName("password")
            .optional()
            .description("user password"),
        parameterWithName("connectedToLdap")
            .optional()
            .description("is user account connected to ldap"),
        parameterWithName("ldapAccountAvailable")
            .optional()
            .description("does is account exist in ldap"),
        parameterWithName("email")
            .optional()
            .description("email address"),
        parameterWithName("maxColor")
            .optional()
            .description("color used for drawing data overlays with max value"),
        parameterWithName("minColor")
            .optional()
            .description("color used for drawing data overlays with min value"),
        parameterWithName("neutralColor")
            .optional()
            .description("color used for drawing data overlays with 0 value"),
        parameterWithName("simpleColor")
            .optional()
            .description("color used for drawing data overlays without value"),
        parameterWithName("name")
            .optional()
            .description("first name"),
        parameterWithName("surname")
            .optional()
            .description("last name"));
  }

  @Test
  public void updateProjectUploadPreferences() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    Map<String, Object> data = new HashMap<>();
    Map<String, Object> preferences = new HashMap<>();
    data.put("preferences", preferences);
    Map<String, Object> uploadPreferences = new HashMap<>();
    preferences.put("project-upload", uploadPreferences);
    uploadPreferences.put("validate-miriam", true);
    uploadPreferences.put("annotate-model", true);
    uploadPreferences.put("cache-data", true);
    uploadPreferences.put("auto-resize", true);
    uploadPreferences.put("semantic-zooming-contains-multiple-overlays", true);
    uploadPreferences.put("sbgn", true);

    String body = new Gson().toJson(data);

    RequestBuilder request = patch("/users/{login}:updatePreferences", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/update_preferences_1",
            pathParameters(parameterWithName("login").description("user login")),
            updatePreferencesRequest(),
            getUserPreferencesResponseFields()))
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      User user = userService.getUserByLogin(TEST_USER_LOGIN);
      assertTrue(user.getAnnotationSchema().getValidateMiriamTypes());
      return null;
    });

  }

  @Test
  public void updateElementAnnotatorsPreferences() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    Map<String, Object> data = new HashMap<>();
    Map<String, Object> preferences = new HashMap<>();
    data.put("preferences", preferences);
    Map<String, Object> elementAnnotators = new HashMap<>();
    preferences.put("element-annotators", elementAnnotators);
    List<Object> annotators = new ArrayList<>();
    elementAnnotators.put("lcsb.mapviewer.model.map.species.Gene", annotators);
    Map<String, Object> uniprotAnnotator = new HashMap<>();
    annotators.add(uniprotAnnotator);
    uniprotAnnotator.put("annotatorClass", UniprotAnnotator.class.getCanonicalName());
    List<Object> parameters = new ArrayList<>();
    uniprotAnnotator.put("parameters", parameters);
    parameters.add(createParameterClass("INPUT", 0, MiriamType.UNIPROT, null));
    parameters.add(createParameterClass("INPUT", 1, MiriamType.UNIPROT, "NAME"));
    parameters.add(createParameterClass("OUTPUT", 2, MiriamType.HGNC_SYMBOL, null));
    parameters.add(createParameterClass("OUTPUT", 3, MiriamType.UNIPROT, null));
    parameters.add(createParameterClass("OUTPUT", 4, MiriamType.EC, null));
    parameters.add(createParameterClass("OUTPUT", 5, MiriamType.ENTREZ, null));

    String body = new Gson().toJson(data);

    RequestBuilder request = patch("/users/{login}:updatePreferences", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/update_preferences_2",
            pathParameters(parameterWithName("login").description("user login")),
            updatePreferencesRequest(),
            getUserPreferencesResponseFields()))
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      User user = userService.getUserByLogin(TEST_USER_LOGIN);
      assertEquals(1, user.getAnnotationSchema().getAnnotatorsForClass(Gene.class).size());
      return null;
    });

  }

  @Test
  public void updateElementRequiredAnnotationsPreferences() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    Map<String, Object> data = new HashMap<>();
    Map<String, Object> preferences = new HashMap<>();
    data.put("preferences", preferences);
    Map<String, Object> elementAnnotators = new HashMap<>();
    preferences.put("element-required-annotations", elementAnnotators);
    Map<String, Object> annotations = new HashMap<>();
    elementAnnotators.put(Protein.class.getCanonicalName(), annotations);
    annotations.put("require-at-least-one", true);
    annotations.put("annotation-list", Arrays.asList(MiriamType.HGNC, MiriamType.HGNC_SYMBOL));

    String body = new Gson().toJson(data);

    RequestBuilder request = patch("/users/{login}:updatePreferences", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/update_preferences_3",
            pathParameters(parameterWithName("login").description("user login")),
            updatePreferencesRequest(),
            getUserPreferencesResponseFields()))
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      User user = userService.getUserByLogin(TEST_USER_LOGIN);
      assertTrue(user.getAnnotationSchema().requiresAtLeastOneAnnotation(Protein.class));
      return null;
    });
  }

  @Test
  public void updateElementValidAnnotationsPreferences() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    Map<String, Object> data = new HashMap<>();
    Map<String, Object> preferences = new HashMap<>();
    data.put("preferences", preferences);
    Map<String, Object> elementAnnotations = new HashMap<>();
    preferences.put("element-valid-annotations", elementAnnotations);
    elementAnnotations.put(Gene.class.getCanonicalName(), Arrays.asList(MiriamType.HGNC));

    String body = new Gson().toJson(data);

    RequestBuilder request = patch("/users/{login}:updatePreferences", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/update_preferences_4",
            pathParameters(parameterWithName("login").description("user login")),
            updatePreferencesRequest(),
            getUserPreferencesResponseFields()))
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      User user = userService.getUserByLogin(TEST_USER_LOGIN);
      assertTrue(user.getAnnotationSchema().requiresAtLeastOneAnnotation(Protein.class));
      return null;
    });

  }

  @Test
  public void updateGuiPreferences() throws Exception {
    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    Map<String, Object> data = new HashMap<>();
    Map<String, Object> preferences = new HashMap<>();
    data.put("preferences", preferences);
    Map<String, Object> elementAnnotations = new HashMap<>();
    preferences.put("gui-preferences", elementAnnotations);
    elementAnnotations.put("admin-projects-datatable-order", "1-asc");

    String body = new Gson().toJson(data);

    RequestBuilder request = patch("/users/{login}:updatePreferences", TEST_USER_LOGIN)
        .content(body)
        .session(session);

    mockMvc.perform(request)
        .andDo(document("user/update_preferences_5",
            pathParameters(parameterWithName("login").description("user login")),
            updatePreferencesRequest(),
            getUserPreferencesResponseFields()))
        .andExpect(status().is2xxSuccessful());

    callInSeparateThread(() -> {
      User user = userService.getUserByLogin(TEST_USER_LOGIN);
      assertTrue(user.getAnnotationSchema().requiresAtLeastOneAnnotation(Protein.class));
      return null;
    });
  }

  private Map<String, Object> createParameterClass(String type, int order, MiriamType annotationType, String field) {
    Map<String, Object> result = new HashMap<>();
    result.put("type", type);
    result.put("order", order);
    if (annotationType != null) {
      result.put("annotation_type", annotationType.name());
    }
    result.put("field", field);
    return result;
  }

  private RequestFieldsSnippet updatePreferencesRequest() {
    RequestFieldsSnippet result = requestFields(
        fieldWithPath("preferences")
            .description("information about default preferences")
            .type("object"));
    for (FieldDescriptor fd : getProjectUploadResponseFields()) {
      if (fd.getDescription() != null) {
        result = result.andWithPrefix("preferences.",
            fieldWithPath(fd.getPath()).description(fd.getDescription()).type(fd.getType()).optional());
      }
    }
    for (FieldDescriptor fd : getElementAnnotatorsResponseFields()) {
      if (fd.getDescription() != null) {
        result = result.andWithPrefix("preferences.",
            fieldWithPath(fd.getPath()).description(fd.getDescription()).type(fd.getType()).optional());
      }
    }
    for (FieldDescriptor fd : getElementRequiredAnnotationResponseFields()) {
      if (fd.getDescription() != null) {
        result = result.andWithPrefix("preferences.",
            fieldWithPath(fd.getPath()).description(fd.getDescription()).type(fd.getType()).optional());
      }
    }
    for (FieldDescriptor fd : getElementValidAnnotationResponseFields()) {
      if (fd.getDescription() != null) {
        result = result.andWithPrefix("preferences.",
            fieldWithPath(fd.getPath()).description(fd.getDescription()).type(fd.getType()).optional());
      }
    }
    for (FieldDescriptor fd : getGuiPreferencesResponseFields()) {
      if (fd.getDescription() != null) {
        result = result.andWithPrefix("preferences.",
            fieldWithPath(fd.getPath()).description(fd.getDescription()).type(fd.getType()).optional());
      }
    }
    return result;
  }

  private ResponseFieldsSnippet getUserPreferencesResponseFields() {
    return responseFields(fieldWithPath("preferences").description("set of user preferences").type("object"))
        .andWithPrefix("preferences.", getElementAnnotatorsResponseFields())
        .andWithPrefix("preferences.", getElementRequiredAnnotationResponseFields())
        .andWithPrefix("preferences.", getElementValidAnnotationResponseFields())
        .andWithPrefix("preferences.", getGuiPreferencesResponseFields())
        .andWithPrefix("preferences.", getProjectUploadResponseFields())

    ;
  }

  private List<FieldDescriptor> getElementAnnotatorsResponseFields() {
    return Arrays.asList(
        fieldWithPath("element-annotators")
            .description("definition of element annotators used by user")
            .type("object"),
        fieldWithPath("element-annotators['*']")
            .description(
                "list of annotators used for specific BioEntity type. Type of BioEntity is a class name that represents specific type in the system, for example: "
                    + BioEntity.class.getName() + ", " + Complex.class.getName())
            .type("array"),
        fieldWithPath("element-annotators['*'][].id")
            .description("id")
            .ignored()
            .type("number"),
        fieldWithPath("element-annotators['*'][].annotatorClass")
            .description("type of annotator")
            .type("string"),
        fieldWithPath("element-annotators['*'][].order")
            .description("defines order in which annotators should be used")
            .type("number"),
        fieldWithPath("element-annotators['*'][].parameters")
            .description("list of parameters for the annotator")
            .type("object"),
        fieldWithPath("element-annotators['*'][].parameters[].type")
            .description(
                "is this parameter INPUT/OUTPUT/CONFIG type. "
                    + "INPUT types define what should be used to identify the element. "
                    + "OUTPUT parameter defines what should be written to the element.")
            .type("string"),
        fieldWithPath("element-annotators['*'][].parameters[].annotation_type")
            .description("specifies annotation type to be used (if empty then field will be used)")
            .optional()
            .type("string"),
        fieldWithPath("element-annotators['*'][].parameters[].field")
            .description("specifies field to be used")
            .optional()
            .type("string"),
        fieldWithPath("element-annotators['*'][].parameters[].order")
            .description(
                "input parameters are checked in order. First match identifies the element. This field defines what is the order of INPUT parameters.")
            .optional()
            .type("number")

    );
  }

  private List<FieldDescriptor> getElementRequiredAnnotationResponseFields() {
    return Arrays.asList(
        fieldWithPath("element-required-annotations")
            .description("definition of required annotation by BioEntity type.")
            .type("object"),
        fieldWithPath("element-required-annotations['*']")
            .description(
                "definition of required annotations for specific BioEntity type")
            .type("object"),
        fieldWithPath("element-required-annotations['*'].require-at-least-one")
            .description("is at least one annotation required")
            .type("boolean"),
        fieldWithPath("element-required-annotations['*'].annotation-list")
            .description("list of annotation types from which at least one is required")
            .type("array"));
  }

  private List<FieldDescriptor> getElementValidAnnotationResponseFields() {
    return Arrays.asList(
        fieldWithPath("element-valid-annotations")
            .description("definition of valid annotation by BioEntity type.")
            .type("object"),
        fieldWithPath("element-valid-annotations['*'][]")
            .description(
                "list of annotation types that is valid for specific BioEntity type")
            .type("array"));
  }

  private List<FieldDescriptor> getGuiPreferencesResponseFields() {
    return Arrays.asList(
        fieldWithPath("gui-preferences")
            .description("definition of user gui preferences.")
            .type("object"),
        fieldWithPath("gui-preferences.*")
            .optional()
            .description(
                "key-value definition of gui preference used for remembering default sort order, default pagination, etc")
            .type("string"));
  }

  private List<FieldDescriptor> getProjectUploadResponseFields() {
    return Arrays.asList(
        fieldWithPath("project-upload")
            .description("definition of preferences for project-upload.")
            .type("object"),
        fieldWithPath("project-upload.annotate-model")
            .description("annotate model after it is uploaded.")
            .type("boolean"),
        fieldWithPath("project-upload.auto-resize")
            .description("auto resize map after parsing")
            .type("boolean"),
        fieldWithPath("project-upload.cache-data")
            .description("cache all the data for project")
            .type("boolean"),
        fieldWithPath("project-upload.sbgn")
            .description("visualize project in sbgn-like view")
            .type("boolean"),
        fieldWithPath("project-upload.semantic-zooming-contains-multiple-overlays")
            .description("display each semantic zoom level in separate overlay.")
            .type("boolean"),
        fieldWithPath("project-upload.validate-miriam")
            .description("validate all miriam annotations")
            .type("boolean"));
  }

}
